#
# sip_request.tcl
# procedures to process an incoming SIP request
# - read the request
# - check if requested media is supported
# - notify callee about incoming request
# - create and send SIP response
# 



#
# checks incoming local multicast packet whether it's a request
# or if we have received our own response
#
proc read_sip_request {msg} {

  set result [catch {exec id} id]
  set id2 $id
  set id [string trim [string range $id2 [string first ( $id] [string first ) $id]] "()"]
  if {[string match "INVITE $id*" $msg] || [string match "INVITE sip://$id*" $msg]} {
    foreach line [split $msg "\t"] {
      if {[string match "CALL-ID:*" [string toupper $line]] || [string match "I:*" [string toupper $line]]} {
          set call_id [string trim [lindex [split $line ":"] 1]]
      }
    }
    process_sip_request $call_id [split $msg "\t"]
  }
} ;# read_sip_request



#
# handler for incoming requests
# calls procedures to 
# - check requested media
# - notify the called user
# - create & send a reply packet
#
proc process_sip_request {call_id msg} {
  global request mailcap_subtypes


  set_mailcap_subtypes

  if [sip_extract_parameters $call_id request $msg] {
    return
  }
    
  # a new request was received, send a "Trying" response to 
  # indicate sipd that isc has received the request
  set request($call_id,reply_code) 100
  set request($call_id,reason_phrase) Trying
  send_sip_reply $call_id
  # check whether the requested media is supported
  if ![check_requested_media $call_id] {
    set request($call_id,reply_code) 150
    set request($call_id,reason_phrase) Ringing
    send_sip_reply $call_id
    notify_callee $call_id
  } else {
    send_sip_reply $call_id
  }
} ;# process_sip_request



#
# check if requested media is supported
# returns "0" if media is supported, "1" otherwise
#
proc check_requested_media {call_id} {
  global request mailcap_rules rtp_list mailcap_types

    mailcap_load
    set mailcap_types [array names mailcap_rules]
    
  set i -1
  foreach media $request($call_id,media_list) {
    incr i
    # create a list containing supported encodings
    set m_rules ""
    foreach rule $mailcap_rules([lindex $media 0]) {
	lappend m_rules [lindex $rule 0]
    }


    foreach encoding [lindex $media 4] {
	if {[lsearch $m_rules $encoding] == -1} {




	    # coding not defined in mailcap
	    set request($call_id,reply_code) 406
	    set request($call_id,reason_phrase) "Not Accaptable"
	    lappend request($call_id,reason) 406.3
	    # replace requested codings by supported ones
	    set new_coding ""
	    foreach encoding $m_rules {
		set encoding [lindex [get_code [get_profile $encoding]] 0]
		if {$encoding != ""} {
		    lappend new_coding $encoding
		}
	    }
	    
	    set new_media [lreplace $media 4 4 $new_coding]
	    set request($call_id,media_list) [lreplace $request($call_id,media_list) $i $i $new_media]
	    break
	}
    }  
  }
    if {$request($call_id,reply_code) == 406} {
	return 1
    } 
    return 0
} ;# check_requested_media



proc ring {call_id} {
  global env request
 
  sip_sound ring.au
  set request($call_id,ringing) [after 5000 "ring $call_id"]
} ;# ring 



#
# inform user about an incoming session invitation
#
proc notify_callee {call_id} {
  global request image SIP conf play

  if ![file exists "/var/tmp/pmm_audioctl.pid"] {
      #start audio control toll from the mailcap entry
	if {[catch { set aid [application audioctl/*]} msg]} {
	    puts stderr "$msg"
	}
  	#catch {exec pmm_audioctl &}
  }
  set w .[noperiod $call_id]
  toplevel $w 
  wm title $w "Incoming Call !"
  wm geometry $w +200+200

    if {$play(port) != "invalid"}  {
	set play(old_port) $play(port)
    }
    if {$play(gain) != "0"}  {
	set play(old_gain) $play(gain)
    }
    control_send "$conf(name)/au_ctrl/* set play(port) speaker"
    control_send "$conf(name)/au_ctrl/* set play(gain) 50"
    ring $call_id


  frame $w.head
  pack \
   [label $w.head.im -image $image(invite)] \
      [label $w.head.title -font *-*-*-bold-i-*-18-*-*-*-*-*-*-* -text "Incoming Call !"] \
   -side left -anchor w  
 
  frame $w.body -relief ridge -borderwidth 4

  pack \
    [label $w.body.label -text "You have received a conference invitation with the following parameters:\n"] \
    [label $w.body.sender -text "Sender: $request($call_id,from)"] \
    -side top -anchor nw

  if {[info exists request($call_id,information)]} {
    pack \
      [label $w.body.info -text "Subject: $request($call_id,information)"] \
      -side top -anchor nw
  }

  pack \
   [label $w.body.label2 -text "\nRequested media types:"] \
   -side top -anchor nw
    set i 0
  foreach entry $request($call_id,media_list) {
      incr i
      set media [lindex $entry 0]
      #set port [lindex $entry 1]
      #set session_addr [lindex $entry 2]
      #set code [lindex $entry 4]
      set info [lindex $entry 5]
  
      frame $w.body.f$i
      frame $w.body.fi$i
      pack \
	  [label $w.body.f$i.img -image $image($media)]\
          [label $w.body.f$i.l -text [string toupper $media]]\
	  -side left -anchor w
      
      if {$info != ""} {
	  pack \
	      [label $w.body.fi$i.info -text "Subject: $info"] \
	      -side top -anchor nw
      }
      pack $w.body.f$i -side top -anchor w
      pack $w.body.fi$i -side top -anchor w
  }

#   frame $w.buttons1 
#   pack \
#     [button $w.buttons1.accept -text "Additional Info" -command "show_additional_info $call_id"] -expand 1

  frame $w.buttons2 
  pack \
      [button $w.buttons2.accept -text "Accept" -command "set request($call_id,reply_code) 200; set request($call_id,reason_phrase) OK; send_sip_reply $call_id"] \
      [button $w.buttons2.reject -text "Reject" -command "specify_reason_phrase $call_id"] \
    -side left -expand 1

  pack $w.head 
  pack $w.body -padx 1m -pady 1m -fill x
  #pack $w.buttons1 $w.buttons2 -side top
  pack $w.buttons2 -side top

  # give callee time to reply
  set request($call_id,notify_timer) [after 60000 "notify_timeout $call_id"]
} ;# notify_callee



#
# invoked if callee doesn't react to a notified call
#
proc notify_timeout {call_id} {
  global request play conf

    set request($call_id,reply_code) 602
  set request($call_id,reason_phrase) "Not Currently Here"
  send_sip_reply $call_id
} ;# notify_timeout



#
# let the user specify a reason why he declines the call
# invoked when pressing the decline button
#
proc specify_reason_phrase {call_id} { 
  global request
   
  set w .[noperiod $call_id]
  if [winfo exists $w.reason_phrase_l] return
  set request($call_id,reason_phrase) ""
  set request($call_id,reply_code) 450
  pack \
    [label $w.reason_phrase_l -text "Please enter a reason\nwhy you decline the call:"] \
    [entry $w.reason_phrase_e -textvariable request($call_id,reason_phrase)] \
    [button $w.reason_phrase_b -text "Send" -command "send_sip_reply $call_id"] \
    -side top
} ;# specify_reason_phrase



#
# show callee data where he can get additional information about 
# the session the was invited to
#
proc show_additional_info {call_id} {
  global request image

  set w .[noperiod $call_id]_info
  if {[winfo exists $w]} return
  toplevel $w
  set title "Additional Session Information"
  wm title $w $title
    
  frame $w.title
  pack \
    [label $w.title.im -image $image(info) -anchor center] \
    [label $w.title.title -text $title -font *-*-*-bold-i-*-18-*-*-*-*-*-*-*] \
    -side left -anchor w

  frame $w.top -relief ridge -borderwidth 4

  foreach tag {Uri Email Phone} {
    set low_tag [string tolower $tag]
    if [info exists request($call_id,$low_tag)] {
      frame $w.top.$low_tag
      pack \
        [label  $w.top.$low_tag.l1 -text "$tag:" -width 8 -anchor w] \
        [label  $w.top.$low_tag.l2 -text "$request($call_id,$low_tag)" -width 40 -anchor w -relief sunken] \
        -side left -anchor nw

      pack $w.top.$low_tag 
    }
  }

  pack \
    $w.title \
    $w.top \
    [button $w.button -text Dismiss -command "destroy $w"] \
    -side top
} ;# show_additional_info



#
# show callee the SIP request payload
#
proc show_payload {call_id} {
  global request image

  set w .[noperiod $call_id]_payload
  if {[winfo exists $w]} return
  toplevel $w
  set title "Invitation Payload"
  wm title $w $title
  
  frame $w.title
  pack \
    [label $w.title.im -image $image(info) -anchor center] \
    [label $w.title.title -text $title -font *-*-*-bold-i-*-18-*-*-*-*-*-*-*] \
    -side left -anchor w

  frame $w.top -relief ridge -borderwidth 4

  set text ""
  foreach line $request($call_id,payload) {
    append text "$line\n"
  }
    
  pack \
    [label $w.top.l -text $text]
    
  pack \
    $w.title \
    $w.top \
    [button $w.button -text Dismiss -command "destroy $w"] \
    -side top
} ;# show_payload



#
# send a SIP reply to the caller
#
proc send_sip_reply {call_id} {
  global request SIP play conf

  catch {destroy .[noperiod $call_id]}

  # create reply
  set answer "$SIP(VERSION) $request($call_id,reply_code) $request($call_id,reason_phrase)\t\t"
  foreach via $request($call_id,via) {
    append answer "Via:$via\t\t"
  }
  append answer "From:$request($call_id,from)\t\t"
  append answer "To:$request($call_id,to)\t\t"
  append answer "Call-id:$call_id\t\t"
  append answer "Contact-host:[idstring %i]\t\t"
    
  switch $request($call_id,reply_code) {
    200 {
      # OK
      catch {unset SIP(started_by_invitation)} 
      lappend request(sessions_to_create) $call_id
      if {[llength $request(sessions_to_create)] == 1} {      
        sip_open_sessions
      }
    }
    406 {
      # Not Acceptable
      foreach reason $request($call_id,reason) {
        append answer "Reason:$reason\t\t"
      }
      append answer "Content-type:application/sdp\t\t"
      set payload "v=$request($call_id,version)\t\t"
      append payload "s=$request($call_id,session_name)\t\t"
      # append alternative descriptions
      if {[lsearch $request($call_id,reason) 406.3] != -1} {
        foreach entry $request($call_id,media_list) {
          set code_list [lindex $entry 4]
          # create profile_list
          set profile_list ""
          foreach c $code_list {
            lappend profile_list [get_profile $c]
          }
          set information [lindex $entry 5]

          append payload \
            "m=[lindex $entry 0] [lindex $entry 1] RTP/AVP $profile_list\t\t"
          append payload \
            "c=IN IP4 [lindex $entry 2]/[lindex $entry 3]\t\t"

          if {$information != ""} {
            append payload \
              "i=$information\t\t"
          } 
        }
      }
    }
  }

  # calculate payload length
  if {[info exists payload]} {
    append answer "Content-length:[string length $payload]\t\t\t\t"
    append answer $payload
  } else {
    append answer "Content-length:0\t\t\t\t"
  }
  regsub -all -- / $call_id _ id
  control_send "$id/sip/0 $answer"
  if {$request($call_id,reply_code) >= 200} {
      catch {after cancel $request($call_id,notify_timer)}
      catch {after cancel $request($call_id,ringing)}
      if {[info exists play(old_port)]} {
	  set play(port) $play(old_port)
	  control_send "$conf(name)/au_ctrl/* set play(port) $play(port)"
      }
      if {[info exists play(old_gain)]} {
	  set play(gain) $play(old_gain)
	  control_send "$conf(name)/au_ctrl/* set play(gain) $play(gain)"
      }

  }
  if {$request($call_id,reply_code) >= 300 && [info exists SIP(started_by_invitation)]} {
      # close the compund window first
      conference_close_comp_win
      exit
  }
} ;# send_sip_reply


