/*
 *  rtp.c --
 *
 *       This file provides functions for RTP protocol handling
 *
 *  (C) 1995 Bernd Deffner, GMD Fokus
 *      1996 Dorgham Sisalem
 *  Some of the functions are borrowed from nevot and from the RTP draft 
 *
 *  See the copyright file for copyright information
 */
#include "bit.h"        /* ROUND() */
#include "util.h"       /* strcpy_pad(), timeval_ntp32(), PAD() */
#include <sys/types.h>
#include <sys/time.h>   /* gettimeofday() */
#include <stdlib.h>     /* free() */
#include <netinet/in.h>
#include <memory.h>
#include <assert.h>
#include <math.h>
#include <tcl.h>
#include <tk.h>
#include "types.h"
#include "ansi.h"
#include "rtp.h"
#include "nevit.h"
#include "debug.h"
#include "xv.h"
#include "ac.h"
#include "bsd-endian.h"

#define MAX_MISORDER 100
#define MAX_DROPOUT  3000
#define MIN_SEQUENTIAL 3

static char rcsid[]  = "$Id: rtp.c,v 1.6 1998/08/05 16:14:56 stb Exp $";

extern Tcl_Interp *interp_;

static Tk_TimerToken RTCP_TimerToken[NUM_LEVELS]; /* token for the time handler of RTCP*/

/*
 *  Initialise the sequence number related fields of the member struct
 */
void init_seq(member_t *m, u_int16 seq, int level)
{
  m->rtp_s[level].base_seq = seq - 1;
  m->rtp_s[level].max_seq  = seq;
  m->rtp_s[level].bad_seq  = RTP_SEQ_MOD + 1;
  m->rtp_s[level].cycles   = 0;
  m->rtp_s[level].received = 0;
  m->rtp_s[level].received_prior = 0;
  m->rtp_s[level].expected_prior = 0;
}


/*
* Update cycle count based on new sequence number.
*
* small positive delta (0 < sdelta < MAX_DROPOUT):
*   new < old: wrap
*   new > old: 
*   both: just update
* small negative delta (0 < -sdelta < MAX_MISORDER):
*   misordered: do nothing
* big |delta| [or MAX_DROPOUT < udelta < RTP_SEQ_MOD - MAX_MISORDER]
*   check if bad
*/
static int update_seq(member_t *m, u_int16 seq, int level)
{
  u_int16 udelta = seq - m->rtp_s[level].max_seq;
  
  /*
   * Source is not valid until MIN_SEQUENTIAL packets with
   * sequential sequence numbers have been received.
   */
  if (m->rtp_s[level].probation) {
    /* packet is in sequence */
    if (seq == m->rtp_s[level].max_seq + 1) {
      m->rtp_s[level].probation--;
      m->rtp_s[level].max_seq = seq;
      if (m->rtp_s[level].probation == 0) {
	init_seq(m, seq,level);
	m->rtp_s[level].received++;
	return 1;
      }
    } else {
      m->rtp_s[level].probation = MIN_SEQUENTIAL - 1;
      m->rtp_s[level].max_seq = seq;
    }
    return 0;
  } else if (udelta < MAX_DROPOUT) {
    /* in order, with permissible gap */
    if (seq < m->rtp_s[level].max_seq) {
      /*
       * Sequence number wrapped - count another 64K cycle.
       */
      m->rtp_s[level].cycles += RTP_SEQ_MOD;
    }
    m->rtp_s[level].max_seq = seq;
    
  } else if (udelta <= RTP_SEQ_MOD - MAX_MISORDER) {
    /* the sequence number made a very large jump */
    if (seq == m->rtp_s[level].bad_seq) {
      /*
       * Two sequential packets -- assume that the other side
       * restarted without telling us so just re-sync
       * (i.e., pretend this was the first packet).
       */
      init_seq(m, seq, level);
    }
    else {
      m->rtp_s[level].bad_seq = (seq + 1) & (RTP_SEQ_MOD-1);
      return 0;
    }
  } else {
    /* duplicate or reordered packet */
    m->rtp_s[level].duplicate++;
    return 0;
  }
  m->rtp_s[level].received++;
  return 1;
} /* update_seq */


/*
 *  Increase timestamp, now we just make a simple  addition
 */
u_int32 RTP_inc_ts(u_int32 old, u_int32 add)
{
  return(old + add);
}

/*
 * check whether some members stopped reporting, if yes delete them
 * otherwise set the activity bit back to 0 and schedule a new event
 */
void RTCP_activity_handler(ClientData session)
{
  session_t* s=(session_t *)session;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search; 
  member_t *m=NULL;
  for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (member_t *) Tcl_GetHashValue(entryPtr);
    if(m->cname&&strcmp(s->local->cname, m->cname)) {
      if(m->active==0) {
	if (s->passive) {
	  dotcl(interp_, "control_send \"$session_name/%s bye\"",
		m->cname);
	}
	delete_entry(s, m->ssrc, DELETE_THIS);
      }else {
	/* the min number of packets to receive before declaring m as active */
	if (m->active > 12 ) {
	  dotcl(interp_, "control_send \"$session_name/%s active\"",
		m->cname);
	} else {
	  dotcl(interp_, "control_send \"$session_name/%s !active\"",
		m->cname);
	}
	m->active=0;
      }
    }
  }
  /* start next event */
  Tcl_CreateTimerHandler((int) 30000,(Tcl_TimerProc *)RTCP_activity_handler,s);
}

/*
 * Write RTP control header
*/
void RTP_fill_hdr(session_t *s, char *buf, int marker, u_int32 ts, int level) 
{
  rtp_hdr_t *h;
  int pad = 0;
  int csrc = 0;

  h = (rtp_hdr_t *) buf;
  h->version = RTP_VERSION;
  h->p       = pad > 0;
  h->x       = 0;
  h->cc      = csrc;
  h->m       = marker > 0;
  h->pt      = s->local->payload_type;
  h->seq     = htons(s->local->rtp_s[level].seq);
  h->ts      = htonl(s->local->rtp_s[level].ts);
  h->ssrc    = htonl(s->local->ssrc);
  gettimeofday(&s->local->rtp_s[level].last[0], 0);/*set last data sent time */
  s->local->rtp_s[level].seq++;
  update_seq(s->local, s->local->rtp_s[level].seq, level);

}


/*
 * convenience function for searching an entry in the hash table
 * an entry is only returned if for the required level all initializations
 * were done, otherwise a NULL is returned to force an add_entry
 * with level=0 or 1  look for the basis level entry
 *
 * returns pointer to member or NULL 
*/
member_t *find_entry(session_t *s, u_int32 ssrc,int level)
{
  Tcl_HashEntry *entryPtr=NULL;
  member_t* m=NULL;
  entryPtr = Tcl_FindHashEntry(&s->member_table, (char*)(unsigned long)ssrc);
  if (entryPtr == NULL)
    return NULL;
  m=(member_t *)Tcl_GetHashValue(entryPtr);
  if(level&&(m->rtp_s[level].buf==NULL))
    return NULL;
  return m;
}

/*
 * Creates a new entry in the hash table with key ssrc and returns
 * a pointer to a dynamically allocated member structure
*/
member_t *add_entry(session_t *s, u_int32 ssrc, int level)
{
  int new;
  Tcl_HashEntry *entryPtr;
  member_t *m=NULL;
  int i;
  if((m=find_entry(s,ssrc,0))==NULL) { /* get the basis level entry */
    entryPtr = Tcl_CreateHashEntry(&s->member_table, (char*)(unsigned long)ssrc, &new);
    assert(new);
    if((m = (member_t *)malloc(sizeof(member_t)))==NULL) {
      perror(" Could not allocate member ");
      return(NULL);
    }
    memset(m, 0, sizeof(m));
    Tcl_SetHashValue(entryPtr, m);
    init_seq(m, 0, level);
    /* initialize the contents of the member structure */
    m->ssrc = ssrc;
    m->cname = m->name = m->email = m->loc = m->phone = m->tool = 
      m->note = m->img= NULL;
    if(s->passive)
      m->preview=0;
    else
      m->preview=1;
    m->theImage=NULL;
    m->theVisual=NULL;
    m->theGC=NULL;
    m->theDisp=NULL;
    m->mainW=NULL;
    m->shminfo=NULL;
    m->decoder=NULL;
    m->width=0;
    m->thumbnail_image=NULL;
    m->decoder_small=NULL;
    s->members++;
    m->active=1;
    if((m->image=malloc(MAX_BUF_LENGTH))==NULL){
      fprintf(stderr," no enough memory avilable -add_entry \n");
      Exit();
    }
    for(i=1;i<NUM_LEVELS;i++){
      int l=0;
      m->rtp_s[i].buf=NULL;
      m->rtp_s[i].probation = MIN_SEQUENTIAL;
      m->rtp_s[i].frames_sent=m->rtp_s[i].octets_sent=0;
      m->rtp_s[i].packets_last=m->rtp_s[i].octets_last=0;
      m->rtp_s[i].created = (u_int32) time(NULL);
      m->rtp_s[i].jitter = 0;
      m->rtp_s[i].seq=0;
      m->rtp_s[i].nbytes=0;
      m->rtp_s[i].ts=0;
      m->rtp_s[i].duplicate=0;
      m->rtp_s[i].rr_lost=m->rtp_s[i].lost=m->rtp_s[i].rr_fraction=
	m->rtp_s[i].fraction=m->rtp_s[i].rr_jitter=
	m->rtp_s[i].jitter=0;
      for(;l<3;l++)
	m->rtp_s[i].last[l].tv_sec=0;
      m->rtp_s[level].thumbnail=0;
    }
  }
  /* initialize the level specific variables */
  memcpy(&(m->rtp_s[level].rmt_addr),&(s->rmt_addr[level]),
	 sizeof(struct sockaddr_in));
  if(s->rmt_addr[level].sin_port%2!=0) {
    m->rtp_s[level].rmt_addr.sin_port--;}
  if((m->rtp_s[level].buf=calloc(MAX_BUF_LENGTH,1))==NULL) {
    fprintf(stderr," Not enough memory available \n");
    Exit();
  }
  m->rtp_s[level].probation = MIN_SEQUENTIAL;
  m->rtp_s[level].frames_sent=m->rtp_s[level].octets_sent=0;
  m->rtp_s[level].packets_last=m->rtp_s[level].octets_last=0;
  m->rtp_s[level].created = (u_int32) time(NULL);
  m->rtp_s[level].jitter = 0;
  m->rtp_s[level].duplicate=0;
  m->rtp_s[level].seq=0;
  m->rtp_s[level].rr_lost=m->rtp_s[level].lost=m->rtp_s[level].rr_fraction=
    m->rtp_s[level].fraction=m->rtp_s[level].rr_jitter=
    m->rtp_s[level].jitter=0;
  gettimeofday(&(m->rtp_s[level].update_time),0);

  /* initialize the app part */
  m->rtp_s[level].app.last_time.tv_sec=0;
  m->rtp_s[level].app.last_time.tv_usec=0;
  m->rtp_s[level].app.last_app.tv_sec=0;  
  m->rtp_s[level].app.last_app.tv_usec=0;  
  m->rtp_s[level].app.seq=0;
  m->rtp_s[level].app.bbw=0;

  /* initialize the bbw variables for this member */
  m->rtp_s[level].max_count=0;
  m->rtp_s[level].average_bbw=0;
  for(i=0;i<5;i++) {
    int l=0;
    for(;l<10;l++){
      m->rtp_s[level].classification[i][l].count=0;
      m->rtp_s[level].classification[i][l].average_bbw=0;
    }
  }
  return m;
    
}



/*
 * convenience function for deleting an entry in the hash table when there are
 * no more TK events to handle
 */

void delete_entry_idle(ClientData dataPtr)
{
  Tcl_HashEntry *entryPtr=dataPtr;
  member_t *m=NULL;      
  int i;
  m = (member_t *)Tcl_GetHashValue(entryPtr);

  if(m->theImage!=NULL) {
     end_TkStuff(m->cname); /* clear the window slot for this member */
  }     
  for(i=1;i<NUM_LEVELS;i++){
    if(m->rtp_s[i].buf) 
      free(m->rtp_s[i].buf);
    m->rtp_s[i].buf=NULL;
  }
  /* delete the thumbanil image */
  delete_thumbnail(m);
  Tcl_DeleteHashEntry(entryPtr);
  s->members--;
  /*free ((char *)m);
  m=NULL;*/
}

/* 
 * flag == DELETE_THIS   deletes an entry from the hash table and frees 
 *                       the allocated memory for the member entry 
 *
 * flag == DELETE_OTHERS deletes all entries from the hash table except 
 *                       the one specified
 * flag == DELETE_ALL    ssrc is ignored, deletes all members
 *
*/
void delete_entry(session_t *s, u_int32 ssrc, int flag)
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  member_t *m=NULL;
  int i;
  
  if(s->control)
    ac_delete_entry(ssrc);
  switch(flag) {
  case DELETE_THIS:
    entryPtr = Tcl_FindHashEntry(&s->member_table, (char*)(unsigned long)ssrc);
    if (entryPtr != NULL) {
      Tcl_DoWhenIdle((Tcl_IdleProc *)delete_entry_idle,entryPtr);
    }
    break;
  case DELETE_OTHERS:
  case DELETE_ALL:
    /* for all members */
    for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
	 entryPtr != NULL;
	 entryPtr = Tcl_NextHashEntry(&search)) {
      m = (member_t *) Tcl_GetHashValue(entryPtr);
      if (!(flag == DELETE_OTHERS && m->ssrc == ssrc)){
	  Tcl_DeleteHashEntry(entryPtr);
	  s->members--;
	  if(m->theImage!=NULL) {
	    end_TkStuff(m->cname);
	  }
	  for(i=1;i<NUM_LEVELS;i++)
	    if(m->rtp_s[i].buf){
	      free(m->rtp_s[i].buf) ;
	      m->rtp_s[i].buf=NULL;
	    }
          if(m->image) {
	    free(m->image);
	    m->image=NULL;
	  }
	  free ((char *)m);
	  m=NULL;
	}
    }
    for(i=1;i<NUM_LEVELS;i++)
      Tcl_DeleteTimerHandler(RTCP_TimerToken[i]);
    end_TkAll(0);
    break;
  } /* switch */
}


/*
 *  Clean up
 */
void RTP_close(session_t *s)
{
  Tcl_DeleteHashTable(&s->member_table);
}


struct {
  rtcp_sdes_type_t type;
  char *name;
  int offset;
} map[] = {
  {RTCP_SDES_CNAME, "CNAME", Offset(member_t,cname)},
  {RTCP_SDES_NAME,  "NAME",  Offset(member_t,name)},
  {RTCP_SDES_EMAIL, "EMAIL", Offset(member_t,email)},
  {RTCP_SDES_LOC,   "LOC",   Offset(member_t,loc)},
  {RTCP_SDES_PHONE, "PHONE", Offset(member_t,phone)},
  {RTCP_SDES_TOOL,  "TOOL",  Offset(member_t,tool)},
  {RTCP_SDES_IMG,  "IMAGE",  Offset(member_t,img)},
  {0,0}
};



/* 
 * called when RTP packet received. Processes only the header, updates the
 * jitter estimate
 */
member_t* RTP_receive(session_t *s, rtp_hdr_t *rtp_hdr, int level)
{
  static member_t *m=NULL;
  static u_int32 ssrc_cache = 0;
  u_int32 ssrc = ntohl(rtp_hdr->ssrc);
  u_int32 arrival, transit;
  double now;
  int d;
  static boolean first = 1;
  static u_int32 first_arrival;

  /* check if the received data is really video */  
  if(rtp_hdr->pt!=RTP_PT_YUV&&rtp_hdr->pt!=RTP_PT_JPEG) {
    fprintf(stderr,"encoding %i not supported!! \n",rtp_hdr->pt);
    return NULL;           
  }
    

  /* search only for member if ssrc of the actual packet has changed */
  if (ssrc_cache != ssrc) {
    if((m = find_entry(s, ssrc,level))==NULL) /* if not found wait till an 
						 RTCP packet arrives */
      return NULL;
    ssrc_cache = ssrc;
  }
  if((m==NULL)||(m->rtp_s[level].buf==NULL)){ 
                                /* buf =NULL indicates that this sender has 
				 * quit the session. but the Bye message 
				 * received before the last data packets, 
				 * so just ignore these late arrivals */
    return NULL;
  }
  
  /* set the activity bit on */
  m->active +=1;
  m->rtp_s[level].active = 1;
  gettimeofday(&m->rtp_s[level].last[0], 0);  /* set last data arrival time */

  /* update jitter estimate */
  now = m->rtp_s[level].last[0].tv_sec + m->rtp_s[level].last[0].tv_usec / 1E6;
  if (first) {
    first = 0;
    first_arrival = (u_int32) now;
  } else {
    arrival = (now - first_arrival) * 90000;
    transit = arrival - ntohl(rtp_hdr->ts);
    d = transit - m->rtp_s[level].transit;
    m->rtp_s[level].transit = transit;
    if (d < 0) d = -d;
    m->rtp_s[level].jitter += (1./16.) * ((double)d - m->rtp_s[level].jitter);
    debug(DEBUG_RTP, "now=%f arrival=%u transit=%u jitter=%f d=%d",
	  now, arrival, transit, m->rtp_s[level].jitter, d);
  }

 debug(DEBUG_RTP, "prob=%u base_seq=%u cycles=%u max_seq= %u bad_seq= %u", 
	m->rtp_s[level].probation, m->rtp_s[level].base_seq, 
	m->rtp_s[level].cycles, m->rtp_s[level].max_seq, 
	m->rtp_s[level].bad_seq);
  debug(DEBUG_RTP, "received=%u recv_prior=%u exp_prior=%u", 
	m->rtp_s[level].received, m->rtp_s[level].received_prior, 
	m->rtp_s[level].expected_prior);
  if(update_seq(m, ntohs(rtp_hdr->seq),(int) level))
    return m;
  return NULL;
}


/*
 *  Computation of the RTCP transmission interval
 */
double rtcp_interval(int members, int senders, double rtcp_bw,
		     int we_sent, int packet_size, int *avg_rtcp_size,
		     int initial)
{
  /*
   * Minimum time between RTCP packets from this site (in seconds).
   * This time prevents the reports from `clumping' when sessions
   * are small and the law of large numbers isn't helping to smooth
   * out the traffic.  It also keeps the report interval from
   * becoming ridiculously small during transient outages like a
   * network partition.
   */
  double const RTCP_MIN_TIME = 5.;
  /*
   * Fraction of the RTCP bandwidth to be shared among active
   * senders.  (This fraction was chosen so that in a typical
   * session with one or two active senders, the computed report
   * time would be roughly equal to the minimum report time so that
   * we don't unnecessarily slow down receiver reports.)  The
   * receiver fraction must be 1 - the sender fraction.
   */
  double const RTCP_SENDER_BW_FRACTION = 0.25;
  double const RTCP_RCVR_BW_FRACTION = (1-RTCP_SENDER_BW_FRACTION);
  double t;                   /* interval */
  double rtcp_min_time = RTCP_MIN_TIME;
  int n;                      /* number of members for computation */
  
  /*
   * Very first call at application start-up uses half the min delay
   * for quicker notification while still allowing some time before
   * reporting for randomization and to learn about other sources so
   * the report interval will converge to the correct interval more
   * quickly.  The average RTCP size is initialized to 128 octets
   * which is conservative (it assumes everyone else is generating
   * SRs instead of RRs: 20 IP + 8 UDP + 52 SR + 48 SDES CNAME).
   */
  if (initial) {
    rtcp_min_time /= 2;
    *avg_rtcp_size = 128;
  }
  
  /*
   * If there were active senders, give them at least a minimum
   * share of the RTCP bandwidth.  Otherwise all participants share
   * the RTCP bandwidth equally.
   */
  n = members;
  if (senders > 0 && senders < members * RTCP_SENDER_BW_FRACTION) {
    if (we_sent) {
      rtcp_bw *= RTCP_SENDER_BW_FRACTION;
      n = senders;
    } else {
      rtcp_bw *= RTCP_RCVR_BW_FRACTION;
      n -= senders;
    }
  }
  
  /*
   * Update the average size estimate by the size of the report
   * packet we just sent.
   */
  *avg_rtcp_size += (packet_size - *avg_rtcp_size) * RTCP_SIZE_GAIN;
  
  /*
   * The effective number of sites times the average packet size is
   * the total number of octets sent when each site sends a report.
   * Dividing this by the effective bandwidth gives the time
   * interval over which those packets must be sent in order to meet
   * the bandwidth target, with a minimum enforced.  In that time
   * interval we send one report so this time is also our average
   * time between reports.
   */
  t = (*avg_rtcp_size) * n / rtcp_bw;
  if (t < rtcp_min_time) t = rtcp_min_time;
  
  /*
   * To avoid traffic bursts from unintended synchronization with
   * other sites, we then pick our actual next report interval as a
   * random number uniformly distributed between 0.5*t and 1.5*t.
   */
  return t * (drand48() + 0.5);
}


/*
 *  processes one SDES entry, stores it in the member hash table and
 *  prints a warning if collision is detected
 */
void member_sdes(session_t *s, u_int32 ssrc, rtcp_sdes_type_t t, char *b,
		 int len, int level)
{
  int i;
  char **p=NULL;
  member_t *m=NULL;

  for (i = 0; map[i].name; i++) {
    if (map[i].type == t) break;
  }

  /* find or create member entry */
  if((m = find_entry(s, ssrc,level))==NULL)
    m = add_entry(s, ssrc, level);
  
  p = (char *)m + map[i].offset;

  /* enter sdes item only if it has changed or it is NULL */
  if (*p == NULL || strncmp(*p, b, len) != 0)
    {
      free(*p);
      *p=NULL;
      if((*p = malloc(len + 1))==NULL){
	fprintf(stderr," no enough memory available -proc member_sdes \n");
	Exit();
      }
      memcpy(*p, b, len);
      (*p)[len] = '\0';
      /* 
       * transmit pmm messages if we are controlled, transmit a hello
       * message if we received a CNAME
       */
      if (map[i].type == RTCP_SDES_CNAME){
        if (s->passive) {
	  dotcl(interp_, "control_send \"$session_name hello %s\"", *p);
	}
      } else {
	if (s->passive) {
	  if(m->cname!=NULL)
	    dotcl(interp_, "control_send \"$session_name/%s %s {%s}\"",
		  m->cname, map[i].name, *p);
	}
      }
    }
  if (s->local->ssrc == ssrc && t == RTCP_SDES_CNAME &&
      !strcmp(s->local->cname, b)) {
    /*
     * Collision Check 
     * we have to check the cname because we might receive our own 
     * RTCP packets
     */
    fprintf(stderr, "* SSRC Collision: received ssrc=%lu \n cname=%s", 
	    (unsigned long )ssrc, b);
  }
} /* member_ses */


/*
* Parse one SDES chunk.
*/
static char *rtp_parse_sdes(session_t *s, char *b, int len, int level)
{
  rtcp_sdes_item_t *rsp;
  u_int32 *src = (u_int32 *)b;

  rsp = (rtcp_sdes_item_t *)(b + 4);
  for (; rsp->type; rsp = (rtcp_sdes_item_t *)((char *)rsp + rsp->length + 2))
    {
      member_sdes(s, ntohl(*src), rsp->type, rsp->data, rsp->length, level);
      /* rtp_db_sdes_update(src, rsp, addr);  */
    }
  b = (char *)rsp;
  return b + ((4 - ((long)b & 0x3)) & 0x3);
}


/*
 *  this function stores the fields of the RR into a member structure 
 */
void rr_handler(member_t *m, rtcp_rr_t *rr, int level)
{
  struct timeval now;
 
  m->rtp_s[level].rr_fraction = rr->fraction;
  m->rtp_s[level].rr_lost = ntohl(SHIFT8L(rr->lost));
  m->rtp_s[level].rr_jitter = ntohl(rr->jitter);
  gettimeofday(&now, 0);
  m->rtp_s[level].rr_rtt=
    timeval_ntp32(&now)/65536.0-ntohl(rr->lsr)/65536.0-ntohl(rr->dlsr)/65536.0;
  if(s->debug)
    fprintf(stderr," receiver reports rtt=%f \n",m->rtp_s[level].rr_rtt);
  

}

/* Classify the calculated bbw 
 * take an array for 100kb/s, 1Mb/s 10Mb/s, 100Mb/s and 1Gb/s and devide it 
 * into 10 parts. */

double  RTCP_classify_bbw(member_t *m, int level, double bbw)
{
  double factors[5]= {100000, 1000000, 10000000, 100000000, 1000000000};
  int i;

  for(i=0;i<5;i++) {
    int result=bbw*10/factors[i];
    if(result>=0 && result<10) {
      m->rtp_s[level].classification[i][result].count++;
      m->rtp_s[level].classification[i][result].average_bbw=
	(((m->rtp_s[level].classification[i][result].count-1)*
	  m->rtp_s[level].classification[i][result].average_bbw)+bbw)/
	m->rtp_s[level].classification[i][result].count;
      if(m->rtp_s[level].classification[i][result].count>
	 m->rtp_s[level].max_count) {
	m->rtp_s[level].max_count=
	  m->rtp_s[level].classification[i][result].count;
	m->rtp_s[level].average_bbw=
	  m->rtp_s[level].classification[i][result].average_bbw;
      }
      return(m->rtp_s[level].average_bbw);
    }
  }
  return 0;
} 

/* claculate the bottleneckbandwidth */
void RTCP_calc_bbw(member_t *m,u_int16 seq,int len,int level)
{
  /* is this the packet we are waiting for */
  if(seq==m->rtp_s[level].app.seq) {
    gettimeofday(&m->rtp_s[level].app.last_time,0);  
    return;
  }
  if(seq==m->rtp_s[level].app.seq+1) {
    if(m->rtp_s[level].app.last_time.tv_sec > 0) {
      double factor=1000000.0;
      m->rtp_s[level].app.bbw= ((float)(len+28)*8)/
	((m->rtp_s[level].last[0].tv_sec-m->rtp_s[level].app.last_time.tv_sec)+
       ((m->rtp_s[level].last[0].tv_usec-m->rtp_s[level].app.last_time.tv_usec)/factor)); 
      m->rtp_s[level].app.bbw= RTCP_classify_bbw(m,level,m->rtp_s[level].app.bbw);
      
      /* set the time the last time an app packet was received to a negative 
       * value, in order to determine if the packet with 
       * seq=m->rtp_s[level].app.seq was received or lost 
       */
      m->rtp_s[level].app.last_time.tv_sec =-1; 
      m->rtp_s[level].app.last_time.tv_usec=-1;
    }
  }  
}

/* analyze the app packet received */
void RTCP_app_handle(session_t *s,rtcp_t *r,int len,int level)
{
  member_t *m=NULL;
  rtcp_app_rec_t *rr;

  rr=r->r.app.r.app_rec;
  if(!strncmp("MEAS",r->r.app.name,4)){
    if((m = find_entry(s, ntohl(r->r.app.src),level))==NULL)
      return;

    /* starting a measurement period*/
    m->rtp_s[level].app.seq=ntohl(r->r.app.r.app_send.seq); 
    rr=r->r.app.r.app_send.app_rec;
  } 
  if(!strncmp("MEAR",r->r.app.name,4)||(ntohs(r->common.count)>0)){
    int i=0;
    for(;i<ntohs(r->common.count);i++) {
      if(s->local->ssrc==ntohl(rr->ssrc)) {
	if(s->debug)
	  fprintf(stderr," bbw=%i source=%i \n",ntohl(rr->bbw),r->r.app.src);
	if(s->control)
	  ac_update_bbw(r->r.app.src,ntohl(rr->bbw));
      }
      rr++;
    }
  }
    
}


/*
 * Decode an RTCP packet.
 */
void RTCP_receive(session_t *s, char *buf, int len, int level) {
  rtcp_t *r; 
  int i,length;
  struct timeval now;
  member_t *m=NULL;
  
  length=len;
  r = (rtcp_t *) buf;
  gettimeofday(&now, 0);
  
  debug(DEBUG_RTCP, "* %8ld.%06ld %4d RTCP: ", now.tv_sec, now.tv_usec, len);
  if (r->common.version != 2) {
    debug(DEBUG_RTCP, "Whoops.  Non-RTP packet.");
    return;
  }
  if (len > 0 && ntohl(r->r.sr.ssrc) != s->local->ssrc) {
    s->avg_rtcp_size[level] += (len + 28 - s->avg_rtcp_size[level]) *
      RTCP_SIZE_GAIN;
  }
  while (len > 0) {
    len -= (ntohs(r->common.length) + 1) << 2;
    if (len < 0)
      {
	/* something wrong with packet format */
	fprintf(stderr, "Illegal length %d", r->common.length);
	break;
      }
    debug(DEBUG_RTCP, "p=%d  cc=%d  len=%d  ", 
	  r->common.p, r->common.count, r->common.length);
    switch (r->common.pt) {
    case RTCP_SR:   /*Sender Report*/
      if (1) {
	m = find_entry(s, ntohl(r->r.sr.ssrc),level);
	if (m == NULL) {
	  m = add_entry(s, ntohl(r->r.sr.ssrc), level);
	}
	m->rtp_s[level].lsr_sec  = ntohl(r->r.sr.ntp_sec);
	m->rtp_s[level].lsr_frac = ntohl(r->r.sr.ntp_frac);
	gettimeofday(&m->rtp_s[level].last[2], 0);    /* set time of last SR */
	m->rtp_s[level].last[1] = m->rtp_s[level].last[2];           
	/* set time of last control packet*/
	
        debug(DEBUG_RTCP, "SR  %0x", r->r.sr.ssrc);
        debug(DEBUG_RTCP, "SR ntp=%u.%u ts=%u psent=%u osent=%u \n",
	      ntohl(r->r.sr.ntp_sec),
	      ntohl(r->r.sr.ntp_frac),
	      ntohl(r->r.sr.rtp_ts),
	      ntohl(r->r.sr.psent),
	      ntohl(r->r.sr.osent));
        for (i = 0; i < r->common.count; i++) {
	  if (s->local->ssrc == ntohl(r->r.sr.rr[i].ssrc)) {
	    /* register RR */
	    rr_handler(m, &r->r.sr.rr[i],level);
	    /* update  the application controller */
	    if(s->control){
	      s->local->rtp_s[level].bandwidth= s->grabber.bw=
		ac_qos_control(m->ssrc,m->rtp_s[level].rr_fraction/(double)256,
			       m->rtp_s[level].rr_rtt,s->grabber.bw);
		if(s->grabber.bw<1)
			s->local->rtp_s[level].bandwidth=
			  s->grabber.bw=1;  /* minimum bandwidth allowed*/
		if(s->debug)
		  fprintf(stderr," bw= %i \n", s->grabber.bw);
	    }
	  }
	}
      }
      break;
    case RTCP_RR: /* Receiver Report */
      m = find_entry(s, ntohl(r->r.rr.ssrc),level);
      if (m == NULL) { 
	m = add_entry(s, ntohl(r->r.rr.ssrc), level);
      }
      gettimeofday(&m->rtp_s[level].last[1], 0);      /* set time of last 
						       * control packet */
      for (i = 0; i < r->common.count; i++) {
	if (s->local->ssrc == ntohl(r->r.rr.rr[i].ssrc)) {
	  /* register RR */
	  rr_handler(m, &r->r.rr.rr[i], level);
	  /* application controller */
	  if(s->control){
	    s->local->rtp_s[level].bandwidth= s->grabber.bw=
	      ac_qos_control(m->ssrc, m->rtp_s[level].rr_fraction/(double)256,
			     m->rtp_s[level].rr_rtt,s->grabber.bw);
		if(s->grabber.bw<1)
		  s->local->rtp_s[level].bandwidth=
		    s->grabber.bw=1;  /* minimum bandwidth allowed*/
		if(s->debug)
		  fprintf(stderr," bw= %i \n", s->grabber.bw);
            }
	}
      }
      break;
    case RTCP_BYE: /* leaving a connection */
      for (i = 0;  i < r->common.count; i++) {
	m = find_entry(s, ntohl(r->r.bye.src[i]),level);
	if (m != NULL) {
	  /* announce BYE of this member */
	  if (s->passive) {
	    dotcl(interp_, "control_send \"$session_name/%s bye\"",
		  m->cname);
	  }
	  delete_entry(s, m->ssrc, DELETE_THIS);
	}
      }
      break;
    case RTCP_SDES: /* sdes packet */
      debug(DEBUG_RTCP, "SDES");
      buf = (char*)&r->r.sdes;
      for (i = 0; i < r->common.count; i++) {
        buf = rtp_parse_sdes(s, buf, len, level);
      } 
      break;
    case RTCP_APP:		/* Uh.. Nothing here. */
      RTCP_app_handle(s,r,length,(int)level);
      debug(DEBUG_RTCP, "Got app-specific report.  Okay...");
      break;
    default:
      /* invalid type */
      debug(DEBUG_RTCP, "Invalid RTCP Type: %d", r->common.pt);       
      break;
    }
    r = (rtcp_t *)((u_int32 *)r + ntohs(r->common.length) + 1);
  }
  /* set the activity bit on */
  if(m!=NULL) m->active+=1;

}


/*
 * Creates a RTCP SDES packet
 */
static int rtp_write_sdes(session_t *s, u_int32 ssrc, rtcp_t *r)
{
  int i, pad;
  int len;
  member_t *m=NULL;
  rtcp_sdes_item_t *rsp;
  int text_len;
  char **p;

  m = find_entry(s, ssrc,0);
  assert(m);
  len = 8;
  r->r.sdes.src = htonl(ssrc);
  /* skip generic header and ssrc */
  rsp = (rtcp_sdes_item_t *) ((char *)r + 8); 
  r->common.count++;
  for (i = 0; map[i].type; i++) {
    rsp->type = map[i].type;
    p = (char *)m + map[i].offset;
    text_len = strlen(*p);
    if (!text_len) continue;             /* skip empty fields */
    rsp->length = text_len; 
    strcpy(rsp->data, *p);
    debug(DEBUG_RTCP, "%x %s(%d): %s", ssrc, map[i].name, i, *p);
    len += 2 + text_len;
    rsp = (rtcp_sdes_item_t *)((char *)rsp + 2 + text_len);
  }
  /* terminate with zero byte */
  memset(&((char *)r)[len++], 0, 1);
  /* if necessary, pad with zeroes to next 4-byte boundary */
  pad = PAD(len, 4);
  memset(&((char *)r)[len], 0, pad);
  len += pad;
  return len / 4;
}

/* calculate the number of lost packets */
void rtp_calc_loss(rtp_t *rtp_s)
{
  u_int32 extended_max, expected, lost_interval, expected_interval;
  u_int32 received_interval;
  extended_max = rtp_s->cycles + rtp_s->max_seq;
  expected = extended_max - rtp_s->base_seq + 1;
  rtp_s->lost = expected - rtp_s->received;
  expected_interval = expected - rtp_s->expected_prior;
  received_interval =
    rtp_s->received - rtp_s->received_prior;
  lost_interval = expected_interval - received_interval;
  if (expected_interval == 0 || lost_interval <= 0)
    rtp_s->fraction = 0;
  else rtp_s->fraction = (lost_interval* 256) / expected_interval;
  return;
}

/*
* Create SR or RR packet. If the station has neither sent nor
* received data, send empty RR.
*/
static int rtp_write_traffic(session_t *s, rtcp_t *r, int level)
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  member_t *m=NULL;
  rtcp_rr_t *rr;
  int len = 1;
  struct timeval now;
  u_int32 now_ntp32;
  s->senders[level] = 0;
  /* sender report */
  r->common.count = 0;
  /* set time of last control packet */
  gettimeofday(&s->local->rtp_s[level].last[1], 0);   

  if (s->local->rtp_s[level].active > 0) {
    s->senders[level]++;
    r->common.pt = RTCP_SR;
    r->r.sr.ssrc = htonl(s->local->ssrc);
    /* active is cleared after RTCP interval computation */
    r->r.sr.rtp_ts    = htonl(s->local->rtp_s[level].ts);
    gettimeofday_ntp(&r->r.sr.ntp_sec);  /* returns two words */
    r->r.sr.ntp_sec   = htonl(r->r.sr.ntp_sec);
    r->r.sr.ntp_frac  = htonl(r->r.sr.ntp_frac);
    r->r.sr.psent     = htonl(s->local->rtp_s[level].frames_sent);
    r->r.sr.osent     = htonl(s->local->rtp_s[level].octets_sent);
    rr = r->r.sr.rr; 
    len += 6;
    debug(DEBUG_RTCP, "Local Site: SR  %0x  "
	  "ntp=%u.%u ts=%u psent=%u osent=%u",
	  ntohl(r->r.sr.ssrc), 
          ntohl(r->r.sr.ntp_sec),
          ntohl(r->r.sr.ntp_frac),
	  ntohl(r->r.sr.rtp_ts),
          ntohl(r->r.sr.psent),
          ntohl(r->r.sr.osent));
    gettimeofday(&s->local->rtp_s[level].last[2],0);/*set time of last SR*/
    /* set the number of sent packets pack to 0 */
    s->local->rtp_s[level].octets_sent=0;
  }
  else {
    r->common.pt = RTCP_RR;
    r->r.rr.ssrc = htonl(s->local->ssrc);
    rr = r->r.rr.rr;
    len+=1;
  }
  /* receiver reports */
  gettimeofday(&now, 0);
  now_ntp32 = timeval_ntp32(&now);
  debug(DEBUG_RTCP, "Local Site: RR  %0x", s->local->ssrc);
  /* for all members */
  for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (member_t *) Tcl_GetHashValue(entryPtr);
    /* update statistics and generate RR */
    if (s->local->ssrc != m->ssrc && m->rtp_s[level].active > 0 &&m->cname) {
      u_int32 extended_max, expected, lost_interval, expected_interval;
      u_int32 received_interval;
      extended_max = m->rtp_s[level].cycles + m->rtp_s[level].max_seq;
      expected = extended_max - m->rtp_s[level].base_seq + 1;
      m->rtp_s[level].lost = expected - m->rtp_s[level].received;

      expected_interval = expected - m->rtp_s[level].expected_prior;
      m->rtp_s[level].expected_prior = expected;
      received_interval =
	m->rtp_s[level].received - m->rtp_s[level].received_prior;
      m->rtp_s[level].received_prior = m->rtp_s[level].received;
      lost_interval = expected_interval - received_interval;
      if (expected_interval == 0 || lost_interval <= 0)
	m->rtp_s[level].fraction = 0;
      else m->rtp_s[level].fraction = (lost_interval* 256) / expected_interval;
      m->rtp_s[level].active--;
      s->senders[level]++;
      rr->ssrc     = htonl(m->ssrc);
      rr->last_seq = htonl(m->rtp_s[level].cycles + m->rtp_s[level].max_seq);
      rr->lost     = htonl(SHIFT8L(m->rtp_s[level].lost));
      rr->fraction = m->rtp_s[level].fraction;
      rr->jitter   = htonl((u_int32)m->rtp_s[level].jitter);
      rr->lsr      = htonl((m->rtp_s[level].lsr_sec << 16) | 
			   (m->rtp_s[level].lsr_frac >> 16));
      /* force to zero if no SR received yet */
      if (m->rtp_s[level].last[2].tv_sec)
        rr->dlsr  = htonl(now_ntp32 - timeval_ntp32(&m->rtp_s[level].last[2]));
      else
        rr->dlsr  = 0;
      r->common.count++;
      len += 6;
     rr++;
    }
  }
  return len;
} /* rtp_write_traffic */

/*
 * write an application defined RTCP packet for measurement of the bottleneck
 * bandwdith
*/
static int rtp_write_app(session_t *s, rtcp_t *r, int level)
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  member_t *m=NULL;
  rtcp_app_rec_t *rr;
  int len = 1;
  s->senders[level] = 0;
  /* sender report */
  r->common.count = 0;

  if (s->local->rtp_s[level].active > 0) {
    r->common.pt = RTCP_APP;
    r->common.count=0;
    r->r.app.src=htonl(s->local->ssrc);
    sprintf(r->r.app.name,"%s","MEAS");
    /* write the seq num of the next data packet to send*/
    r->r.app.r.app_send.seq= htons(s->local->rtp_s[level].seq+1);
    rr=r->r.app.r.app_send.app_rec; 
    len+=3;
  }
  else {
    r->common.pt = RTCP_APP;
    r->common.count=0;
    r->r.app.src = htonl(s->local->ssrc);
    sprintf(r->r.app.name,"%s","MEAR");
    rr=r->r.app.r.app_rec;
    len+=2;
 }
 /* receiver measurement reports */
  /* for all members */
  for (entryPtr = Tcl_FirstHashEntry(&s->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (member_t *) Tcl_GetHashValue(entryPtr);
    if(m->rtp_s[level].app.bbw>0) {
      rr->ssrc=m->ssrc;
      rr->bbw=m->rtp_s[level].app.bbw;
      r->common.count++;
      len+=2;
      rr++;
      m->rtp_s[level].app.bbw=0;
    }
  }
  return len;
} /* rtp_write_app */


/*
* Create a control packet. Add items to end. For BYE, add reason.
*/
int RTP_write_control(session_t *s, int type, char *reason, int level)
{
  rtcp_t *r;
  int len=0;         /* length of the RTCP packet in bytes */
  int words=0;       /* length of one stacked packet in bytes */
  int bytes=0;
  
  static struct {
    int type;
    rtcp_type_t pt;  
  } list[] = {
    {P_traffic, RTCP_RR},
    {P_sdes,    RTCP_SDES}, 
    {P_bye,     RTCP_BYE}, 
    {P_measure, RTCP_APP},
    {0,0}
  };
  int i;     
  
  r = (rtcp_t *)s->local->rtp_s[level].buf;
  r->common.version = RTP_VERSION;
  r->common.p       = 0;
  
  len = 0;
  for (i = 0; list[i].type; i++) {
    if (type & list[i].type) {
      r->common.pt      = list[i].pt;
      switch(list[i].type) {
      case P_traffic:
        r->common.version = RTP_VERSION;
        words = rtp_write_traffic(s, r, level);
        break;
      case P_sdes:
        /* XXX if bridge, report on all sources */
        r->common.version = RTP_VERSION;
	r->common.pt      = list[i].pt;
        r->common.count  = 0;
	words = rtp_write_sdes(s, s->local->ssrc, r);
        break;
      case P_bye:
        r->common.version = RTP_VERSION;
        r->r.bye.src[0] = htonl(s->local->ssrc);
        bytes = (8 + strcpy_pad((char *)&r->r.bye.src[1], reason));
	r->r.bye.length = strlen(reason);
	words = bytes >> 2;
        r->common.count  = 1;
	debug(DEBUG_RTCP, "Local Source: BYE %x  reason=%s", s->local->ssrc,
	      reason);
	break;
      case P_measure:
        r->common.version = RTP_VERSION;
	words=rtp_write_app(s, r, level);
        break;
      }
      r->common.length = htons(words - 1);  
      len += words<<2;
      r = (rtcp_t *)((u_int8 *)r + words * 4);
    }
  }
  return len;
} /* rtp_write_control */



/*
 *  transmit a control packet
 */
void RTP_transmit_control(session_t *s, int type, char *reason, int level)
{
  int len, sent;
  len = RTP_write_control( s,type, reason, level);
  s->rtcp_pkt_size[level] = len + 28;   /* length including UDP/IP headers */

  
  sent = s->net.write(RTCP+level-1, s->local->rtp_s[level].buf, len, NULL, 0);
  if (len != sent) 
    perror("sent RTCP");
}

/* the handler procedures that calles a delayed handler when an RTCP message
 * is to be sent
 */
int rtcp_timer_handler(ClientData level)
{
  rtcp_timer_handler_delayed(level);
  return 1;  
}

/*
 *  RTCP interval timer handler. At the first time it is invoked the timer is
 *  set and no packet is generated
 *  dor data = -1 stop the timer
*/


int rtcp_timer_handler_delayed(int level)
{
  struct timeval t;
  static int first[] = {1,1,1,1,1,1,1,1,1,1};
  double secs;
  int next_rtcp;

  if ((int)level <=0) {   
    if(level==0){
      int i;
      for(i=0;i<NUM_LEVELS;i++){
	/* stop timer */
	Tcl_DeleteTimerHandler(RTCP_TimerToken[i]);
	first[i] = 1;
      }
      return 1;
    }
    /* stop timer */
    Tcl_DeleteTimerHandler(RTCP_TimerToken[abs((int)level)]);
    first[abs((int)level)] = 1;
    return 1;
  }
  
  secs = rtcp_interval(s->members,
		       s->senders[(int)level],
		       TARGET_RTCP_BANDWIDTH,
		       s->local->rtp_s[(int)level].active,
		       s->rtcp_pkt_size[(int)level],  
		       &s->avg_rtcp_size[(int)level],
		       first[(int)level]);
  debug(DEBUG_RTCP, "rtcp_bw=%f interval=%f senders=%d members=%d "
	"we_sent=%d packet_size=%d avg_rtcp_size=%d  first=%d", 
	(double)TARGET_RTCP_BANDWIDTH,
	secs,
	s->senders[(int)level],
	s->members,
	s->local->rtp_s[(int)level].active,
	s->rtcp_pkt_size[(int)level],  
	s->avg_rtcp_size[(int)level],
	first[(int)level]);

  t.tv_sec = (int)secs;
  t.tv_usec = (secs - (int)secs) * 1e6;
  next_rtcp=(t.tv_sec*1000)+(t.tv_usec/1000);
  if (first[(int)level]) 
    first[(int)level] = 0;
  else{
    RTP_transmit_control(s, P_traffic|P_measure|P_sdes, "", (int)level);    
  }

  s->local->rtp_s[(int)level].active = 0;
  Tcl_DeleteTimerHandler(RTCP_TimerToken[(int)level]);   
  RTCP_TimerToken[(int)level]=
	 Tcl_CreateTimerHandler((int) next_rtcp,
		       (Tcl_TimerProc *)rtcp_timer_handler,(ClientData)level); 
  return 1;
}


/* Initialize the global session structure */
void session_init(session_t *s, int level)
{
  /* initailize the data structure needed for controlling the grabbing 
   * operation */
  s->grabber.frames  = MAX_FRAME_RATE;
  s->grabber.bw      = 9999; /* bandwidth in kb/s */
  s->codec.quality   = 50;   /* qulaity factor for JPEG */
  s->local->width   = 320;  /* default picture width */
  s->local->height  = 240;  /* default picture length */
  s->grabber.pb      = 9999; /* peak bandwidth for ATM */
  s->grabber.mb      = 66;   /* mean bandwidth for  ATM */
  s->grabber.whichinput = 1; /* input video port the IN 1 input of the card*/
  s->local->preview = 0;    /* do not preview the sent data */

  /* initailize the network interface */
  s->net.network     = 0;    /* default network ip(0) or ATM(1) */
  s->net.connect     = 0;

  /* initialize the statistics structure to 0 */
  s->stats.frame_size = 0;   
  s->stats.bw = 0;
  s->stats.frame_rate = 0;
  
  /* misc init */
  s->passive = 0;                       /* passive mode for remote control */
  s->ssrc[level]=0;
  s->send[level]=0;                               /* default value; */
  s->control= 0;  /* do not use the application controller by default */
 
  gettimeofday(&(s->local->rtp_s[level].update_time),0);
}

/*
 *  init a sub session for the sender 
 */


/*
*  RTP initialisation
*/
member_t *RTP_init(session_t *s, int level)
{
  member_t *m=NULL;
  u_int32 ssrc;

  if(level==1){ /* init the basis structure of the sender */
    Tcl_InitHashTable(&s->member_table, TCL_ONE_WORD_KEYS);
    srand48(random32(0));  /* initialize random number generator */  
    
    ssrc = (long unsigned )random32((int)s);
    s->members = 0;   
    s->delay_rtcp=1000;
    s->local=NULL;
    s->rtcp_pkt_size[level] = 120;                      /* initial value */
    m = add_entry(s, ssrc, level);                      /* add local site */
    
    /* initialize the member structure of the local user */
    s->local = m;
    s->local->ssrc = ssrc;
    s->local->image=NULL;
    s->local->cname = idstring("%u@%i",100);
    s->local->name = idstring("%N", 100);
    s->local->email = idstring("%N@%3D", 100);
    s->local->loc = idstring("%o (%Z %z)", 100);
    s->local->phone = idstring("%e", 100);
    s->local->tool = "nevit";
    s->local->note = "";
    s->local->img = "";
    if((s->local->image = malloc(MAX_BUF_LENGTH))==NULL){ /* image buffer */
      fprintf(stderr," no enough memory available -RTP_init\n");
      Exit();
    }
    s->local->rtp_s[level].ts = random32((int)s);
    s->local->rtp_s[level].seq = random32((int)s);
    s->local->rtp_s[level].first.seq = s->local->rtp_s[level].seq;
    s->local->rtp_s[level].first.ts = s->local->rtp_s[level].ts;
    s->scale_factor=1;
    session_init(s, level);
    return(s->local);
  }
  /* init the structures for the sending entities of the sub sessions */
  add_entry(s, s->local->ssrc, level);
  s->local->rtp_s[level].ts = random32((int)s);
  s->local->rtp_s[level].seq = random32((int)s);
  s->local->rtp_s[level].first.seq = s->local->rtp_s[level].seq;
  s->local->rtp_s[level].first.ts = s->local->rtp_s[level].ts;
  s->local->rtp_s[level].thumbnail=0;
  s->local->thumbnail_image=NULL;
  s->local->decoder_small=NULL;
  return NULL;
}


/* 
 * check if the data of this ssrc are to be displayed or can simply be ignored
 * return 1 diplay the data stream
 * return 0 do not diplay the data stream
 */
 
int display_ssrc(session_t* s,u_int32 ssrc)
{

  if((s->receive->preview==1)||(s->receive->rtp_s[FIRST_LEVEL].thumbnail==1))
    return(1);
  if(s->receive->preview==0)
    if(s->receive->theImage) {
      end_TkStuff(s->receive->cname); /* clear the window slot of this 
				       * memeber */
      s->receive->preview=-1;
    }
  return(0);
}






