/*
 *  xil_capture.c --
 *
 *       This file provides functions for grabbing and compressing video
 *       images using the sun video card
 *
 *  (C) 1996 Dorgham Sisalem GMD Fokus
 *    
 *  Some of the functions are borrowed from the xil examples, therefore there 
 *  are lots of code handling MPEG even though they are still not
   supported by nevit.
 *
 *  See the copyright file for copyright information
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <tcl.h>
#include "types.h"
#include "rtp.h"
#include "nevit.h"
#include "xil_capture.h"
#include "vic_jpeg.h"
#include "xv.h"





static char rcsid[]  = "$Id: xil_capture.c,v 1.2 1998/04/06 08:24:36 dor Exp $";



typedef struct {
    int weight_i;
    int weight_p;
    int weight_b;
} RtvcMpeg1Weights;

#ifdef XIL /* start of XIL related part */
#include <xil/xil.h>

static xil_grab_parameters param;
XilImage tmp_display;
static xil_parameters exp_par;
static XilSystemState state=NULL;
static XilImage image=NULL, rtvc_image=NULL;

/************* Local procedures declaration ************************/


/* open th xil device and check if it is allowed to use it */
char* xil_open_device()
{
  if(state==NULL){
    state = xil_open();  /* open the video device */
    if (state == NULL) {
      fprintf(stderr, "unable to open xil library\n");
      return(NULL);
    }
    {
      XilDevice device;
      
      if (!(device = xil_device_create(state, "SUNWrtvc"))) {
	fprintf(stderr, "Unable to create a device object\n");
	xil_close(state);
	state=NULL;
	return(NULL);
      }
      xil_device_set_value(device, "DEVICE_NAME", (void *) param.devname);
      if (!(image = xil_create_from_device(state, "SUNWrtvc", device))) {
	fprintf(stderr, "ERROR: failed to open SUNWrtvc device\n");
	xil_close(state);
	state=NULL;
	return(NULL);
      }
      xil_device_destroy(device);
    }
    rtvc_image=image;
    return ((char *)image);
  }
  return ((char *)image);
}

/* open the video input */

int xil_open_input(unsigned int input)
{
  int format;
  if(state!=NULL) /* still too early for input initailization */
    {
      xil_end_grab();
      xil_open_device();
      xil_set_device_attribute(image, "PORT_V", (void *) input);
      xil_get_device_attribute(rtvc_image, "FORMAT_V", (void **) &format);
      if (format == 0) {
	fprintf(stderr,"Unknown video format,");
	fprintf(stderr," or no camera attached  .. exiting\n");
	return(-1);
      }
      return format;
    }
  return(0);
}

/*
 * initialize the grabber structure and open the video device and set the 
 * compression parameters
 * widht and height are ignored 
 */
char * xil_init_grab(unsigned  int *width_,
		     unsigned  int *height_,
		     unsigned  int quality,
		     int        input,
		     compressor_t compressor)
{
  XilMpeg1Pattern mpattern;
  int scalex = 1, scaley = 1;
  XilDataType datatype;
  XilImage scaled_image;
  XilCis cis=NULL;
  unsigned int width, height, nbands;
  struct timeval t0;
  const char *video_format;
  int bytes_for_first_frame;
  char *XIL_OK="-1";
  /*********************************/
  /* init grap parameters          */
  /*********************************/
  
  switch(compressor) {
  case NEVIT_JPEG: 
    param.do_cis = JPEG;
    param.cis_type = "Jpeg";
    break;
  case NEVIT_YUV:
    param.do_cis = UYVY;
    param.cis_type = "UYVY";
    break;
  default:
    fprintf(stderr,"{encoding not supported}");
    return NULL;
  }
  param.quality = quality;
  param.x = -1;
  param.y = -1;
  param.width = *width_;
  param.width=param.width>>1;
  param.height = *height_;
  param.height =  param.height>>1;
  param.scalex = 2;
  param.scaley = 2;
  param.windowname = (char *) malloc(32);
  param.max_buffer = 2;
  param.datafile = NULL;
  param.cellb_xy_addr = 0;
  param.polling = 0;
  param.pattern = "IIII";
  param.devname = NULL;

  /*********************************/
  /* end init grap parameters      */
  /*********************************/
  if(state==NULL)
    xil_open_device();

  cis = xil_cis_create(state, param.cis_type);

  xil_set_device_attribute(image, "IMAGE_SKIP", (void *) param.image_skip);
  xil_set_device_attribute(image, "MAX_BUFFERS", (void *) param.max_buffer);
  xil_get_device_attribute(image, "FORMAT", (void **) &video_format);
  xil_set_device_attribute(image, "PORT_V", (void *) input);

  xil_get_info(image, &width, &height, &nbands, &datatype);
  /*
   * If the user has not changed the default to something else, reset the
   * default capture width and height
   */
  if (*width_ == 0){
    param.width = width;
    *width_=width>>1;
  }
  if (*height_ == 0){
    param.height = height;
    *height_=height>>1;
  }
  
    /* check for width or height too big */
  if ((param.width > width) || (param.height > height)) {
    fprintf(stderr, "ERROR: Maximum width & height are %d & %d\n",
	    width, height);
    xil_close(state);
    state=NULL;
    return(NULL);
  }
  
  if (param.do_cis == UYVY) {  
    /*
     * UYVY requires that the width of the image used with xil_compress
     * each be a multiple of 2 */
    if ((int) (((float) param.width / (float) scalex) + 0.5) % 2) {
      param.width = ((param.width / scalex) & ~1) * scalex;
      fprintf(stderr, "capture width adjusted to %d\n", param.width);
    }
  }
  
  if (param.do_cis == CELLB || param.do_cis == UYVY || param.scale) {
    /*  *height_*=2;
    *width_*=2;
    param.width =*width_ ;
    param.height =*height_;
    param.scalex = 1;
    param.scaley = 1;*/
  }
  
  if (param.do_cis == MPEG)
    if (strpbrk(param.pattern, "pPbB") != NULL) {
      
      /*
       * Only SunVideo's Mpeg1 Constant Bit Rate Compression supports P
       * and B frame types (Variable Bit Rate Compression supports I
       * frames only).
       * 
       * To select SunVideo's Mpeg1 Constant Bit Rate Compression the
       * SunVideo device attribute MPEG_RATE_CONTROL must be set. Since
       * SunVideo's Mpeg1 Constant Bit Rate Compression produces
       * 320x240 for PAL (and not 384x288), this attribute needs to be
       * set before the call to xil_get_info().
       */
      
      xil_set_device_attribute(image, "MPEG_RATE_CONTROL", (void *) TRUE);
    }
  
    /*
     * if user specifies '-s 0' for no scaling, set local 'scalefactor' to
     * 1.0
     */
  scalex = (param.scalex ? param.scalex : 1);
  scaley = (param.scaley ? param.scaley : 1);
  
  /* CellB compression will produce 0.25 bytes/pixel for the first frame */
  bytes_for_first_frame = param.width * param.height / (scalex * scaley * 4);
  
  
  if ((param.do_cis == MPEG) &&
	    ((scalex != 2) || (scaley != 2))) {
    fprintf(stderr, "ERROR: scalefactor is: %d ", scalex);
    fprintf(stderr, "       should be 2 for Mpeg1\n");
    xil_close(state);
    state=NULL;
    return(NULL);
  }
  
  /* Currently, Mpeg1 only captures PAL or NTSC */
  if (param.do_cis == MPEG) {
    if ((width == 640) && (height == 480) &&
	((param.width != 640) || (param.height != 480))) {
      fprintf(stderr, "ERROR: width,height is: (%d,%d).\n   It should be ",
	      param.width, param.height);
      fprintf(stderr, "valid NTSC=(640,480) for Mpeg1.\n");
      xil_close(state);
      state=NULL;
      return(NULL);
    }
    
    if ((width == 768) && (height == 576) &&
	((param.width != 768) || (param.height != 576))) {
      fprintf(stderr, "ERROR: width,height is: (%d,%d).\n   It should be ",
	      param.width, param.height);
      fprintf(stderr, "valid PAL=(768,576) for Mpeg1.\n");
      xil_close(state);
      state=NULL;
      
      /* return 0 so testscript does not stop */
      return(NULL);
    }
  }
  
  if (param.do_cis == CELLB) {
    
    /*
     * CellB requires that the dimensions of the image used with
     * xil_compress each be a multiple of 4
     */
    if ((int) (((float) param.width / (float) scalex) + 0.5) % 4) {
      param.width = ((param.width / scalex) & ~3) * scalex;
      fprintf(stderr, "capture width adjusted to %d\n", param.width);
    }
    
    if ((int) (((float) param.height / (float) scaley) + 0.5) % 4) {
      param.height = ((param.height / scaley) & ~3) * scaley;
      fprintf(stderr, "capture height adjusted to %d\n", param.height);
    }
  }
  
  if (param.do_cis == UYVY) {
    
    /*
     * UYVY requires that the width of the image used with xil_compress
     * each be a multiple of 2*/
    if ((int) (((float) param.width / (float) scalex) + 0.5) % 2) {
      param.width = ((param.width / scalex) & ~1) * scalex;
      fprintf(stderr, "capture width adjusted to %d\n", param.width);
    }
    
  }
  
  if ((param.width < width) || (param.height < height)) {
    image = xil_create_child(rtvc_image,
			     (width - param.width) / 2,
			     (height - param.height) / 2,
			     param.width, param.height,
			     0, nbands);
    if (image == NULL) {
      fprintf(stderr, "ERROR: failed to open child image on SUNWrtvc\n");
      xil_close(state);
      state=NULL;
      return(NULL);
    }
    xil_get_info(image, &width, &height, &nbands, &datatype);
  }

  if (scalex || scaley) {
    width = (int) ((float) width / (float) scalex + 0.5);
    height = (int) ((float) height / (float) scaley + 0.5);
    scaled_image = xil_create(state, width, height, nbands, XIL_BYTE);
    tmp_display = xil_create(state, width, height, nbands, XIL_BYTE);
  }

  
  
  /*
   * set the buffers to make sure that you don't keep information you do
   * not need, but Mpeg1 currently needs to keep at least 6 frames around.
   */
  if (param.do_cis == MPEG) {
    xil_cis_set_keep_frames(cis, 6);
    xil_cis_set_max_frames(cis, 6);
  }
  else {
    xil_cis_set_keep_frames(cis, 1);
    xil_cis_set_max_frames(cis, 1);
  }
  
  /*
   * We need to be able to seek backwards to both get-bits, and decompress
   * (but Mpeg1 doesn't support this attribute).
   */
  if ((param.do_cis == JPEG) || (param.do_cis == CELLB))
    xil_cis_set_attribute(cis, "IGNORE_HISTORY", (void *) TRUE);
  
  if (param.do_cis == JPEG) {
    /* this attribute must be TRUE for acceleration */
    xil_cis_set_attribute(cis, "ENCODE_411_INTERLEAVED", (void *) TRUE);
    xil_cis_set_attribute(cis, "COMPRESSION_QUALITY",(void *)param.quality); 
  }
  
  /*
   * if MPEG, Set Mpeg1 pattern and Bit Rate/Fixed Quant
   */
  if (param.do_cis == MPEG) {
    xil_cis_set_attribute(cis, "COMPRESSOR_INSERT_VIDEO_SEQUENCE_END", 
			  (void *) TRUE);
    
    mpattern.pattern = param.pattern;
    mpattern.repeat_count = 0;
    xil_cis_set_attribute(cis, "COMPRESSOR_PATTERN", (void *) &mpattern);
    
    if (strpbrk(param.pattern, "pPbB") != NULL) {
      
      RtvcMpeg1Weights w;
      
      /*
       * The bit rate of SunVideo's Mpeg1 Constant Bit Rate Compression
       * is set by the XIL Mpeg1 CIS attribute
       * COMPRESSOR_BITS_PER_SECOND.
       */
      
      xil_cis_set_attribute(cis, "COMPRESSOR_BITS_PER_SECOND",(void *)1150000);
      
      /* SunVideo's Mpeg1 Constant Bit Rate Compression Weights */
      w.weight_i = 10;
      w.weight_p = 6;
      w.weight_b = 1;
      xil_set_device_attribute(image, "MPEG_WEIGHTS", (void *) &w);

    }
    else
      
      /*
       * SunVideo's Mpeg1 Variable Bit Rate Compression quality is set
       * by the device attribute MPEG_FIXED_QUANT. Valid values range
       * from 1 to 31 where 1 yields the highest quality (and the
       * largest bit rate).
       */
      
      xil_set_device_attribute(image, "MPEG_FIXED_QUANT", (void *) 4);
  }
  
  /******************************************************************/
  /* collect the initialized value */
  /******************************************************************/
  param.width=width;
  param.height=height;
  exp_par.scalex=scalex;
  exp_par.scaley=scaley;
  exp_par.state=state;
  exp_par.rtvc_image=rtvc_image;
  exp_par.image=image;
  exp_par.scaled_image=scaled_image;
  exp_par.cis=cis;
  gettimeofday(&t0, NULL);
  exp_par.t=t0;
  return(XIL_OK);
}
/************************************************************************/

XilImage imageYCC,image_display;
/*
*  this function returns the number of bytes grabbed
*  image points to the grabbed data
*/
int xil_grab_image(char **image)
{
  int nbytes;
  int nframes = 0;
  Xil_unsigned8 *data=NULL;
  
  if (exp_par.scalex|| exp_par.scaley) {
    xil_scale(exp_par.image, exp_par.scaled_image, "nearest",
	      1.0 / (float) exp_par.scalex, 1.0 / (float) exp_par.scaley);
  }
  else {
    exp_par.scaled_image = exp_par.image;
  }
  xil_compress(exp_par.scaled_image, exp_par.cis);
  xil_cis_sync(exp_par.cis);
  xil_toss(exp_par.scaled_image);
  
  if (xil_cis_has_frame(exp_par.cis)) {
            
    data = (void *) xil_cis_get_bits_ptr(exp_par.cis, &nbytes, &nframes);
    exp_par.nbytes=nbytes;
    exp_par.nframes=nframes;
  }
  
  *image = (char *)data;
  return(nbytes);
}


/**************************************************************************/
/*
 * close the video device and end the xil stuff
 */
/**************************************************************************/
int xil_end_grab()
{
    /*
     * cleanup
     */
    if(rtvc_image!=NULL){
	xil_destroy(rtvc_image);
	rtvc_image=NULL;
    }
    if(exp_par.cis!=NULL) xil_cis_destroy(exp_par.cis);
    exp_par.cis=NULL;

    /*
     * shutdown
     */
    if(exp_par.state!=NULL) xil_close(exp_par.state);
    exp_par.state=NULL;
    state=NULL;
    return 0;
}

/**************************************************************************/
/*  change the Qunatization Table of the jpeg decoder
 */
/**************************************************************************/

void setQ(int quality)
{
  int qt[8][8];
  XilJpegQTable xq;
  if(state!=NULL){
    s->codec.quality=quality;
    if( s->codec.quality>75)
      s->codec.quality=70;
    param.quality= s->codec.quality;
    jpeg_luma_qt(s->codec.quality, (int*)qt);
    xq.table = 0;
    xq.value = qt;
    (void)xil_cis_set_attribute(exp_par.cis, "QUANTIZATION_TABLE", (void*)&xq);
    jpeg_chroma_qt(s->codec.quality, (int*)qt);
    xq.table = 1;
    xq.value = qt;
    (void)xil_cis_set_attribute(exp_par.cis, "QUANTIZATION_TABLE", (void*)&xq);
    xil_cis_set_attribute(exp_par.cis,"COMPRESSION_QUALITY",
			  (void *)(s->codec.quality));
  
    if(s->local->decoder!=NULL)
      init_jpeg_q(s->local->decoder, s->codec.quality);
  }
}
#endif  /* end of xil related part */




