/*
audio_stats  --  compute audio statistics

Copyright 1994 by Henning Schulzrinne
*/

#include "types.h"
#include "audio_generic.h"
#include "pcmu_l.h"
#include <math.h>     /* log10() */
#include <stdlib.h>   /* abs() */

#define PCMU_CLIP 32124         /* translated to linear */
#define L8_CLIP   32767
#define L16_CLIP  32767
#define DC_FILTER 0.0001220703  /* should 1/2^N to avoid round-off */



static float L16_dB[32768];   /* 16-bit sample to dB table */

/*
* Since this routine is called very frequently, we do the
* minimum necessary. We sample every sixtenth sample for a statistical
* impression of the volume.
*
* For mu-law, the energy estimate is 1/n sum log (|x_i|) = 1/n log prod
* |x_i| = log prod |x_i|^(1/n), which is the log of the geometric mean
* of the amplitudes.  for linear, the energy estimate is log (1/n sum
* |x_i|), the log of the average of the amplitudes.
*
* Return 0 if successful, -1 on error (wrong encoding).
*/
int audio_stats(void *buf, size_t bytes, audio_descr_t *ad,
  audio_stats_t *as)
{
  int i, j;
  u_int8 *cp;
  int16  *sp;
  long energy = 0;
  int count = 0;
  int32 sample;
  int32 peak[2];

  peak[0] = peak[1] = 0;
  switch(ad->encoding) {
  case AE_PCMU:
    if (ad->channels == 1) {
      for (i = 0, cp = (u_int8 *)buf; i < bytes; i++, cp++) {
        sample = pcmu_l16(*cp);
        as->c[0].dc = as->c[0].dc + DC_FILTER * (sample - as->c[0].dc);
        sample -= as->c[0].dc;
        if (sample < 0) sample = -sample;
        if (sample > peak[0]) peak[0] = sample;
      }
    }
    break;

  case AE_L8:
    for (i = 0, cp = (u_int8 *)buf; i < bytes; i += 16, cp++) {
      energy += (*cp) & 0x7f;
      count++;
    }
    break;

  case AE_L16:
    bytes >>= 1;
    if (ad->channels == 1) {
      for (i = 0, sp = (int16 *)buf; i < bytes; i++, sp++) {
        sample = *sp;
        as->c[0].dc = as->c[0].dc + DC_FILTER * (sample - as->c[0].dc);
        sample -= as->c[0].dc;
        if (sample < 0) sample = -sample;
        if (sample > peak[0]) {
          peak[0] = sample;
        }
      }
    }
    else {
      for (i = 0, sp = (int16 *)buf; i < bytes; i += ad->channels) {
        for (j = 0; j < ad->channels; j++) {
          sample = sp[i];
          as->c[j].dc = as->c[j].dc + DC_FILTER * (sample - as->c[j].dc);
          sample -= as->c[j].dc;
          if (sample < 0) sample = -sample;
          if (sample > peak[j]) peak[j] = sample;
        }
      }
    }
    break;

  default:
    return -1;
    break;
  }

  for (i = 0; i < ad->channels; i++) {
    if (peak[i] > 32767) peak[i] = 32767;
    as->c[i].peak = L16_dB[peak[i]];
  }
  return 0;
} /* audio_stats */

/*
* Initialize dB-lookup table, which maps 16-bit value to .
*/
void audio_stats_init(void)
{
  int i;
  const int L16_min = 33;     /* corresponds to -60 dB */
  const int L16_max = 32767;  /* corresponds to 0 dB */
  float dB_max, dB_min;

  dB_max = 20.0 * log10(2.0 * L16_max);
  dB_min = 20.0 * log10(2.0 * L16_min);

  for (i = 0; i <= L16_min; i++) L16_dB[i] = 0.0;

  for (; i < L16_max; i++) {
    L16_dB[i] = (dB_max - dB_min - (-20.0 * log10((float)i/L16_max))) /
        (dB_max - dB_min);
  }
  L16_dB[L16_max] = 1.0;
} /* audio_stats_init */
