/*
audio  --  audio utility functions

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include "types.h"         /* u_int32 */
#include "audio_generic.h"
#include "arch_sys.h"      /* infinity() */
#include "pcmu_l.h"
#include <memory.h> 
#include <values.h>        /* MAXINT */
#include <math.h>          /* infinity() */
#include <string.h>        /* strlen() */
#include <ansi.h>



/*
* Return 1 if the two audio descriptors are the same.
*/
int audio_cmp(audio_descr_t *ad1, audio_descr_t *ad2)
{
  int sample_rate;

  /* implied */
  if (ad2->sample_rate == 0) {
    sample_rate = (ad2->encoding < AE_L16) ? 8000 : 16000;
  }
  else sample_rate = ad2->sample_rate;
  return ad1->encoding    == ad2->encoding &&
         ad1->channels    == ad2->channels &&
         ad1->sample_rate == sample_rate;
} /* audio_cmp */


unsigned audio_bytes_to_samples(audio_descr_t *ad, int bytes)
{
  if (ad->channels > 1) bytes /= 2;
  switch (ad->encoding) {
  case AE_L16:
    return bytes / 2;

  case AE_L8:
  case AE_PCMA:
  case AE_PCMU:
    return bytes;

  /* does not include header, since block can be any size */
  case AE_DVI4:
    return bytes * 2;

  /* block of 20 ms (160 samples) into 33 bytes */
  case AE_GSM:
    return (bytes / 33) * 160;

  /* block of 20 ms (160 samples) into 13 bytes */
  case AE_LPC:
    return (bytes / 13) * 160;

  default:
    return 0;
  }
} /* audio_bytes_to_samples */


/*
* Return number of seconds corresponding to audio descriptor and
* data length. If 'bytes' is negative, return infinity.
*/
double audio_bytes_to_seconds(audio_descr_t *ad, int bytes)
{
  if (bytes < 0) return infinity();
  return audio_bytes_to_samples(ad, bytes) / (double)ad->sample_rate;
} /* audio_bytes_to_seconds */


unsigned audio_seconds_to_bytes(audio_descr_t *ad, double secs)
{
  return audio_samples_to_bytes(ad, secs * ad->sample_rate);
} /* audio_seconds_to_bytes */


unsigned audio_samples_to_bytes(audio_descr_t *ad, int samples)
{
  samples *= ad->channels;
  switch (ad->encoding) {
  case AE_L16:
    return samples * 2;

  case AE_L8:
  case AE_PCMA:
  case AE_PCMU:
    return samples;

  /* does not include header */
  case AE_DVI4:
    return samples / 2;

  /* block of 20 ms (160 samples) into 33 bytes */
  case AE_GSM:
    return (samples / 160) * 33;

  /* block of 20 ms (160 samples) into 13 bytes */
  case AE_LPC:
    return (samples / 160) * 13; 

  default:
    return 0;
  }
} /* audio_samples_to_bytes */


/*
* Set audio buffer 'buffer' to silence.
*/
int audio_silence(void *buffer, size_t bytes, audio_encoding_t encoding)
{
  u_int8 s = 0;

  if (encoding == AE_PCMU) s = l16_to_pcmu(s);

  memset(buffer, s, bytes);
  return 0;
} /* audio_silence */
