/*
list  --  list manipulation routines.

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include <string.h>
#include <stdlib.h>   /* atoi() */
#include <stdio.h>    /* sprintf() */
#include <ctype.h>    /* isdigit(), isspace() */
#include "util.h"     /* itoa() */
#include "list.h"
#include "bit.h"



/*
* Return index of 's' in list 'list', -1 if not found.
*/
int list_to_index(char *list[], char *s)
{
  int i;

  /* handle integers */
  if (!s) return 0;
  if (isdigit(*s) || *s == '-') return atoi(s);

  /* handle booleans */
  if (!strcmp(s, "on") || !strcmp(s, "true") || !strcmp(s, "yes")) return 1;
  else if (!strcmp(s, "off") || !strcmp(s, "false") || !strcmp(s, "no")) return 0;

  /* search in list */
  for (i = 0; list[i]; i++) {
    if (strcmp(s, list[i]) == 0) return i;
  }
  return -1;
} /* list_to_index */

typedef struct {
  char *start;
  int len;
} word_t;


/*
* Determine word beginnings within 's'.
*/
static int list_words(char *s, word_t *word, int max_count)
{
  enum {white, black} state = white;
  int count = 0;

  for (; ; s++) {
    switch(state) {
    case white:
      if (!isspace(*s)) {
        word[count].start = s;
        word[count].len = 0;
        state = black;
      }
      break;
    case black:
      if ((*s == 0) || isspace(*s)) {
        word[count].len = s - word[count].start;
        count++;
        if (count >= max_count) return count;
        if (*s == 0) return count;
        state = white;
      }
      break;
    }
  }
}


/*
* return bit mask; use old mask if s == '*'
* list: in: array of strings from which match is chosen
* s:    in: list of space-separated items
* mask: in: old mask
*/
int list_to_mask(char *list[], char *s, int mask)
{
  int count;
  int i, j;
  int negate;
  word_t word[20];

  /* handle integers */
  if (!s) return 0;
  if (*s == '*') return mask;
  if (isdigit(*s) || *s == '-') return atoi(s);

  /* handle booleans if no list present */
  if (!list || !list[0]) {
    if (!strcmp(s, "true") || !strcmp(s, "on")) mask = 1;
    else if (!strcmp(s, "false") || !strcmp(s, "off")) mask = 0;
    return mask;
  }

  count = list_words(s, word, 20);
  for (i = 0, mask = 0; i < count; i++) {
    char *w = word[i].start;
    int len = word[i].len;
    negate = (*w == '!');
    if (negate) {
      w++;
      len--;
    }

    for (j = 0; list[j]; j++) {
      if ((strlen(list[j]) == len) && (strncmp(list[j], w, len) == 0)) {
        if (negate) mask &= ~SET(j); 
        else mask |= SET(j);
      }
    }
  }
  return mask;
} /* list_to_mask */


/*
* list:  in: zero-terminated string list; if entry = nil, show "*"
* index: in: index into list; if > list size, show number
*/
char *index_to_list(char *list[], int index)
{
  static char b[80];
  int lsize;          /* number of elements in list */

  for (lsize = 0; list[lsize]; lsize++) ; 
  if (index >= 0 && index < lsize)
    sprintf(b, "%s", list[index] ? list[index] : "*");
  else
    sprintf(b, "%d", index);

  return b;
} /* index_to_list */


/*
Return space-separated list of list entries corresponding to bit
positions.
list: in: zero-terminated list
mask: in: bit mask
*/
char *mask_to_list(char *list[], int mask)
{
  static char b[80];
  char *s = b;   /* pointer within buffer */
  int i;

  if (!list || !list[0]) return itoa(mask);
  b[0] = '\0';
  for (i = 0; list[i]; i++) {
    if (BIT(mask, i)) {
      if (s != b) strcpy(s++, " ");
      strcpy(s, list[i]);
      s += strlen(list[i]);
    }
  }
  return b;
} /* mask_to_list */

/*
* Search through key/value array.
* Return last value if no match.
*/
int list_key2value(list_map_t *l, char *key, int ignore_case)
{
  int i;

  if (ignore_case) {
    for (i = 0; l[i].key; i++) {
      if (strcasecmp(l[i].key, key) == 0) return l[i].value;
    }
  }
  else {
    for (i = 0; l[i].key; i++) {
      if (strcmp(l[i].key, key) == 0) return l[i].value;
    }
  }
  return l[i].value;
} /* list_key2value */


/*
* Search through key/value array.
* Return last value if no match.
*/
char *list_value2key(list_map_t *l, int value)
{
  int i;

  for (i = 0; l[i].key; i++) {
    if (l[i].value == value) return l[i].key;
  }
  return l[i].key;
} /* list_value2key */
