/*
lpc  --  linear predictive audio codec.

Ron Frederick (frederic@parc.xerox.com) or Xerox PARC,  Palo Alto,  CA,
contributed this LPC codec which is based on an implementation done by
Ron Zuckerman (ronzu@isu.comm.mot.com) of Motorola which was posted to
the Usenet group comp.dsp on 26 June 1992. The original code is
available as
   ftp://parcftp.xerox.com/pub/net-research/lpc.tar.Z

The code was modified for use within NEVOT by H. Schulzrinne.
*/

#include <math.h>
#include <string.h>     /* memcpy() */
#include <stdlib.h>     /* rand(), malloc() */
#include "types.h"
#include "audio_descr.h"
#include "lpc.h"
#include "pcmu_l.h"



#define MAXWINDOW       1000    /* Max analysis window length */
#define FS              8000.0  /* Sampling rate */

#define DOWN            5       /* Decimation for pitch analyzer */
#define PITCHORDER      4       /* Model order for pitch analyzer */
#define FC              600.0   /* Pitch analyzer filter cutoff */
#define MINPIT          50.0    /* Minimum pitch */
#define MAXPIT          300.0   /* Maximum pitch */

#define MINPER          (int)(FS/(DOWN*MAXPIT)+.5)  /* Minimum period */
#define MAXPER          (int)(FS/(DOWN*MINPIT)+.5)  /* Maximum period */

#define WSCALE          1.5863  /* Energy loss due to windowing */

#define BIAS 0x84       /* define the add-in bias for 16 bit samples */
#define CLIP 32635

static void auto_correl(float w[], int n, int p, float r[])
{
  int i, k, nk;

  for (k=0; k <= p; k++) {
    nk = n-k;
    r[k] = 0.0;
    for (i=0; i < nk; i++) r[k] += w[i] * w[i+k];
  }
} /* auto_correl */


static void durbin(float r[], int p, float k[], float *g)
{
  int i, j;
  float a[LPC_FILTORDER+1], at[LPC_FILTORDER+1], e;

  for (i=0; i <= p; i++) a[i] = at[i] = 0.0;

  e = r[0];
  for (i=1; i <= p; i++) {
    k[i] = -r[i];
    for (j=1; j < i; j++) {
      at[j] = a[j];
      k[i] -= a[j] * r[i-j];
    }
    k[i] /= e;
    a[i] = k[i];
    for (j=1; j < i; j++) a[j] = at[j] + k[i] * at[i-j];
    e *= 1.0 - k[i]*k[i];
  }

  *g = sqrt(e);
} /* durbin */


static void inverse_filter(int buflen, float w[], float k[])
{
  int i, j;
  float b[PITCHORDER+1], bp[PITCHORDER+1], f[PITCHORDER+1];

  for (i = 0; i <= PITCHORDER; i++) b[i] = f[i] = bp[i] = 0.0;

  for (i = 0; i < buflen/DOWN; i++) {
    f[0] = b[0] = w[i];
    for (j = 1; j <= PITCHORDER; j++) {
      f[j] = f[j-1] + k[j] * bp[j-1];
      b[j] = k[j] * f[j-1] + bp[j-1];
      bp[j-1] = b[j-1];
    }
    w[i] = f[PITCHORDER];
  }
} /* inverse_filter */


static void calc_pitch(int buflen, float w[], float *per)
{
  int i, j, rpos;
  float d[MAXWINDOW/DOWN], k[PITCHORDER+1], r[MAXPER+1], g, rmax;
  float rval, rm, rp;
  float a, b, c, x, y;
  static int vuv=0;

  for (i=0, j=0; i < buflen; i+=DOWN) d[j++] = w[i];
  auto_correl(d, buflen/DOWN, PITCHORDER, r);
  durbin(r, PITCHORDER, k, &g);
  inverse_filter(buflen, d, k);
  auto_correl(d, buflen/DOWN, MAXPER+1, r);
  rpos = 0;
  rmax = 0.0;
  for (i = MINPER; i <= MAXPER; i++) {
    if (r[i] > rmax) {
      rmax = r[i];
      rpos = i;
    }
  }

  rm = r[rpos-1];
  rp = r[rpos+1];
  rval = rmax / r[0];

  a = 0.5 * rm - rmax + 0.5 * rp;
  b = -0.5*rm*(2.0*rpos+1.0) + 2.0*rpos*rmax + 0.5*rp*(1.0-2.0*rpos);
  c = 0.5*rm*(rpos*rpos+rpos) + rmax*(1.0-rpos*rpos) + 0.5*rp*(rpos*rpos-rpos);

  x = -b / (2.0 * a);
  y = a*x*x + b*x + c;
  x *= DOWN;

  rmax = y;
  rval = rmax / r[0];
  if (rval >= 0.4 || (vuv == 3 && rval >= 0.3)) {
    *per = x;
    vuv = (vuv & 1) * 2 + 1;
  } else {
    *per = 0.0;
    vuv = (vuv & 1) * 2;
  }
} /* calc_pitch */


/*
* Initialize LPC codec with period 'period' measured in seconds. 
* Needs to be called only once.  Return 0 on success, -1 if resulting
* buffer size is too large.
*/
void *lpc_init(void *vp, double period)
{
  int i;
  float r, v, w, wcT;
  struct lpc_state *state = (struct lpc_state *)vp;

  if (period == 0.) return (void *)0;
  if (!state) state = (void *)malloc(sizeof(struct lpc_state));
  if (!state) return (void *)0;

  state->framelen = period * FS;
  state->buflen   = state->framelen * 3/2;
  if (state->buflen > MAXWINDOW) return (void *)0;

  for (i = 0; i < state->buflen; i++) {
    state->s[i] = 0.0;
    state->h[i] = WSCALE*(0.54 - 0.46 * cos(2 * M_PI * i / 
                    (state->buflen - 1.0)));
  }

  wcT = 2 * M_PI * FC / FS;
  r = 0.36891079 * wcT;
  v = 0.18445539 * wcT;
  w = 0.92307712 * wcT;
  state->fa[1] = -exp(-r);
  state->fa[2] = 1.0 + state->fa[1];
  state->fa[3] = -2.0 * exp(-v) * cos(w);
  state->fa[4] = exp(-2.0 * v);
  state->fa[5] = 1.0 + state->fa[3] + state->fa[4];

  state->u1 = 0.0;
  state->yp1 = 0.0;
  state->yp2 = 0.0;

  state->per = 0.0;
  state->G = 0.0;
  for (i = 1; i <= LPC_FILTORDER; i++) state->k[i] = 0.0;
  for (i = 0; i <= LPC_FILTORDER; i++) 
    state->b[i] = state->bp[i] = state->f[i] = 0.0;
  state->pitchctr = 0;

  return (void *)state;
} /* lpc_init */


static void lpc_encode_frame(void *in_buf, audio_descr_t *ad, 
  struct lpc_state *state, lpc_t *lpc)
{
  int i, j;
  float w[MAXWINDOW], r[LPC_FILTORDER+1];
  float per, G, k[LPC_FILTORDER+1];
  u_int8 *in_buf_c = 0;
  int16  *in_buf_s = 0;
  int framelen = state->framelen;
  int buflen   = state->buflen;

  if (ad->encoding == AE_L16) in_buf_s = (int16 *)in_buf;
  else in_buf_c = (u_char *)in_buf;

  for (i = 0, j = buflen - framelen; i < framelen; i++, j++) {
    /* convert input sample to +/-1.0 range
       and place at *end* of s[] buffer */
    if (ad->encoding == AE_L16)
      state->s[j] = in_buf_s[i] / 32768.0;
    else
      state->s[j] = pcmu_l16(in_buf_c[i])/32768.0;

    state->u = state->fa[2] * state->s[j] - state->fa[1] * state->u1;
    state->y[j] =   state->fa[5] * state->u1
                  - state->fa[3] * state->yp1
                  - state->fa[4] * state->yp2;
    state->u1  = state->u;
    state->yp2 = state->yp1;
    state->yp1 = state->y[j];
  }

  calc_pitch(state->buflen, state->y, &per);

  for (i=0; i < buflen; i++) w[i] = state->s[i] * state->h[i];
  auto_correl(w, buflen, LPC_FILTORDER, r);
  durbin(r, LPC_FILTORDER, k, &G);

  lpc->period = per * (1<<8);   /* XXX change this to shift? */
  lpc->gain   = G * (1<<8);
  for (i=0; i < LPC_FILTORDER; i++) lpc->k[i] = k[i+1] * (1<<7);

  memcpy((char *)state->s, (char *)state->s+framelen, (buflen-framelen)*sizeof(state->s[0]));
  memcpy((char *)state->y, (char *)state->y+framelen, (buflen-framelen)*sizeof(state->y[0]));
} /* lpc_encode_frame */


/*
* If in_size (converted to samples) exceeds the defined codec frame
* size, we call the actual coded several times and pack a number
* of descriptors into a single buffer.
*/
int lpc_encode(void *in_buf, int in_size, audio_descr_t *ad,
  void *out_buf, int *out_size, void *state, int header_flag)
{
  int i;
  int framelen = ((struct lpc_state *)state)->framelen;
  int frames;  /* number of frames */
  int bpf;     /* bytes per frame */
  lpc_t *lpc = (lpc_t *)out_buf;
  char *b    = (char *)in_buf;

  /* compute number of frames */
  switch(ad->encoding) {
  case AE_L16:
    frames = in_size / (framelen*2);
    bpf    = framelen*2;
    break;
  case AE_PCMU:
    frames = in_size / framelen;
    bpf    = framelen;
    break;
  default:
    frames = 0;
    bpf    = 0;
    break;
  }

  for (i = 0; i < frames; i++, lpc++, b += bpf) {
    lpc_encode_frame(b, ad, state, lpc);
  }
  *out_size = frames * sizeof(lpc_t);
  return 0;
} /* lpc_encode */


static void lpc_decode_frame(lpc_t *lpc, audio_descr_t *ad, 
  struct lpc_state *state, void *out_buf)
{
  double speed = 1.0;
  int framelen = state->framelen;
  int i, j, flen = framelen/speed;
  float per, G, k[LPC_FILTORDER+1];
  float u, NewG, Ginc, Newper, perinc;
  float Newk[LPC_FILTORDER+1], kinc[LPC_FILTORDER+1];

  /* data type conversion */
  int16  *out_buf_s = (int16 *)out_buf;
  u_int8 *out_buf_c = (u_int8 *)out_buf;

  per  = (float)lpc->period / (1<<8);
  G    = (float)lpc->gain / (1<<8);
  k[0] = 0.0;
  for (i = 0; i < LPC_FILTORDER; i++) k[i+1] = (float)lpc->k[i] / (1<<7);

  if (per == 0.0) {
    G /= sqrt(state->buflen/3.0);
  } else {
    i = state->buflen / per;
    if (i == 0) i = 1;
    G /= sqrt((float)i);
  }

  Newper = state->per;
  NewG   = state->G;
  for (i=1; i <= LPC_FILTORDER; i++) Newk[i] = state->k[i];
    
  if (state->per != 0 && per != 0) {
    perinc = (per - state->per) / flen;
    Ginc   = (G   - state->G) / flen;
    for (i=1; i <= LPC_FILTORDER; i++) 
      kinc[i] = (k[i] - state->k[i]) / flen;
  } else {
    perinc = 0.0;
    Ginc   = 0.0;
    for (i = 1; i <= LPC_FILTORDER; i++) kinc[i] = 0.0;
  }
    
  if (Newper == 0) state->pitchctr = 0;
    
  for (i = 0; i < flen; i++) {
    /* generate noise if unvoiced */
    if (Newper == 0) {
      u = (rand()/2147483648.0) * NewG;
    } else {
      if (state->pitchctr == 0) {
        u = NewG;
        state->pitchctr = (int) Newper;
      } else {
        u = 0.0;
        state->pitchctr--;
      }
    }
      
    state->f[LPC_FILTORDER] = u;
    for (j = LPC_FILTORDER; j >= 1; j--) {
      state->f[j-1] = state->f[j] - Newk[j] * state->bp[j-1];
      state->b[j]   = Newk[j] * state->f[j-1] + state->bp[j-1];
      state->bp[j]  = state->b[j];
    }
    state->b[0] = state->bp[0] = state->f[0];
    
    /* output value */
    if (ad->encoding == AE_PCMU)
      out_buf_c[i] = l16_to_pcmu((int) (state->b[0] * 32768.0));
    else
      out_buf_s[i] = (int) (state->b[0] * 32768.0);

    Newper += perinc;
    NewG   += Ginc;
    for (j = 1; j <= LPC_FILTORDER; j++) Newk[j] += kinc[j];
  } /*for*/
  
  state->per = per;
  state->G   = G;
  for (i = 1; i <= LPC_FILTORDER; i++) state->k[i] = k[i];
} /* lpc_decode_frame */


/*
* Translate from LPC to local audio format.
*/
int lpc_decode(void *in_buf, int in_size, audio_descr_t *ad, 
  void *out_buf, int *out_size, void *state)
{
  int i;       /* frame loop index */
  int frames;  /* number of frames */
  int framelen = ((struct lpc_state *)state)->framelen;
  int bpf;     /* bytes per frame */
  lpc_t *lpc = (lpc_t *)in_buf;
  char *b    = (char *)out_buf;

  frames = in_size / sizeof(lpc_t);
  switch(ad->encoding) {
  case AE_L16:
    bpf = framelen*2;
    break;
  case AE_PCMU:
    bpf = framelen;
    break;
  default:
    bpf = 0;
    break;
  }

  for (i = 0; i < frames; i++, lpc++, b += bpf) {
    lpc_decode_frame(lpc, ad, state, b);
  }
  *out_size = bpf * frames;
  return 0;
} /* lpc_decode */
