

/*
* Compute control packet period.
* members:        number of session members
* senders:        senders among members
* bw:             total session bandwidth
* we_sent:        we sent recently
* packet_size:    current packet size
* avg_rtcp_size:  smoothed average
* initial:        initial call
*/
double rtcp_period(int members, int senders, double bw, int we_sent, 
  int packet_size, int *avg_rtcp_size, int initial)
{
  /*
   * Minimum time between control packets from this site (in seconds).
   * This time prevents the reports from 'clumping' when sessions are
   * small and the law of large numbers isn't helping to smooth
   * out the traffic. It also keeps the report interval from becoming
   * ridiculously small during transient outages like a network
   * partition.
   */
  double const RTCP_MIN_TIME = 5.;

  /*
   * Fraction of the control bandwidth to be shared among active
   * senders. (This fraction was chosen so that in a typical session
   * with one or two active senders, the computed report time would be
   * roughly equal to the minimum report time so that we don't
   * unnecessarily slow down receiver reports.) The receiver fraction
   * must be 1 - the sender fraction.
   */
  double const RTCP_SENDER_BW_FRACTION = 0.25;
  double const RTCP_RECEIVER_BW_FRACTION = (1 - RTCP_SENDER_BW_FRACTION);

  /*
   * Gain (smoothing constant) for the low pass filter that estimates
   * the average RTCP packet size (see Cadzow reference).
   */
  double const RTCP_SIZE_GAIN = (1./16.);

  double t = 0.;   /* interval */
  double rtcp_min_time = RTCP_MIN_TIME;
  int n;      /* number of members for computation */

  /*
   * Very first call at application start-up uses half the minimum
   * delay for quicker notification while still allowing some time before
   * reporting for randomization and to learn about other sources.
   * The avg. RTCP size is initialized to 128 bytes which is
   * conservative (it assumes everyone else is generating SRs instead
   * of RRs).
   */
  if (initial) {
    rtcp_min_time /= 2;
    *avg_rtcp_size = 128;
  }

  /*
   * If there were active senders, split the RTCP bandwidth between
   * senders and receivers. Otherwise, the receivers share all the
   * RTCP bandwidth.
   */
  n = members;
  if (senders > 0 && senders < members * RTCP_SENDER_BW_FRACTION) {
    if (we_sent) {
      bw *= RTCP_SENDER_BW_FRACTION;
      n   = senders;
    }
    else {
      bw *= RTCP_RECEIVER_BW_FRACTION;
      n  -= senders;
    }
  }

  /*
   * Update the average size estimate by the size of the report
   * packet we just sent.
   */
  *avg_rtcp_size += (packet_size - *avg_rtcp_size) * RTCP_SIZE_GAIN;

  /*
   * The effective number of sites times the average packet size is
   * the total number of bytes sent when each site sends a report. 
   * Dividing this by the effective bandwidth gives the time interval
   * over which those packets must be sent in order to meet the
   * bandwidth target. In that time interval we send one report so
   * this time is also our average time between reports.
   */
  if (bw > 0)
    t = (*avg_rtcp_size) * n / bw;

  /* Enforce minimum spacing. */
  if (t < rtcp_min_time) t = rtcp_min_time;

  /*
   * To avoid traffic bursts from unintended synchronization with
   * other sites, we then pick our actual next report interval as a
   * random number uniformly distributed between 0.5*t and 1.5*t.
   * We do this in the calling routine so that we can use the period
   * for timing out sources.
   */
  return t;
} /* period */
