/*
notify  --  Tk-specific wrapper for input function handler

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <tk.h>
#include <sys/time.h>
#include <signal.h>
#include <assert.h>
#include <malloc.h>
#include "notify.h"


/*
* Call table indexed by file descriptor.
*/
#define MAX_FD 64
static struct {
  Notify_func_input func;  /* function to be called */
  Notify_client client;    /* first argument to that function (from set_input_func) */
} fd_table[MAX_FD];

static struct timer_table {
  Notify_func func;     /* actual function to be called */
  Notify_client client; /* value to be passed to function */
  int interval;         /* scheduling interval (milliseconds) */
  int pending;          /* a timer is pending */
  Tk_TimerToken token;  /* internal Tk handle for timer */
} *timer_table;
static int timer_count;


/*
* Called by Tk; invokes actual event handler.
*/
static void file_handler(ClientData fd, int mask)
{
  assert((long)fd < MAX_FD);
  (*fd_table[(long)fd].func)(fd_table[(long)fd].client, (long)fd);
} /* file_handler */


/*
* The function callback differs from the xview notifier, as the
* second argument for Tk contains the event mask rather than the file
* descriptor. We are thus forced to keep another layer of indirection,
* indexed by file descriptor.
*/
Notify_func_input notify_set_input_func(Notify_client client, Notify_func_input func, int fd)
{
  assert(fd < MAX_FD);
  if (func) {
    fd_table[fd].client = client;
    fd_table[fd].func   = func;
    Tk_CreateFileHandler(fd, TK_READABLE, file_handler, (ClientData)fd);
  }
  else {
    fd_table[fd].client = 0;
    fd_table[fd].func   = 0;
    Tk_DeleteFileHandler(fd);
  }
  return func;
} /* notify_set_input_func */

typedef void (*sig_handler)();

/*
* Tk does not handle signals itself, so we can use the straight signal
* call.
*/
Notify_func_signal notify_set_signal_func(Notify_client client,
  Notify_func_signal func, int sig, Notify_signal_mode when)
{
  signal(sig, (sig_handler)func);
  return func;
} /* notify_set_signal_func */


/*
* The call to the actual timer handler (func) must be before setting
* up the replacement timer since that function may call
* set_timer_func() to change the periodicity.  Also, to avoid
* cascading calls, we need the pending flag for the same case.  This
* calling order extends the period by the execution time of the
* function called.  Alternatively, on discovering a change of timer
* value, we could delete the just created timer.  Probably not worth
* the hassle.
*/
static void timer_handler(ClientData client)
{
  int index = (long)client; 

  /* call actual handler function */
  assert(index < timer_count);
  timer_table[index].pending = 0;
  (*timer_table[index].func)(timer_table[index].client);

  /*
   * Don't set up another timer if the function called
   * notify_set_periodic_func() for this client/func pair already.
   */
  if (!timer_table[index].pending) {
    timer_table[index].token =
      Tk_CreateTimerHandler(timer_table[index].interval, timer_handler,
      client);
  }
} /* timer_handler */


/*
* Periodic invocation function.  A 'interval' value of zero disables
* the particular function.  A non-zero interval for an existing
* periodic timer identified by the (func, client) pair changes the
* interval.  The interval is specified in milliseconds.
*/
void notify_set_periodic_func(Notify_client client, Notify_func func, int interval)
{
  int index = 0;

  if (interval == 0) {
    /* search for function and client */
    for (index = 0; index < timer_count; index++) {
      if (timer_table[index].func   == func && 
          timer_table[index].client == client) break;
    }
    /* if found in table, delete all record of this timer */
    if (index < timer_count) {
      Tk_DeleteTimerHandler(timer_table[index].token);
      timer_table[index].interval = 0;
      timer_table[index].pending  = 0;
      timer_table[index].func     = 0;
      timer_table[index].client   = 0;
      timer_table[index].token    = 0;
    }
  }
  /* change period or establish new timer */
  else {
    /* check if existing timer */
    for (index = 0; index < timer_count; index++) {
      if (timer_table[index].client == client) break;
    }
    if (index >= timer_count) {
      /* search for available timer slot to create new timer */
      for (index = 0; index < timer_count; index++) {
        if (timer_table[index].func == 0) break;
      }
      /* If no empty slot found, extend array. */
      if (index >= timer_count) {
        timer_count++;
        timer_table = timer_table ? 
          (struct timer_table *)realloc(timer_table, timer_count * sizeof(struct timer_table)) :
          (struct timer_table *)calloc(timer_count, sizeof(struct timer_table));
      }
      timer_table[index].pending = 0;
    }
    timer_table[index].interval  = interval;
    timer_table[index].func      = func;
    timer_table[index].client    = client;
    if (!timer_table[index].pending) {
      timer_table[index].token   =
        Tk_CreateTimerHandler(interval, timer_handler, (ClientData)index);
      timer_table[index].pending = 1;
    }
  }
} /* notify_set_periodic_func */
