/*
sun4_audio  --  audio interface for Sun 4 SPARCstation.

The recording gain rounds to the following values:
0 2 8 14 20 26 32 38 44 50 56 62 68 74 80 85 91 97 103 109 115 121 127
133 139 145 151 157 163 169 174 180 186 192 198 204 210 216 222 228
234 240 246 252 255

For some systems (SS20/Solaris, others?), the recording gains are:
15 31 47 63 79 95 111 127 143 159 175 191 207 223 239 255

audio_u2s() translates mu-law into the range -32124 to 32124 (full
16 bits) rather than the actual 13-bit range.

The BSD (Berkeley) audio driver written by Van Jacobson does not have
a control device.  Thus, if we can't open the control device, we just
use the regular audio device.  If the control device is unavailable,
the audio settings are stored until the audio device is opened and
then set.

The sample_rate, channels, precision, and encoding fields are treated
as read-only for /dev/audioctl, so that applications can be guaranteed
that the existing audio format will stay in place until they
relinquish the audio device.  AUDIO_SETINFO will return EINVAL when
the desired configuration is not possible, or EBUSY when another
process has control of the audio device.

Once set, the following values persist through subsequent open() and
close() calls of the device: play.gain, record.gain, play.balance,
record.balance, output_muted, monitor_gain, play.port, and
record.port.  All other state is reset when the corresponding I/O
stream of /dev/audio is closed.

SIGPOLL handling is slow and may lead to overflow cycles feeding on
themselves; thus, we disable the SIGPOLL handling if 'track'
is zero. Also, when there is a device error (over/underflow), we do
not update the application state. The latter has the unfortunate
consequence that once an error has occured, outside changes are no
longer tracked. The ability to reset the error indicator is needed.

The .error flag needs to be explicitly reset with AUDIO_SETINFO.

SPARC ELC, SLC with non-multicast 4.1.3
---------------------------------------
These systems won't read from the audio device if it is opened with
O_RDWR. Thus, we need to maintain to separate file descriptors, one
for reading and another for writing.

SPARCstation 5 with Solaris 2.3
-------------------------------
The SPARCstation 5 has to open the audio device O_RDWR, otherwise
setting values will yield an EINVAL error.  For the first 400,000
samples (for PCMU), reading seems to happen in spurts of about 8192
bytes, but then the system 'catches itself' and continues normally. 
This is visible in a rather sluggish VU meter display.

SPARCstation 10 with Solaris 2.1 (SunOS 5.1) or SunOS 4.1.3:
------------------------------------------------------------
Precision=16 only works with AUDIO_ENCODING_LINEAR; otherwise
precision=8 remains unchanged.

The following combinations are possible:
encoding  sample_rate  precision  channels
ALAW/ULAW    8000         8         1
LINEAR       any         16        1/2 

Play and record format must be the same. They cannot be changed
(EBUSY) if play and record are on separate file descriptors.
Thus, we need to use a single descriptor.

sun/audioio.h is replaced by sys/audioio.h, which also needs
<sys/types.h>. Use multimedia/audio_device.h instead. (Under
/usr/demo/SOUND/include or /usr/demo/SOUND.)

The AUDIO_GETDEV return value differs: instead of an int, it is a
structure of text fields.

Solaris 2.x
-----------
The audio_prinfo_t structure contains a new field 'buffer_size'.

On Solaris, signal() has the property "the system first sets the
signal's disposition to SIG_DFL before executing the signal handler"
(signal(2)).  In other words, SVR4 implementes old-style unreliable
signals.  This causes a small race condition, in that additional
SIGPOLLs are not caught and trigger SIG_DFL (which for SIGPOLL is
exit(), with the error message "Stream I/O possible").  Thus, we use
sigaction() with the appropriate flags.  See also Stevens, "Advanced
Programming in the UNIX Environment", p.  298.

#ifdef AUDIO_DIAG_LOOPBACK is used to test for advanced features like
'buffer_size' and 'balance' fields.

Copyright 1995 by Henning Schulzrinne; all rights reserved.
*/

#include <fcntl.h>       /* O_RDWR */
#include <sys/ioctl.h>   /* ioctl() */
#include <unistd.h>      /* close(), ioctl() */
#ifdef sun4
#include <sun/audioio.h>
#else
#include <sys/audioio.h>
#endif
#include <multimedia/audio_device.h>
#include <multimedia/libaudio.h>
#include "audio_generic.h" /* AE_, audio_descr_t */
#include "ahw.h"
#include <errno.h>
#include <stropts.h>     /* I_SETSIG */
#include <signal.h>      /* SIGPOLL */
#include <ansi.h>

#define AUDIO_DATA "/dev/audio"
#define AUDIO_CTL  "/dev/audioctl"
#ifndef AUDIO_LINE_OUT
#define AUDIO_LINE_IN  (0x02)
#define AUDIO_LINE_OUT (0x04) 
#endif

static int ac = -1;                 /* audio control file descriptor */
static int au[2] = {-1,-1};         /* audio file descriptors */
static int precision = 0;           /* 8: amd; 16: dbri */
static audio_info_t ai_global;      /* audio properties when opening /dev/audio */
static signal_func_t track;         /* handler that tracks audio state */
static long play_count, rec_count;  /* number of bytes played, recorded */

int ahw_changed = 0;                /* there have been changes */

/* gcc include files don't define these */
#ifndef AUDIO_ENCODING_LINEAR
#define AUDIO_ENCODING_LINEAR    (3)
#define AUDIO_ENCODING_FLOAT   (100)
#define AUDIO_ENCODING_G721    (101)
#define AUDIO_ENCODING_G722    (102)
#define AUDIO_ENCODING_G723    (103)
#endif
#ifndef AUDIO_GETDEV
#define AUDIO_GETDEV  _IOR('A', 4, int)
#define AUDIO_DEV_AMD  (1) /* audioamd device */
#endif
/* Solaris 2.3 */
#ifndef AUDIO_ENCODING_DVI
#define AUDIO_ENCODING_DVI     (104)
#define AUDIO_ENCODING_LINEAR8 (105)
#endif
static char rcsid[] = "$Id: sun4_audio.c,v 1.1.1.1 1998/04/21 17:05:50 sho Exp $";

/*
* Open audio control device. The BSD audio driver doesn't have one, so
* ac and au_* are treated as one and the same in that case.
* Also, set up SIGPOLL handler if track_audio_state flag is TRUE.
* Return -1 on error, 0 on success.
*/
static int ahw_open_control(void)
{
  if (ac < 0) {
    AUDIO_INITINFO(&ai_global);
    ac = open(AUDIO_CTL, O_RDWR);
    /* no control device */
    if (ac < 0) {
      if (errno == ENODEV) ac = au[0];
      else return -1;
    }
  }
  return 0;
} /* ahw_open_control */

/*
* Set precision of current audio device.
* SunOS and Solaris use the same call, but different parameter types.
*/
static int ahw_getdev(void)
{
#ifdef sun4
  int dev;       /* device type */ 
#else
  audio_device_t dev;
#endif
  int status;    /* ioctl() status */

  /* return if already set */
  if (precision > 0) return 0;

  /* need control device */
  if (ahw_open_control() < 0) return -1;

  /* determine what kind of device we have */
  status = ioctl(ac, AUDIO_GETDEV, &dev);
  /* 
   * If it's 4.1.1, it doesn't know AUDIO_GETDEV and we'll always
   * have to treat this as an 8-bit (amd) device.
   */
  if (status < 0) precision = 8;
#ifdef sun4
  else if (dev > AUDIO_DEV_AMD) precision = 16;
  else precision = 8;
#else
  else {
    if      (strcmp(dev.name, "SUNW,dbri")    == 0) precision = 16;
    else if (strcmp(dev.name, "SUNW,CS4231")  == 0) precision = 16;
    else if (strcmp(dev.name, "SUNW,am79c30") == 0) precision = 8;
    else precision = 8;
  }
#endif
  return 0;
} /* ahw_getdev */

/*
* Open audio hardware and set file descriptor to non-blocking.  Return
* file descriptor number or -1 on error.  If the output device is
* busy, we don't fail, but would try again later, thus grabbing the
* audio device as quickly as possible.  This allows short
* announcements (e.g., biff) to grab the audio device without crashing
* this program.  If the other program changes the device
* characteristics, there is potential for disaster.  Several nevot's
* can run concurrently if the nevot.talk is used to disable the
* recording device (iff there is no audio output waiting).
*/
int ahw_open(int ndelay)
{
  int fd_flags;     /* fcntl flags */
  audio_info_t ai;
  int i;

  if (ahw_getdev() < 0) return -1;     /* get precision (device type) */

  /* 
   * If we haven't called ahw_open_control(), initialize ai_global.
   */
  if (ai_global.record.sample_rate == 0) AUDIO_INITINFO(&ai_global);

  /*
   * Set up audio format parameters.  Since AUDIO_SETINFO also
   * retrieves the current settings, we use a temporary variable. 
   * Otherwise, we are saving the settings even though they (like port
   * or gain) may have been changed elsewhere.
   */
  AUDIO_INITINFO(&ai);
  ai.play.sample_rate = ai_global.play.sample_rate;
  ai.play.encoding    = ai_global.play.encoding;
  ai.play.precision   = ai_global.play.precision;
  ai.play.channels    = ai_global.play.channels;
  ai.record = ai.play;

  /* set input buffering depth, if supported (SunOS 5.x) */
#ifdef AUDIO_DIAG_LOOPBACK
  ai.record.buffer_size = 160;
#endif 

  /*
   * If 8-bit device, we open separate file descriptors to
   * accomodate broken sun4c/4.1.3 hosts that need it. For others
   * we use a single file descriptor (SPARC 10's can't handle
   * separate descriptors...).
   */
  if (precision == 8) {
    for (i = 0; i < 2; i++) {
      if ((au[i] = open(AUDIO_DATA, (i ? O_WRONLY:O_RDONLY))) < 0) {
        return -1;
      }
      if (ioctl(au[i], AUDIO_SETINFO, &ai) < 0) {
        return -1;
      }
    }
  }
  else {
    if ((au[0] = open(AUDIO_DATA, O_RDWR)) < 0) {
      return -1;
    }
    audio_pause(au[0]);
    if (ioctl(au[0], AUDIO_SETINFO, &ai) < 0) {
      return -1;
    }
    audio_flush(au[0]);
    audio_resume(au[0]);
    au[1] = au[0];
  }

  /* if desired, set input to non-blocking */
  if (ndelay) {
    fd_flags = fcntl(au[0], F_GETFL, 0);
    (void)fcntl(au[0], F_SETFL, fd_flags|O_NDELAY);
  }
  rec_count = play_count = 0;

  return au[0];
} /* ahw_open */

/*
* Close audio hardware (reading and writing); also drains audio
* device.  Return old (now closed) input file descriptor to allow
* disabling input handlers.  Return -1 on error.
*/
int ahw_close(void)
{
  int au_save;

  au_save = au[0];
  if (au[0] > 0) {
    if (close(au[0]) < 0) return -1;
  }
  if (au[1] > 0 && (au[0] != au[1])) {
    if (close(au[1]) < 0) return -1;
  }
  au[0] = au[1] = -1;
  return au_save;
} /* ahw_close */

/*
* Establish function that tracks audio device changes.
*/
void ahw_track(signal_func_t func)
{
  if (ahw_open_control() < 0) return;

  if (track == func) return;
  track = func;
  if (track) {
    signal(SIGPOLL, track);
    if (ioctl(ac, I_SETSIG, S_MSG) < 0) return;
  }
  else {
    signal(SIGPOLL, SIG_IGN);
  }
} /* ahw_track */

/*
* Set audio hardware through control device. 
* If value < 0, don't change.
* Disable polling signal. Return 0 if ok, -1 if not.
* pr     in   0: record, 1: play, 2: monitor
* p      in   parameter to set
* value  in   value to be used; -1: don't change
*/
int ahw_set(int pr, audio_set_t p, int value)
{
  int success = -1;              /* return status */
  audio_info_t ai;               /* use if we can actually set it right now */
  audio_prinfo_t *ap;            /* play/record properties */

  /* nothing to change */
  if (value < 0) return 0;
  if (ahw_open_control() < 0) return -1;

  AUDIO_INITINFO(&ai);
  ap = (pr == 0) ? &ai.record : &ai.play;

  switch(p) {
  case AS_port:
    ap->port = 0;
    if (pr == 0) {
      if      (value == AP_in_line)       ap->port = AUDIO_LINE_IN;
      else if (value == AP_in_microphone) ap->port = AUDIO_MICROPHONE;
      else return -1;
    }
    else {
      if (value & AP_out_speaker)   ap->port |= AUDIO_SPEAKER;
      if (value & AP_out_headphone) ap->port |= AUDIO_HEADPHONE;
      if (value & AP_out_line)      ap->port |= AUDIO_LINE_OUT;
      if (value & ~(AP_out_speaker|AP_out_headphone|AP_out_line))
        return -1;
    }
    value = (ap->port > 0) ? ap->port : -1;
    break;

  case AS_gain_left:
  case AS_gain_right:
  case AS_gain:
    if (pr == 2) {
      ai_global.monitor_gain = ai.monitor_gain = value;
    }
    else ap->gain = value;
    break;

  case AS_gain_step:
    break;

  case AS_balance:
#ifdef AUDIO_DIAG_LOOPBACK
    ap->balance = value;
#endif
    break;

  case AS_fillpoint:
    break;
  } /* switch */

  if (track) signal(SIGPOLL, SIG_IGN);
  success = ioctl(ac, AUDIO_SETINFO, &ai);
  if (track) signal(SIGPOLL, track);
  if (success >= 0) {
     if (p == AS_port)      
        success = (value == ap->port ? 0 : -1);
  }
  return success;
} /* ahw_set */

/*
* Set audio device to best approximation of 'record' and 'play'. 
* Since playing and recording and the individual parameters are not
* independent, we have to set them together.  On return, 'record' and
* 'play' contain the setting used.  Record overrides play.  For Suns,
* they will be forced to be the same.  Return -1 on error (illegal
* rate for amd device).
*/
int ahw_fmt(audio_descr_t *record, audio_descr_t *play)
{
  int status;
  double m;

  /* set precision value */
  ahw_getdev();

  /* we treat sampling rates close to multiples of 8000 Hz as the same */
  m = record->sample_rate / 8000.;
  if (m - (int)m < 0.1) {
    record->sample_rate = (int)m * 8000;
  }

  /*
   * For amd (8-bit only), we don't have much choice.  We *can* convert
   * between mono and stereo, so we simply force mono.
   */
  if (precision == 8) {
    if (record->encoding != AE_PCMA)
      record->encoding = AE_PCMU;
    if (record->sample_rate != 8000) return -1;
    record->channels = 1;
  }
  /*
   * If the rate is not close to 8000 or we want stereo, we need to use
   * LINEAR encoding.
   */
  else if (precision == 16) {
    if (record->encoding == AE_L16 ||
        record->sample_rate != 8000 ||
        record->channels > 1) {
      record->encoding = AE_L16;
    } 
    else if (record->encoding != AE_PCMA) {
      record->encoding = AE_PCMU;
    }
  }
  /*
   * Undefined device type.
   */
  else return -1;

  /*
   * Play and record format have to be the same; we can't compensate
   * for sampling rate variation.
   */
  *play = *record;

  ai_global.record.channels    = record->channels;
  ai_global.record.sample_rate = record->sample_rate;
  if (record->encoding == AE_L16) {
    ai_global.record.precision = 16;
    ai_global.record.encoding  = AUDIO_ENCODING_LINEAR;
  }
  else if (record->encoding == AE_PCMU) {
    ai_global.record.precision = 8;
    ai_global.record.encoding  = AUDIO_ENCODING_ULAW;
  }
  else if (record->encoding == AE_PCMA) {
    ai_global.record.precision = 8;
    ai_global.record.encoding  = AUDIO_ENCODING_ALAW;
  }
  /*
   * Since ai_global.play and record are the same, we can check whether
   * we need to adjust the audio device at all.
   */
  if (ai_global.record.precision   != ai_global.play.precision ||
      ai_global.record.channels    != ai_global.play.channels ||
      ai_global.record.encoding    != ai_global.play.encoding ||
      ai_global.record.sample_rate != ai_global.play.sample_rate) {
    ai_global.play = ai_global.record;

    /* 
     * Set physical audio device if that device is open.
     * Otherwise, this will be done when it is opened.
     * Can only do this if both descriptors are the same.
     * Only touch sample_rate, encoding, precision, channels.
     * Pause and flush audio device when changing its parameters.
     */
    if (au[0] >= 0) {
      audio_info_t ai;
      AUDIO_INITINFO(&ai);
      ai.play.sample_rate = ai_global.play.sample_rate;
      ai.play.encoding    = ai_global.play.encoding;
      ai.play.precision   = ai_global.play.precision;
      ai.play.channels    = ai_global.play.channels;
      ai.record = ai.play;
      audio_pause(au[0]);
      audio_flush(au[0]);
      status = ioctl(au[0], AUDIO_SETINFO, &ai);
      audio_resume(au[0]);
      if (status < 0) return -1;
    }
  }

  return 0;
} /* ahw_fmt */

/*
* Return number of samples in write queue, <0 if underflow.  If the
* number is negative, an underflow has occurred and we alse set
* 'underflow'.
*/
int ahw_write_queue(int *underflow)
{
  audio_info_t audio_info;

  ioctl(au[1], AUDIO_GETINFO, (char *)&audio_info);
  *underflow = audio_info.play.error; 
  if (audio_info.play.error) {
    AUDIO_INITINFO(&audio_info);
    audio_info.play.error = 0;
    ioctl(au[1], AUDIO_SETINFO, (char *)&audio_info);
  }
  return play_count * (audio_info.play.precision >> 3) - audio_info.play.samples;
} /* ahw_write_queue */


/*
* Return occupancy of read queue in samples.  Set 'overflow' if read
* queue lost audio samples.
*/
int ahw_read_queue(int *overflow)
{
  audio_info_t audio_info;

  ioctl(au[0], AUDIO_GETINFO, (char *)&audio_info);
  *overflow = audio_info.record.error;
  if (audio_info.record.error) { /* reset error indicator */
    AUDIO_INITINFO(&audio_info);
    audio_info.play.error = 0;
    ioctl(au[0], AUDIO_SETINFO, (char *)&audio_info);
  }
  return audio_info.record.samples - rec_count * (audio_info.record.precision >> 3);
} /* ahw_read_queue */

/*
* Get current hardware audio status.
* Return -1 if device error, 0 if success.
*/
int ahw_get(audio_state_t *as)
{
  audio_info_t a;
  audio_prinfo_t *pr;
  audio_ws_t *ws;
  static int errors[2];
  int i;

  if (ahw_open_control() < 0) return -1;
  if (ioctl(ac, AUDIO_GETINFO, (char *)&a) < 0) return -1;
  ahw_changed = 0;  /* reset flag */

  for (i = 0; i < 2; i++) {
    pr = (i == 0) ? &a.play : &a.record;
    ws = (i == 0) ? &as->play : &as->record; 

    ws->waiting = pr->waiting;
    ws->gain[0] = ws->gain[1] = pr->gain;
#ifdef AUDIO_DIAG_LOOPBACK
    ws->buffer_size = pr->buffer_size;
#else
    ws->buffer_size = 0;
#endif 

    if      (pr->encoding == AUDIO_ENCODING_ULAW)    ws->ad.encoding = AE_PCMU;
    else if (pr->encoding == AUDIO_ENCODING_ALAW)    ws->ad.encoding = AE_PCMA;
    else if (pr->encoding == AUDIO_ENCODING_LINEAR)  ws->ad.encoding = AE_L16;
    else if (pr->encoding == AUDIO_ENCODING_DVI)     ws->ad.encoding = AE_DVI4;
#ifdef AUDIO_ENCODING_LINEAR8
    else if (pr->encoding == AUDIO_ENCODING_LINEAR8) ws->ad.encoding = AE_L8;
#endif

    ws->ad.sample_rate = pr->sample_rate;
    ws->ad.channels    = pr->channels;

    if (pr->error) errors[i]++;
    ws->errors    = errors[i];
    ws->samples   = pr->samples;
    ws->active    = pr->active;
  }

  as->play.port = 0;
  if (a.play.port & AUDIO_SPEAKER)   as->play.port |= AP_out_speaker;
  if (a.play.port & AUDIO_HEADPHONE) as->play.port |= AP_out_headphone;
  if (a.play.port & AUDIO_LINE_OUT)  as->play.port |= AP_out_line;

  as->record.port = 0;
  if (a.record.port & AUDIO_MICROPHONE) as->record.port |= AP_in_microphone;
  if (a.record.port & AUDIO_LINE_IN)    as->record.port |= AP_in_line;

  as->monitor_gain[0] = a.monitor_gain;
  as->monitor_gain[1] = a.monitor_gain;

  return 0;
} /* ahw_get */

/*
* Return > 0 if another process is waiting for audio input or output.
* Return -1 if error.
*/
int ahw_waiting(void)
{
  audio_info_t a;

  if (ioctl(ac, AUDIO_GETINFO, (char *)&a) < 0) return -1;
  return (a.play.waiting << 1) + a.record.waiting;
} /* ahw_waiting */


/*
* Write audio buffer to audio device.
*/
int ahw_write(char *buffer, int bytes)
{
  if (bytes == 0) return bytes;
  if (au[1] >= 0) {
    bytes = write(au[1], buffer, bytes);
    if (bytes > 0) play_count += bytes;  
  }
  return bytes;
} /* ahw_write */


/*
* Read audio samples from audio device.
* buffer  out   buffer to receive audio data
* bytes   in    number of audio bytes to read
*/
int ahw_read(char *buffer, int bytes)
{
  bytes = read(au[0], buffer, bytes);
  if (bytes > 0) rec_count += bytes;  
  return bytes;
} /* ahw_read */
