/*
 * @(#) $Id:
 */

/************************ rsvp.cc  ***********************************
 *                                                                   *
 *      Routines to work the RSVP generic API to use RSVP features   *
 *		in vic.                                            *
 *                                                                   *
 *                                                                   *
 *********************************************************************/

/****************************************************************************

            	RSVP code for use with VIC

                USC Information Sciences Institute
                Marina del Rey, California

	    Current Version:  Steven Berson, July 1996.

  Copyright (c) 1996 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for any purpose and without
  fee is hereby granted, provided that both the above copyright notice
  and this permission notice appear in all copies, and that any
  documentation, advertising materials, and other materials related to
  such distribution and use acknowledge that the software was developed
  in part by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of this software and are so
  noted when applicable.

********************************************************************/


#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>



#include "Tcl.h"
#include "tk.h"




extern int debug;
extern "C" {
int close(int);
}



extern "C" {
#	include "rapi_lib.h"
#	include "rapi_err.h"
}





/* Defines the source sending path packests seen by RSVP daemon */
class source {
public:
  source();
  source(source*, rapi_filter_t*, rapi_flowspec_t*);
  source* next() {return(next_);}
  void next(source *sp) {next_ = sp;}
  int eq(int addr, unsigned short port);
  int eq(int addr);
  void setsid(int sid) {sid_ = sid;}
  void setReservationState(int state) {reservation_ = state; }
  int sid() {return(sid_);}
  int reservationState() {return(reservation_);}
  rapi_flowspec_t* flowspec() {return(&flowSpec_);}
  rapi_tspec_t* tspec() {return(&tSpec_);}
  rapi_filter_t* filter() {return(&filtSpec_);}
  void update_tspec_old(rapi_flowspec_t*);
  void update_tspec(char *, const char *const *argv, int);
  void update_flowspec_old(rapi_flowspec_t*);
  void update_flowspec(char *, const char *const *argv, int);
private:
  int sid_;
  int newState_;
  int sending_;
  int reservation_;
  rapi_tspec_t tSpec_;
  rapi_flowspec_t flowSpec_;
  rapi_filter_t filtSpec_;
  source* next_;
};

/* Defines the RSVP session */
class session {
public:
  session();
  session(session*);
  ~session();
  
  void update_flowspec(char *, const char *const *argv, int);
  void update_Tspec(char *, const char *const *argv, int);
  session* next() {return(next_);}
  session* next(session *ses) { next_ = ses;}
  source* get_source(int);
  source* find_source(struct sockaddr_in, rapi_filter_t*, rapi_flowspec_t*);
  void delete_source(source*);
  source* get_sources_() {return(sources_);}
  void set_sources_(source *src) {sources_ = src;}
  qos_tspecx_t get_Tspec() {return(Tspec);}
  qos_flowspecx_t get_flowspec() {return(flowspec);}
  int get_flownumber(){return(flownumber);}
  void set_flownumber(int i){ flownumber += i;}
  int get_filtertype(){return(filtertype);}
  void set_filtertype(int i){filtertype = i;}
  int get_pathstate_rcv_(){return(pathstate_rcv_);}
  void set_pathstate_rcv_(int i){pathstate_rcv_ = i;}
  int get_pathstate_snt_(){return(pathstate_snt_);}
  void set_pathstate_snt_(int i){pathstate_snt_ = i;}
  int get_sid(){return(sid_);}
  void set_sid(int sid){sid_ = sid;}
  int get_dstaddr(){return(dstaddr_);}
  void set_dstaddr(int addr){dstaddr_ = addr;}
  unsigned short get_dstport(){return(dstport_);}
  void set_dstport(int port){dstport_ = port;}
  int localaddr(){return(local_addr);}
  void localaddr(u_long localaddr){local_addr = localaddr;}
protected:
private:
  int sid_;
  unsigned int dstaddr_;
  unsigned short dstport_;
  int pathstate_rcv_;
  int pathstate_snt_;
  int flownumber;
  int filtertype;
  qos_flowspecx_t flowspec;
  qos_tspecx_t Tspec;
  source* sources_;
  session* next_;
  u_long local_addr;
};

/* manages the set of sessions and sources */
class Rsvp: public TclObject {
public:
  Rsvp(const char *);
  ~Rsvp();
  virtual int command(int argc, const char*const* argv);
  int send_reserve(int, struct sockaddr_in *);
  source* find_source(int);
  session* get_session(int);
protected:
private:
  int do_session(char *, const char *const *argv);
  int do_sender(char *, const char *const *argv, int);
  int redo_sender(char *);
  int get_sid(int, unsigned short);
  session* get_session(int, unsigned short);
  int get_path_state(int);
  void delete_session(int);
  
  session* session_;
  int fd_;
  int sid_;
  int sending_;
  unsigned short dstport_;
  int dstaddr_;
  unsigned short srcport_;
  int srcaddr_;
  u_char ttl_;
};

Rsvp::Rsvp(const char *name) : TclObject(name) {
  session_ = NULL;
  fd_ = -1;
  sid_ = -1;
  sending_ = 0;
  dstport_ = 0;
  dstaddr_ = 0;
  srcport_ = 0;
  srcaddr_ = 0;
  ttl_ = 0;
}

Rsvp::~Rsvp() {
  close(fd_);
  fd_ = -1;
}

static Rsvp RsvpObject("rsvp");

session::session() {
  sid_ = -1;
  dstaddr_ = 0;
  dstport_ = 0;
  next_ = NULL;
  flownumber = 0;
  sources_ = NULL;
  local_addr = 0;
}

session::session(session *sp) {
  next_ = sp;
  pathstate_snt_ = 0;
  pathstate_rcv_ = 0;
  flownumber = 0;
  filtertype = 0;
  sources_ = NULL;
  local_addr = 0;
  Tspec.spec_type = QOS_TSPEC;
  Tspec.xtspec_b = 128000/8;
  Tspec.xtspec_r = 512000/8;
  Tspec.xtspec_p = 0;//500000/8;
  Tspec.xtspec_m = 0;//9188;
  Tspec.xtspec_M = 65535;
  flowspec.spec_type = QOS_CNTR_LOAD;
  flowspec.xspec_b = 128000/8;
  flowspec.xspec_r = 512000/8;
  flowspec.xspec_p = 0;
  flowspec.xspec_m = 0;
  flowspec.xspec_M =  65535;
  flowspec.xspec_R = 0;
  flowspec.xspec_S = 0;}

session::~session() {
}

/* update of the flow specificication for the session, used while working with WILDACRD filter */
void session::update_flowspec(char* cp, const char*const* argv, int sid) {
  //flowspec.spec_type = QOS_TSPEC;
  flowspec.spec_type = atoi(argv[4]);
  flowspec.xspec_b = atoi(argv[5]);
  flowspec.xspec_r = atoi(argv[6]);
  flowspec.xspec_p = atoi(argv[7]);
  flowspec.xspec_m = atoi(argv[8]);
  flowspec.xspec_M = atoi(argv[9]);
  flowspec.xspec_R = atoi(argv[10]);
  flowspec.xspec_S = atoi(argv[11]);
}



/* update of the tspec specificication for the session, needed while sending PATH packets */
void session::update_Tspec(char* cp, const char*const* argv, int sid) {
  //Tspec.spec_type = QOS_TSPEC;
  Tspec.spec_type = atoi(argv[4]);
  Tspec.xtspec_b = atoi(argv[5]);
  Tspec.xtspec_r = atoi(argv[6]);
  Tspec.xtspec_p = atoi(argv[7]);
  Tspec.xtspec_m = atoi(argv[8]);
  Tspec.xtspec_M = atoi(argv[9]);
}

source::source() {
  sending_ = 0;
  newState_ = 0;
  next_ = NULL;
}

source::source(source *sp, rapi_filter_t* filtp, rapi_flowspec_t* flowp) {
  sending_ = 0;
  reservation_ = 0;
  newState_ = 0;
  memcpy(&filtSpec_, filtp, filtp->len);
  memcpy(&tSpec_, flowp, flowp->len);
  next_ = sp;
}

/* updates the flow spec of the source on the base of the tspec coming in PATH packets */
void source::update_tspec_old(rapi_flowspec_t* flowp) {
  memcpy(&tSpec_, flowp, flowp->len);
}

/* updates the flow spec of the source on the base of the tspec coming in PATH packets */
void source::update_flowspec_old(rapi_flowspec_t* flowp) {
  memcpy(&flowSpec_, flowp, flowp->len);
}

/* compares the sources on the base of the address */
int
source::eq(int addr) {
  struct sockaddr_in sin = filtSpec_.filt_u.base.sender;
  return(sin.sin_addr.s_addr == addr);
}

/* compares the sources on the base of the address and port number*/
int
source::eq(int addr, unsigned short port) {
  struct sockaddr_in sin = filtSpec_.filt_u.base.sender;
  return( (sin.sin_addr.s_addr == addr) && (sin.sin_port == port) );
}

/* updates the flow spec of the source when user defines it*/
void source::update_flowspec(char* cp, const char*const* argv, int sid) {
  //flowSpec_.specbody_qosx.spec_type = 5;
  flowSpec_.specbody_qosx.spec_type = atoi(argv[4]);
  flowSpec_.specbody_qosx.xspec_b = atoi(argv[5]);
  flowSpec_.specbody_qosx.xspec_r = atoi(argv[6]);
  flowSpec_.specbody_qosx.xspec_p = atoi(argv[7]);
  flowSpec_.specbody_qosx.xspec_m = atoi(argv[8]);
  flowSpec_.specbody_qosx.xspec_M = atoi(argv[9]);
}

/* updates the flow spec of the source when user defines it*/
void source::update_tspec(char* cp, const char*const* argv, int sid) {
  //tSpec_.tspecbody_u.qosxt.spec_type = 5;
  tSpec_.tspecbody_u.qosxt.spec_type = atoi(argv[4]);
  tSpec_.tspecbody_u.qosxt.xtspec_b = atoi(argv[5]);
  tSpec_.tspecbody_u.qosxt.xtspec_r = atoi(argv[6]);
  tSpec_.tspecbody_u.qosxt.xtspec_p = atoi(argv[7]);
  tSpec_.tspecbody_u.qosxt.xtspec_m = atoi(argv[8]);
  tSpec_.tspecbody_u.qosxt.xtspec_M = atoi(argv[9]);
}
static int rerrno;

void
Tk_dispatch_read(ClientData data, int mask) {
  if (mask == TK_EXCEPTION) {
if(debug) {
    printf("TK_EXCEPTION signalled - shouldn't happen\n");
}
  } else if (mask == TK_READABLE) {
    int err = rapi_dispatch();
    if (err != 0) {
if(debug) {
      printf("rapi_dispatch failed, %s\n", rapi_errlist[err]);
}
    }
    if (err == RAPI_ERR_NORSVP) {
if(debug) {
      printf("Deleting file handler\n");
}
      Tcl_DeleteFileHandler(data);
    }
  } else {
if(debug) {
    printf("Unknown condition signalled - %d\n", mask);
}
  }
}

/* handles the asynchronous callbacks */
int
async(
      rapi_sid_t sid, 
      rapi_eventinfo_t event, 
      int style, 
      int code, 
      int value, 
      sockaddr *node,
      u_char flags,
      int filterCount,
      rapi_filter_t *filterList,
      int flowCount,
      rapi_flowspec_t *flowList,
      int adCount,
      rapi_adspec_t *adList,
      void *arg) {
  
  char buf[1024];
  char command[1000];
  int i;
  
  Tcl& tcl = Tcl::instance();
  char* cp = tcl.result();
  session* se = RsvpObject.get_session(sid);
  if(se==NULL)
    return 0;
  struct sockaddr_in tmp;
  tmp.sin_addr.s_addr = se->get_dstaddr();
  tmp.sin_port = se->get_dstport();
  if (event == RAPI_PATH_ERROR) {
if(debug) {
    printf("%d, value = %d\n", rsvp_errlist[code], value);
}
    bzero(command,strlen(command));
    sprintf(command,"Message %s %d {RAPI_PATH_ERROR %s,value= %d}",inet_ntoa(tmp.sin_addr),ntohs(tmp.sin_port),rsvp_errlist[code],value);
    tcl.evalc(command);
    return code;
  } else if (event == RAPI_RESV_ERROR) {
    bzero(command,strlen(command));
    sprintf(command,"Message %s %d {RAPI_RESV_ERROR %s,value= %d}",inet_ntoa(tmp.sin_addr),ntohs(tmp.sin_port),rsvp_errlist[code],value);
    tcl.evalc(command);
if(debug) {
    printf("%s, value = %d\n", rsvp_errlist[code], value);
    rapi_fmt_flowspec(flowList, buf, 1024);
    printf("flowspec = %s\n", buf);
}
    return code;
  } else if (event == RAPI_RESV_EVENT) {
    bzero(command,strlen(command));
    sprintf(command,"Message %s %d RAPI_RESV_EVENT",inet_ntoa(tmp.sin_addr),ntohs(tmp.sin_port));
    tcl.evalc(command);
    
if(debug) {
    printf("RESV_EVENT\n");
    rapi_fmt_flowspec(flowList, buf, 1024);
    printf("flowspec = %s\n", buf);
}
    return 0;
  } else if (event != RAPI_PATH_EVENT) {
if(debug) {
    printf("Unknown event type %d\n", event);
}
    return -1;
    bzero(command,strlen(command));
    sprintf(command,"Message %s %d {Unknown event type %d}",inet_ntoa(tmp.sin_addr),ntohs(tmp.sin_port), event);
    tcl.evalc(command);
  }
      

  sprintf(command,"Message %s %d RAPI_PATH_EVENT",inet_ntoa(tmp.sin_addr),ntohs(tmp.sin_port));
  tcl.evalc(command);
if(debug) {
  printf("PATH_EVENT \n");  
  printf("tspec = %s\n", buf);
  printf("filterCount = %d\n", filterCount);
}
  if (filterCount != flowCount) {
if(debug) {
    printf("flowCount != filterCount\n");
}
  }
  rapi_filter_t* filtp = filterList;
  rapi_flowspec_t* flowp = flowList;
  if (se->get_flownumber() <= flowCount) {
    for (i = 0; i < filterCount; i++) {
      if (filtp->form != RAPI_FILTERFORM_BASE) {
if(debug) {
	printf("Bad filtertype\n");
}
      }
      struct sockaddr_in sin = filtp->filt_u.base.sender;
      if(debug) {
	printf("Source %d = %s/%d\n", i,inet_ntoa(sin.sin_addr), sin.sin_port);
      }
      if (flowp->form != RAPI_TSPECTYPE_Simplified) {
	printf("Bad flowspec form - %d\n", flowp->form);
      }
      qos_flowspecx_t * cszp = &flowp->specbody_u.qosx;
      if (cszp->spec_type != QOS_TSPEC) {
	printf("Bad qos_type %d\n", cszp->spec_type);
      }
if(debug) {
      printf("Bandwidth = %f\n", cszp->xspec_r);
}
      source *sp = se->find_source(sin, filtp, flowp);
      sp->setsid(sid);
      sp->update_tspec_old(flowp);
      
      filtp = (rapi_filter_t *)((char *)filtp + filtp->len);
      flowp = (rapi_flowspec_t *)((char *)flowp + flowp->len);
      if (se != NULL) {
	se->set_pathstate_rcv_(1);
      }
      
      struct sockaddr_in tmp;
      tmp.sin_addr.s_addr = se->get_dstaddr();
      tmp.sin_port = se->get_dstport();
      
      char addr1[16];
      char addr2[16];
      bzero(addr1,16);
      bzero(addr2,16);
      strcpy(addr1, inet_ntoa(tmp.sin_addr));
      strcpy(addr2, inet_ntoa(sin.sin_addr));
      
      /* update the flowspec for the reservation */
      bzero(command,strlen(command));
      bzero(buf, 1024);
      rapi_fmt_tspec((rapi_tspec_t *)(flowList+i*sizeof(rapi_tspec_t)), buf, 1024);
      sprintf (command,"add_src %s %d %s {%s}",addr1, ntohs(tmp.sin_port), addr2,buf);
      tcl.evalc(command);
      if (se->get_flownumber() < flowCount) {
	se->set_flownumber(1);
      }
    }
  } else if (se->get_flownumber() > flowCount) {
    int res = 0;
    source* src = se->get_sources_();
    while ((res == 0)&&(src != NULL)&&(flowCount > 0)) {
      int i = 0;
      while ((i <= flowCount)&&
	     (!src->eq(filtp->filt_u.base.sender.sin_addr.s_addr))) {
	filtp = (rapi_filter_t *)((char *)filtp + filtp->len);
	flowp = (rapi_flowspec_t *)((char *)flowp + flowp->len);
	i++;
      }
      if((i<= flowCount)&&(src->eq(filtp->filt_u.base.sender.sin_addr.s_addr))) {
	src = src->next();
      } else {
	res = 1;
      }
    }
    
    rapi_filter_t* filtspec = src->filter();
    struct sockaddr_in sin = filtspec->filt_u.base.sender;
    
    char addr1[12];
    char addr2[12];
    strcpy(addr1, inet_ntoa(tmp.sin_addr));
    strcpy(addr2, inet_ntoa(sin.sin_addr));
    
    bzero(command,strlen(command));
    sprintf (command, "rm_src %s %d %s", addr1, ntohs(tmp.sin_port), addr2);
    tcl.evalc(command);
    se->set_flownumber(-1);
    se->delete_source(src);
    source *sources = se->get_sources_();
    if (sources == NULL) {
      se->set_pathstate_rcv_(0);
    }
    struct sockaddr_in localaddr;
    if (se->localaddr() != 0) {
	    localaddr.sin_family = AF_INET;
	    localaddr.sin_addr.s_addr = se->localaddr();
	    /*
	      localaddr.sin_port = se->get_dstport();*/
    } else {
	    localaddr.sin_family = INADDR_ANY;
    }
    RsvpObject.send_reserve(sid, &localaddr);
  }
}

/* determines the source while receiving PATH packets */
source*
session::find_source(struct sockaddr_in sin, 
		     rapi_filter_t* filtp, 
		     rapi_flowspec_t* flowp) {
  source* sp;
  for (sp = get_sources_(); sp != NULL; sp = sp->next()) {
    if (sp->eq(sin.sin_addr.s_addr, sin.sin_port)) {
      return sp;
    }
  }
  sp = new source(get_sources_(), filtp, flowp);
  sources_ = sp;
  return(sp);
}

/* sends rapi_reserve to RSVP daemon */
int
Rsvp::send_reserve(int sid, struct sockaddr_in *local) {
  source* src;
  int nspec = 0, fspec = 0;
  char specbuf[4096];
  char filtbuf[4096];
  rapi_flowspec_t *specp = (rapi_flowspec_t *)specbuf;
  rapi_filter_t *filtp = (rapi_filter_t *)filtbuf;
  qos_flowspecx_t cszxp;
  session *sp = get_session(sid);
  rapi_styleid_t type_of_filter;
  if ((sp->get_filtertype() == 0)&&(sp->get_pathstate_rcv_() == 1)) {
    specp->form = RAPI_FLOWSTYPE_Simplified;
    specp->len = sizeof(rapi_flowspec_t);
    qos_flowspecx_t* cszxp = &specp->specbody_qosx;
    memcpy(cszxp, &sp->get_flowspec(), sizeof(qos_tspecx_t));
    //cszxp->spec_type = QOS_CNTR_LOAD;
    cszxp->xspec_R = 0;
    cszxp->xspec_S = 0;
    nspec = 1;
    fspec = 0;
    type_of_filter = RAPI_RSTYLE_WILDCARD;
    specp = (rapi_flowspec_t *)specbuf;
    filtp = (rapi_filter_t *)filtbuf;
  } else {
    for (src = sp->get_sources_(); src!= NULL; src = src->next()) {
      if ((src->sid() != sid)|| (src->reservationState() == 0))
	continue;
      rapi_tspec_t* tsp = src->tspec();
if(debug) {
      printf("Tspec %f \n", tsp->tspecbody_qosx.xtspec_r);
}
      // Copy the generic header
      memcpy(specp, tsp, sizeof(rapi_hdr_t));
      qos_flowspecx_t* cszxp = &specp->specbody_qosx;
      memcpy(cszxp, &tsp->tspecbody_qosx, sizeof(qos_tspecx_t));
      //cszxp->spec_type = QOS_CNTR_LOAD;
      cszxp->xspec_R = 0;
      cszxp->xspec_S = 0;
      specp->form = RAPI_FLOWSTYPE_Simplified;
      specp->len += (sizeof(qos_flowspecx_t) - sizeof(qos_tspecx_t));
      rapi_filter_t* fsp = src->filter();
      memcpy(filtp, fsp, fsp->len);
      specp = (rapi_flowspec_t*) ((char *)specp+specp->len);
      
      filtp = (rapi_filter_t*) ((char *)filtp+filtp->len);
      nspec++;
      fspec++;
    }
    type_of_filter = RAPI_RSTYLE_FIXED;
    specp = (rapi_flowspec_t *)specbuf;
    filtp = (rapi_filter_t *)filtbuf;
  }
  int code = rapi_reserve(sid, 0, (struct sockaddr *)local, type_of_filter, 0, 0, fspec, 
			  filtp, nspec, specp);
  if (code != 0) {
    Tcl& tcl = Tcl::instance();
    char* cp = tcl.buffer();
    tcl.result(cp);
if(debug) {
    printf("error in rapi_reserve() %d\n", code);
}
    sprintf(cp, "rapi_reserve() error, sid_out = %d\n", sid);
    return(TCL_ERROR);
  }
  return(TCL_OK);
}

/* handles tcl/tk commands */
int
Rsvp::command(int argc, const char*const* argv) {
  Tcl& tcl = Tcl::instance();
  char* cp = tcl.buffer();
  sprintf(cp,"");
  tcl.result(cp);
  
  if (strcmp(argv[1], "open") == 0) {
    if (argc == 2) {
      struct sockaddr_in addr;
      memset(&addr, 0, sizeof(addr));
      addr.sin_family = AF_INET;
      char cp[12];
      sprintf (cp, "224.1.1.1"); 
      addr.sin_addr.s_addr = inet_addr(cp);
      addr.sin_port = 3333;
      sid_ = rapi_session((sockaddr *)&addr, IPPROTO_UDP, 0, async, 0, &rerrno);
      if (sid_ == -1) {
if(debug) {
	printf("Do some error handling stuff\n");
}
	return(TCL_ERROR);
      }
      fd_ = rapi_getfd(sid_);
      int mask = TK_READABLE | TK_EXCEPTION;
      Tcl_CreateFileHandler((ClientData)fd_, mask, 
			   Tk_dispatch_read, (ClientData)fd_);
      return(TCL_OK);
    }
  }	
  if (strcmp(argv[1], "close") == 0) {
    if (argc == 2) {
      int code = rapi_release(sid_);
      if (code != 0) {
	sprintf(cp, "rapi_release() error %d", code);
	return(TCL_ERROR);
      }
      return(TCL_OK);
    }	
  }	
  if (strcmp(argv[1], "rapi_release") == 0) {
    if (argc == 4) {
if(debug) {
  fprintf(stderr," rapi_releasing %s %i \n",argv[2],atoi(argv[3]));
}
      int sid = get_sid(inet_addr(argv[2]), atoi(argv[3]));
      if(sid<0) return(TCL_OK);
      /*      int code = rapi_sender(sid, 0, NULL, 0, 0, 0, 0, ttl_);*/
      int code = rapi_release(sid);
      if (code != 0) {
	sprintf(cp, "rapi_release() error %d", code);
	return(TCL_ERROR);
      }
      delete_session(sid);
      return(TCL_OK);
    } 
  }
  if (strcmp(argv[1], "rapi_session") == 0) {
    if (argc == 4) {
      int ztest = do_session(cp, argv);
      return(ztest);
    }
  } 
  if (strcmp(argv[1], "set_local") == 0) {
      session* sp = 
	get_session(inet_addr(argv[2]), atoi(argv[3]));
      struct hostent* hep = gethostbyname(argv[4]);
      struct in_addr in;
      in.s_addr = *(u_long *)(*hep->h_addr_list);
      sp->localaddr((u_long)in.s_addr);
      return(TCL_OK);
  }		
  if (strcmp(argv[1], "set_reservation") == 0) {
    //rsvp set_reservation host port args source 
    session* sp =
      get_session(inet_addr(argv[2]), atoi(argv[3]));
    if(sp==NULL) {
      fprintf(stderr," no session found \n");
      return;
    }
    source* src = sp->get_source(inet_addr(argv[5]));
    if(src!=NULL) {
      src->setReservationState(atoi(argv[4]));
    }
   return(TCL_OK);
  }
  if (strcmp(argv[1], "rapi_sender") == 0) {
    int sid = get_sid(inet_addr(argv[2]), atoi(argv[3]));
    session* sp = 
      get_session(inet_addr(argv[2]), atoi(argv[3]));
    if (sid == -1) {
      printf ("No sid for these address\n");
    } else if (atoi(argv[4]) == 1) {
      sp->set_pathstate_snt_(1);
      return(do_sender(cp, argv, sid));
    } else if ((atoi(argv[4]) == 0)&&
	       (sp->get_pathstate_snt_() == 1)){
      int code = 
	rapi_sender(sid, 0, NULL, 0, 0, 0, 0, ttl_);
      sending_ = 0;
      if (code != 0) {
	sprintf(cp, "rapi_sender() error %d", code);
	return(TCL_ERROR);
      }
      sp->set_pathstate_snt_(0);
      return(TCL_OK);
    }
  }

  if (strcmp(argv[1], "reserve") == 0) {
    int sid = get_sid(inet_addr(argv[2]), atoi(argv[3]));
    session* sp = 
      get_session(inet_addr(argv[2]), atoi(argv[3]));
    if (sid == -1) {
      printf ("No sid for these addressaaa\n");
    } else {
      struct sockaddr_in localaddr;
      if (sp->localaddr() != 0) {
	localaddr.sin_family = AF_INET;
	localaddr.sin_addr.s_addr = sp->localaddr();
	/* 
	   localaddr.sin_port = sp->get_dstport(); */
      } else {
	localaddr.sin_family = INADDR_ANY;
      }
      
      if ((atoi(argv[4]) == 1)&&
	  (sp->get_pathstate_rcv_() == 1)) {
	return(send_reserve(sid, &localaddr));
      } else if ((atoi(argv[4]) == 0)&&
		 (sp->get_pathstate_rcv_() == 1)) {
	
	int code = 
	  rapi_reserve(sid,0,(struct sockaddr *)&localaddr,
		       RAPI_RSTYLE_FIXED,0,0,0,0,0,0);
	return(TCL_OK);
      }
    }
  }
  if (strcmp(argv[1], "rapi_filtertype") == 0) {
    session* sp = get_session(inet_addr(argv[2]), atoi(argv[3]));
    if (sp == NULL) {
      printf ("No session for these address\n");
    } else if (atoi(argv[4]) == 1) {
      sp->set_filtertype(1);
    } else if (atoi(argv[4]) == 0) {
      sp->set_filtertype(0);
    }
    return(TCL_OK);
    
  }
  if (strcmp(argv[1], "rapi_flowspec") == 0) {
    if (argc == 13) {
	
      int sid = get_sid(inet_addr(argv[2]), atoi(argv[3]));
      session* sp = get_session(inet_addr(argv[2]), atoi(argv[3]));
      source* src = sp->get_source(inet_addr(argv[12]));
      if (sp->get_filtertype() == 0) {
	if(debug) {
	  printf("rapi_flowspec WF\n");
	}
	sp->update_flowspec(cp, argv, sp->get_sid());
      } else {
	if (src!=NULL&&src->sid() == sid) {
	  if(debug ){
	    printf("rapi_flowspec FF\n");
	  }
	  src->update_tspec(cp, argv, sid);
	}
      }
      return(TCL_OK);
    } 
  } 
  if (strcmp(argv[1], "rapi_tspec") == 0) {
    if (argc == 10) {
      int sid = get_sid(inet_addr(argv[2]), atoi(argv[3]));
      session* sp = get_session(inet_addr(argv[2]), atoi(argv[3]));
      sp->update_Tspec(cp, argv, sid);
      return(TCL_OK);
    }
  }
  return(TclObject::command(argc, argv));
}

/* sends rapi_session in order to establish a rsvp session */
int
Rsvp::do_session(char *cp, const char*const* argv) {
  struct sockaddr_in addr;
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = inet_addr(argv[2]);
  dstaddr_ = addr.sin_addr.s_addr;
  addr.sin_port = htons(atoi(argv[3]));
  dstport_ = addr.sin_port;
  int sid = rapi_session((sockaddr *)&addr, IPPROTO_UDP, 0, async, 0, &rerrno);
  if (sid == NULL_SID) {
	sprintf(cp, "Could not open connection to RSVP Daemon");
	return(TCL_OK);
  }
  /*if (IN_MULTICAST(addr.sin_addr.s_addr))
		sid_in_ = sid_out_;
	else {
		char name[MAXHOSTNAMELEN];
		if (gethostname(name,sizeof(name)) == -1) {
			sprintf(cp,"Could not gethostname()");
			return(TCL_ERROR);
		}
		struct hostent *hp;
		hp = gethostbyname(name);
		if (hp == NULL) {
			sprintf(cp,"Could not gethostbyname()");
			return(TCL_ERROR);
		}
		if (hp->h_addrtype != AF_INET)
			abort();		// can't happen
		memcpy(&addr.sin_addr.s_addr, hp->h_addr_list[0], hp->h_length);
		sid_in_ = rapi_session((sockaddr *)&addr, IPPROTO_UDP, 0, async, 0, &rerrno);
	} */
  fd_ = rapi_getfd(sid);
  session *sp;
  sp = new session(session_);
  sp->set_sid(sid);
  sp->set_dstaddr(dstaddr_);
  sp->set_dstport(dstport_);
  session_ = sp;
   
  int mask = TK_READABLE | TK_EXCEPTION;
  Tcl_CreateFileHandler((ClientData)fd_, mask, 
		       Tk_dispatch_read, (ClientData)fd_);
  sprintf(cp, "");
  return(TCL_OK);
}

/* sends rapi_sende in order to send PATH packets */
int
Rsvp::do_sender(char *cp, const char*const* argv, int sid_) {
  struct sockaddr_in srcaddr;
  memset(&srcaddr, 0, sizeof(srcaddr));
  char specbuf[4096];
  
  rapi_tspec_t* tspec = (rapi_tspec_t *)specbuf;
  session* s = get_session(sid_);
  if (s->localaddr() != 0) {
    srcaddr.sin_family = AF_INET;
    srcaddr.sin_addr.s_addr = s->localaddr();
    srcaddr_ = srcaddr.sin_addr.s_addr;
    srcaddr.sin_port = s->get_dstport();
    srcport_ = htons(srcaddr.sin_port);
  } else {
    srcaddr.sin_family = INADDR_ANY;
  }
  tspec->form = RAPI_TSPECTYPE_Simplified;
  tspec->len = sizeof(rapi_flowspec_t);
  tspec->tspecbody_qosx = s->get_Tspec();
  ttl_ = 16;
  int code = rapi_sender(sid_, 0, (sockaddr *)&srcaddr, 0, tspec, 0, 0, ttl_);
  if (code != 0) {
if(debug) {
    printf("error in rapi_sender()\n");
}
    return(TCL_ERROR);
  }
  sending_ = 1;
  return(TCL_OK);
}

/* gets the sid of the session on the base of address and port number */
int Rsvp::get_sid(int addr_, unsigned short port_) {
  session* sp=NULL;
  int sid=-1;
  sp = session_;
  if (sp == NULL ) return sid;
  while(((sp->get_dstaddr()!= addr_)||(sp->get_dstport()!=htons(port_)))&&(sp!=NULL)){
    if (sp->next() != NULL) {
      sp = sp->next();
    } else {
      sp = NULL;
    }
  }
  if (sp != NULL) {
    sid = sp->get_sid();
  } else {
    sid = -1;
  }
  return(sid); 
}

/* gets the session on the base of address and port number */
session* Rsvp::get_session(int addr_, unsigned short port_) {
  session* sp;
  int sid;
  sp = session_;
  /*  while( (   (sp->get_dstaddr()!=addr_) || (sp->get_dstport()!= htons(port_))  )&&(sp!=NULL)){
    if (sp->next() != NULL) {
      sp = sp->next();
    } else {
      sp = NULL;
    }
  }
  
  if (sp != NULL) {
    sid = sp->get_sid();
  } else {
    sid = -1;
  }
  return(sp); 
  */


  while(   (sp!=NULL) &&  ( (sp->get_dstaddr()!=addr_) || (sp->get_dstport()!= htons(port_)) )    ){
    if ((sp->next() != NULL) && (sp!=NULL)) {
      sp = sp->next();
    } else {
      sp = NULL;
    }
  }

  if (sp != NULL) {
    sid = sp->get_sid();
  } else {
    sid = -1;
  }
  return(sp);            

}

/* gets the sid of the session on the base of the sid number */
session* Rsvp::get_session(int sid) {
  session* sp=NULL;
  
  sp = session_;
  while((sp != NULL)&&(sp->get_sid() != sid)) {
    if (sp->next() != NULL) {
      sp = sp->next();
    } else {
      sp = NULL;
    }
  }
  return(sp); 
}

/* deletes the particular session */
void Rsvp::delete_session(int sid_) {
  session *sp, *sp_previous;
  
  sp = session_;
  sp_previous = session_;
  while((sp->get_sid() != sid_)&&(sp->next() != NULL)) {
    sp_previous = sp;
    sp = sp->next();
  }
  if (sp == session_) {
    if (sp->next() == NULL) {
      session_ = NULL;
    } 
    else {
      session_ = sp->next();
    }
  }		
  sp_previous->next(sp->next());
  delete sp;
}

/* deletes particular source */
void session::delete_source(source* src) {
  source* sp, *sp_previous;
  
  sp = get_sources_();
  sp_previous = get_sources_();
  while((sp != src)&&(sp->next() != NULL)) {
    sp_previous = sp;
    sp = sp->next();
  }
  if (sp == get_sources_()) {
    if (sp->next() == NULL) {
      set_sources_(NULL);
    } else {
      set_sources_(sp->next());
    }
  }		
  sp_previous->next(sp->next());
  delete sp;
}

/* establishes the source on the base of address */
source* session::get_source(int addr_) {
  source* src=NULL;
  
  if(addr_==0) return NULL;
  for (src = get_sources_(); src != NULL; src = src->next()) {
    if (src->eq(addr_)) {
      return(src);
    }	
  }
  return(NULL);
}
