/***********************************************************************************
 *
 * procedures which handle invitation and location lists 
 *  
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/



#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include "debug.h"
#include "defines.h"
#include "list.h"



extern inv_list_t *invitation_list;



/* inv_list_search **********************************************************************
 *
 *  Find pointer to a list entry in the invitation list
 *
 *  in:
 *  - search_str: call_id of searched entry
 *
 *  returns:
 *  - pointer to the found invitation_list entry
 *  - NULL, if no entry specified by search_str was found in the list
 *
 ***********************************************************************************/
inv_list_t *inv_list_search (char *search_str)
{
  static char     func[] = "inv_list_search";
  inv_list_t      *l;


  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)

  l = invitation_list;
  while (l != NULL) {
    if (!strcmp(l->call_id, search_str)) 
      return l;
    l = l->next;
  }
  return NULL;
} /* inv_list_search */
     

     
/* loc_list_search ******************************************************************
 *
 *  Find pointer to a special location_list entry in invitation_list
 *
 *  in:
 *  - inv: pointer to invitation_list entry in which location_list entry should be 
 *         searched
 *  - search_str: address of searched entry
 *
 *  returns:
 *  - pointer to found location_list entry
 *  - NULL, if no entry specified by search_str was found in the list
 *
 ***********************************************************************************/
loc_list_t *loc_list_search (inv_list_t *inv, char *search_str)
{
  static char     func[] = "loc_list_search";
  loc_list_t      *l;


  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)

  l = inv->location_list;
  while (l != NULL) {
    if (!strcmp(l->addr, search_str))
      return l;
    l = l->next;
  }
  return NULL;
} /* loc_list_search */



/* inv_list_add *********************************************************************
 *
 *  Add a new entry to invitation_list
 *
 *  returns:
 *  - 0:  entry was added to list
 *  - -1: adding to list failed
 *
 ***********************************************************************************/
int inv_list_add ()
{
  static char     func[] = "inv_list_add";
  inv_list_t      *l;


  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)
  
  if ((l = (inv_list_t *)malloc(sizeof(inv_list_t))) == NULL) 
    return (-1);
    
  l->next = invitation_list; /* put new entry at head of list */
  invitation_list = l;

  strcpy(invitation_list->call_id, "");
  strcpy(invitation_list->user, "");
  invitation_list->userEntry = NULL;
  invitation_list->socket_to_client = 0;
  strcpy(invitation_list->request, "");
  strcpy(invitation_list->response, "");
  invitation_list->response_priority = 0;
  strcpy(invitation_list->addr_of_client, "");
  invitation_list->socket_to_client = 0;
  invitation_list->port_to_client = 0;
  invitation_list->transport_to_client = 0;
  invitation_list->socket_to_server = 0;
  invitation_list->transport_to_server = 0;
  invitation_list->state = 0;
  invitation_list->locations = 0;
  invitation_list->location_list = NULL;
  
  return 0;
} /* inv_list_add */



/* loc_list_add ********************************************************************
 *
 *  Add a new entry to location_list in invitation_list entry
 *
 *  in:
 *  - inv: pointer to invitation_list entry in which a location entry should be 
 *         added
 *
 *  returns:
 *  - 0:  entry was added to list
 *  - -1: adding to list failed
 *
 ***********************************************************************************/
int loc_list_add (inv_list_t * inv)
{
  static char     func[] = "inv_list_add";
  loc_list_t      *l;


  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)
  
  if ((l = (loc_list_t *)malloc(sizeof(loc_list_t))) == NULL) 
    return (-1);
    
  l->next = inv->location_list; /* put new entry at head of list */
  inv->location_list = l;
  
  strcpy(inv->location_list->addr, " ");
  strcpy(inv->location_list->hostname, " ");
  strcpy(inv->location_list->user, " ");
  inv->location_list->retransmissions = 0;
  inv->location_list->response_received = FALSE;
  inv->location_list->inv_parent = inv;
       
  return 0;
} /* inv_list_add */



/* inv_list_remove *****************************************************************
 *
 *  Remove an entry from invitation_list
 *  
 *  in:
 *  - str: call_id of entry which sould be removed
 *  
 ***********************************************************************************/
void inv_list_remove (char *str)
{
  static char     func[] = "inv_list_remove";
  inv_list_t      *l, *prev;
  

  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)
  
  l = invitation_list;
  prev = NULL;
  while (l != NULL) {
    if (!strcmp(l->call_id,str)) {
      if (prev==NULL) 
        invitation_list = l->next;
      else 
        prev->next = l->next;
      free(l);
    }
    prev = l;
    l = l->next;
  }
  return;
} /* inv_list_remove */
