/***********************************************************************************
 *
 * functions to handle new incoming requests
 * - location of users
 * - check of call acceptance
 *
 * partially based on cd by Christian Zahl
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/



#include <ctype.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <tcl.h>
#include <unistd.h>
#include "debug.h"
#include "defines.h"
#include "list.h"
#include "misc.h"
#include "reply_codes.h"
#include "servers.h"

#include "user_locate.c"
#include "auto_reply.c"
#include "expand.c"

extern struct in_addr   host2ip(char *host);
extern char             ls_db[255];
extern char             alias_db[255];
extern char             sipd_bin_path[255];
extern char             LocationService[255];
extern char             ExpandService[255];
extern char             ForwardService[255];
extern char             myAddr[17];
extern Tcl_Interp        *interp;



/* locate_user **********************************************************************
 *
 *  Locate invited user in the local domain
 *  
 *  - creates child process which executes locate script with parameter user-id
 *     (called script sends found locations (IP addresses) to stdout)
 *  - creates list of locations in inv->location_list
 *
 *  in:
 *  - inv: pointer to entry in invitation_list
 *  - user-id: local user name
 *  
 *  returns:
 *  - number of found locations
 *  - -1 on failure 
 *
 ***********************************************************************************/
int locate_user (inv_list_t *inv)
{
  static char     func[] = "locate_user";
  FILE            *f;
  int             fd[2];
  pid_t           pid;
  int             child_status;
  char            addr[17];
  char            host[50];
  char            *p;
  char            *via_end; /* pointer to the end of via headers */
  

  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)

  inv->locations = 0;
  inv->location_list = NULL;
 
  /***********************************************************************************
   * get pointer to the end of Via headers to check if the request was already sent 
   * to one of the returned locations 
   ***********************************************************************************/
  for (via_end = strchr(inv->request, '\n'); tolower(*(++via_end)) == 'v';  via_end=strchr(via_end, '\n')); 

  if (pipe (fd) == -1) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: pipe",func);)
    return -1;
  }

  /* execute the script in the child */
  signal (SIGCHLD, SIG_DFL);
  pid = fork ();
  if (pid == -1) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: fork",func);)
    close (fd[0]);
    close (fd[1]);
    return -1;
  }

  /***********************************************************************************
   * child process, which executes the user_locate script 
   ***********************************************************************************/
  if (pid == 0) {
    dup2 (fd[1], 1);
    close (fd[0]);
    close (fd[1]);
    signal (SIGTERM, SIG_DFL);
    signal (SIGINT, SIG_DFL);
    signal (SIGALRM, SIG_DFL);

    if (*LocationService == NULL) {
      Tcl_SetVar(interp, "user", inv->userEntry->pw_name, 0);
      Tcl_SetVar(interp, "lswhod_db", ls_db, 0);
      exit(Tcl_Eval(interp, user_locate_tcl));
    }
    else {
      execle (LocationService, LocationService, inv->userEntry->pw_name, NULL, NULL);

    }
  }

  /***********************************************************************************
   * parent process, which reads hosts  returned from the user_locate script 
   ***********************************************************************************/
  close(fd[1]);
  if ((f = fdopen (fd[0], "r")) == NULL) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: fdopen",func);)
    kill (pid, SIGKILL);
    waitpid (pid, NULL, 0);
    close (fd[0]);
    return -1;
  }

  /* read host which is returned from the  script via stdout */
  while (fgets (host, sizeof (host), f)) {
    *strchr(host, '\n') = '\0';
    DEBUG (2, syslog (LOG_DEBUG, "%s, loc_host: %s", func, host);)
   /* get IP addr */
    strcpy(addr, inet_ntoa(host2ip(host))); 
      
    /* check if addr is not already in Via header */
    if ((p = str_search(inv->request, "\nv")) != NULL)
      if ((p = strstr(p, addr)) != NULL)
        if (p < via_end)
          continue;
   
    /* add addr to location_list */
    if (loc_list_add(inv)) {
      DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: malloc",func);)
      fclose (f);
      close (fd[0]);
      kill (pid, SIGKILL);     /* make sure don't haveing zombies */
      waitpid (pid, &child_status, 0);
      return -1;
    }

    /* fill location list entry */
    strcpy(inv->location_list->addr, addr);
    strcpy(inv->location_list->hostname, host);
    strcpy(inv->location_list->user, inv->userEntry->pw_name);
    inv->location_list->inv_parent=inv;
    inv->locations++;    /* increase counter for locations */
  } /* while */

  
  fclose (f);
  close (fd[0]);
  kill (pid, SIGKILL);     /* make sure don't haveing zombies */
  waitpid (pid, &child_status, 0);
  DEBUG (2, syslog (LOG_DEBUG, "%s: child exited with status %x", func, child_status);)
    
  /* child killed? */
  if (WIFSIGNALED (child_status)) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: child terminated by signal %d", func, WTERMSIG (child_status));)
    return -1; 
  }
  
  /* user not found? */
  if (((WEXITSTATUS (child_status)) == 1) || (inv->locations <= 0)) {
    return 0;
  }
  
  /* error ? */
  if ((WEXITSTATUS (child_status)) != 0) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: unexpected retcode (%d) from child", func, WEXITSTATUS (child_status));)
    return -1;
  }

  /* return number of found locations */
  return inv->locations;
} /* locate_user */



/* automatic_reply ******************************************************************
 *
 *  checks if the invited user wants automatically react to the incoming call 
 *  depending on the invitation's from and information field
 *  - creates child process which executes auto_reply script with two arguments:
 *     entry of "from" and  first information field of the request
 *  - reads returned string which consists of a reply code and a string
 *  - sends reply with regard of the returned reply code and string
 *     
 *  in:
 *  -inv: pointer to entry in invitation_list
 *  
 *  returns:
 *  - 0: automatic reply was sent
 *  - 1: no automatic reply
 *
 ***********************************************************************************/
int automatic_reply (inv_list_t *inv)
{
  static char     func[] = "automatic_reply";
  char            buf[4*1024];
  char            *envv[10]; /* environment to start the auto_reply script */
  int             i;
  int             fd[2];
  FILE            *f;
  pid_t           pid;
  int             child_status;
  char            from[MAX_USER_NAME_LENGTH]; /* from field of SIP header */
  char            subject[MAX_LINE_LENGTH]; /* subject string of SDP payload */
  char            reply_string[MAX_LINE_LENGTH]; /* string returned from the script */
  char            reply_code_str[5]; /* reply code returned from the script */
  int             reply_code;
  char            *fwd_loc; /* SIP location if froward reply */ 
  char            *user; /* user of fwd_loc */
  char            addr[17]; /* IP addr of fwd_loc */
  char            *hostname; /* host (domain) name of fwd_loc */
  int             new_loc_flag = FALSE;
  char            *via_end; /* end of via headers in request */
  char            *p,*c;
  server_t        *s;


  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)
    
  /* get from field of the request */
  if ((p = search_SIP_header(inv->request, "\nFrom:", "\nf:")) != NULL) {
    if ((p = get_field (p, from, sizeof(from)-1)) == NULL) 
      strcpy(from, "-");
  }
  else
    strcpy(from, "-");
  
  /* get first subject field (i=) of the SDP payload */
  if ((p = strstr(inv->request, "\ni=")) != NULL) { 
    p += strlen("\ni=");
  if ((p = get_line (p, subject, sizeof(subject)-1)) == NULL) 
    strcpy(subject, "-");
  }
  else
    strcpy(subject, "-");

  if (pipe (fd) == -1) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: pipe",func);)
    return 1;
  }

  /* execute the script in the child */
  signal (SIGCHLD, SIG_DFL);
  pid = fork ();
  if (pid == -1) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: fork",func);)
    close (fd[0]);
    close (fd[1]);
    return 1;
  }

  /***********************************************************************************
   * child process, which executes the auto_reply script 
   ***********************************************************************************/
  if (pid == 0) {
    dup2 (fd[1], 1);
    close (fd[0]);
    close (fd[1]);
    signal (SIGTERM, SIG_DFL);
    signal (SIGINT, SIG_DFL);
    signal (SIGALRM, SIG_DFL);

    if (*ForwardService == NULL) {
      Tcl_SetVar(interp, "from", from, 0);
      Tcl_SetVar(interp, "subject", subject, 0);
      Tcl_SetVar(interp, "HOME", inv->userEntry->pw_dir, 0);
      exit(Tcl_Eval(interp, auto_reply_tcl));
    }
    else {
      /* setup the environment for the child to start auto_reply script*/
      i = 0;
      envv[i++] = p = buf;
      sprintf (p, "HOME=%s", inv->userEntry->pw_dir);
      envv[i++] = p += strlen (p) +1;
      sprintf (p, "USER=%s", inv->userEntry->pw_name);
      envv[i++] = NULL;
      
      execle (ForwardService, ForwardService, from, subject, NULL, envv);
      exit (1);
    }

  }

  /***********************************************************************************
   * parent process, which reads reply_string  returned from the auto_reply script 
   ***********************************************************************************/
  close(fd[1]);
  if ((f = fdopen (fd[0], "r")) == NULL) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: fdopen",func);)
    kill (pid, SIGKILL);
    waitpid (pid, NULL, 0);
    close (fd[0]);
    return 1;
  }

  /* read reply_string  which is returned from the  script via stdout */
  while (fgets (reply_string, sizeof (reply_string), f)) {
    *strchr(reply_string, '\n') = '\0';
    DEBUG (2, syslog (LOG_DEBUG, "%s, repl_str: %s", func, reply_string);)
  }
  fclose (f);
  close (fd[0]);
  kill (pid, SIGKILL);     /* make sure don't haveing zombies */
  waitpid (pid, &child_status, 0);
  DEBUG (2, syslog (LOG_DEBUG, "%s: child exited with status %x", func, child_status);)
    
  if (WIFSIGNALED (child_status)) 
    return 1;
   
  if ((WEXITSTATUS (child_status)) != 0) 
    return 1;
    
  /***********************************************************************************
   * react to returned reply_string
   * if the returned reply code specifies a redirect response the additional string is 
   * interpreted as the new forwarding locations, otherwise the string specifies the 
   * response reason phrase
   ***********************************************************************************/
  p = reply_string;
  /* get the reply code */
  if ((p = get_field (p, reply_code_str, sizeof(reply_code_str)-1)) == NULL)
    return 1;
  
  reply_code = atoi(reply_code_str);

  if ((reply_code < 300) || (reply_code >= 400)) { 
    /***********************************************************************************
     * the reply code doesn't specify a redirect response 
     * send the reply to the client 
     ***********************************************************************************/
    reply(inv, reply_code, skip_white_spaces(p));
    return 0; 
  } 

  /***********************************************************************************
   * redirect response 
   * - spilt the returned response string into several forwarding locations 
   * - check if the request was already sent to one of these locations (examine via 
   *   headers)
   ***********************************************************************************/
  for (fwd_loc = strtok(p, " "); fwd_loc != NULL; fwd_loc = strtok(NULL, " ")) {
    if ((c = strchr(fwd_loc, '@')) == NULL) { 
      user = inv->userEntry->pw_name; 
      hostname = fwd_loc; 
    } 
    else { 
      *c = '\0'; 
      user = fwd_loc;
      hostname = ++c;
    } 
      
    s = servers(hostname);
    strcpy (addr, inet_ntoa(s->addr));
    
    if (!strcmp(addr, myAddr) && !strcmp(user, inv->userEntry->pw_name)) 
      continue;
        
    /***********************************************************************************
     * get pointer to the end of Via headers to check if the request was already sent 
     * to one of the returned locations 
     ***********************************************************************************/
    for (via_end = strchr(inv->request, '\n'); tolower(*(++via_end)) == 'v';  via_end = strchr(via_end, '\n')); 
    
    /***********************************************************************************
     * append fwd_loc to location_list if:
     *  - addr not in via headers
     *  - user != pw_name
     ***********************************************************************************/
    if ((p = strstr(str_search(inv->request, "\nv"), addr)) != NULL)
      if ((p < via_end) && !(strcmp(user, inv->userEntry->pw_name))) 
        continue;
    
    new_loc_flag = TRUE;
    /* create new location_list entry */
    if (loc_list_add(inv)) { 
      DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: malloc",func);) 
      return 1; 
    } 
    inv->location_list->inv_parent=inv;  
    strcpy(inv->location_list->user, user); 
    strcpy(inv->location_list->addr, addr); 
    strcpy(inv->location_list->hostname, hostname);
  } /* for */
  
  if (new_loc_flag) {
    /***********************************************************************************
     * send the redirect response to the client 
     ***********************************************************************************/
    if (reply_code_str == "301")  
      reply(inv, atoi(reply_code_str), " Permanent Froward"); 
    else  
      reply(inv, atoi(reply_code_str), "Temporary Forward"); 
    return 0; 
  }
  return 1;
} /* automatic_reply */



/* expand_user_name ****************************************************************
 *
 *  expands the given user name into the real user-id
 *  - creates child process which executes expand script with the user name as 
 *    argument
 *  - reads returned user-ids
 *  - sends appropriate reply if user name is not known or is ambigious in the 
 *    domain 
 *     (protection of data privacy)
 *
 *  in: 
 *  - inv: pointer to entry in invitation_list
 *  
 *  returns:
 *  - name: expanded user-id
 *  - NULL if error or reply was sent
 *
 ***********************************************************************************/
char *expand_user_name (inv_list_t *inv)
{
  static char     func[] = "expand_user_name";
  FILE            *f;
  int             fd[2];
  pid_t           pid;  
  int             child_status;
  loc_list_t      *tmp;
  loc_list_t      *list = NULL; /* list of returned user-ids */
  char            name[MAX_USER_NAME_LENGTH]; 
  static char     expanded_name[MAX_USER_NAME_LENGTH]; /* user-ids returned from the script */
  int             n; /* counter for user-ids returned from the script */ 
  char            *p;
  char            c;
  

  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)

  /* get user name which should be expanded to user-id */
  strcpy (name, inv->user);
  if (p = strchr (name, '@'))
    *p = '\0';

  n = 0;

  if (pipe (fd) == -1) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: pipe",func);)
    return NULL;
  }

  /* execute the script in the child */
  signal (SIGCHLD, SIG_DFL);
  pid = fork ();
  if (pid == -1) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: fork",func);)
    close (fd[0]);
    close (fd[1]);
    return NULL;
  }

  /***********************************************************************************
   * child process, which executes the expand script 
   ***********************************************************************************/
  if (pid == 0) {
    dup2 (fd[1], 1);
    close (fd[0]);
    close (fd[1]);
    signal (SIGTERM, SIG_DFL);
    signal (SIGINT, SIG_DFL);
    signal (SIGALRM, SIG_DFL);

    if (*ExpandService == NULL) {
      Tcl_SetVar(interp, "user", name, 0);
      Tcl_SetVar(interp, "sipd_bin", sipd_bin_path, 0);
      Tcl_SetVar(interp, "alias_data_base", alias_db, 0);
      exit(Tcl_Eval(interp, expand_tcl));
    }
    else {
      execle (ExpandService, ExpandService, name, NULL, NULL);
      exit (1);
    }
  }

  /***********************************************************************************
   * parent process, which reads names returned from the expand script 
   ***********************************************************************************/
  close(fd[1]);
  if ((f = fdopen (fd[0], "r")) == NULL) {
    DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: fdopen",func);)
    kill (pid, SIGKILL);
    waitpid (pid, NULL, 0);
    close (fd[0]);
    return NULL;
  }

  while (fgets (expanded_name, sizeof(expanded_name), f)) {
    *strchr(expanded_name, '\n') = '\0';
 
    DEBUG (3, syslog (LOG_DEBUG, "%s: exp_name: %s", func, expanded_name);)

    if ((tmp = malloc(sizeof(loc_list_t))) == NULL) {
      reply (inv, SERVER_INTERNAL_ERROR, "Server Internal Error");
      LIST_DEL(list);
      fclose (f);
      close (fd[0]);
      kill (pid, SIGKILL);     /* make sure don't haveing zombies */
      waitpid (pid, &child_status, 0);
      return NULL;
    }
    tmp->next = list;
    strcpy(tmp->user, expanded_name);
    list = tmp;
    ++n;
  } /* while */
 
  fclose (f);
  close (fd[0]);
  kill (pid, SIGKILL);     /* make sure don't haveing zombies */
  waitpid (pid, &child_status, 0);
  DEBUG (2, syslog (LOG_DEBUG, "%s: child exited with status %x", func, child_status);)
    
  if (WIFSIGNALED (child_status)) {
    reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error");
    DEBUG (2, syslog (LOG_DEBUG, "%s: unexpected retcode (%d) from child", func, WEXITSTATUS (child_status));)
    LIST_DEL(list);
    return NULL;
  }

  if ((WEXITSTATUS (child_status)) != 0) {
    reply(inv, SERVER_INTERNAL_ERROR, "Server Internal Error");
    DEBUG (2, syslog (LOG_DEBUG, "%s: unexpected retcode (%d) from child", func, WEXITSTATUS (child_status));)
    LIST_DEL(list);
    return NULL;
  }

  /* name unknown? */
  if (n <= 0) {
    reply(inv, NOT_FOUND, "Not Found");
    DEBUG (2, syslog (LOG_DEBUG, "%s: NOT_FOUND", func);)
    LIST_DEL(list);
    return NULL;
  }

  /* expands to more then one */
  if (n > 1) {
    reply(inv, NOT_KNOWN_HERE, "Ambiguous");
    DEBUG (2, syslog (LOG_DEBUG, "%s: AMBIGUOUS: expands more than one", func);)
    LIST_DEL(list);
    return NULL;
    }

  LIST_DEL (list);
  return expanded_name;
} /* expand_user_name */



/* process_request *****************************************************************
 *
 *  handles new incoming requests
 *  - calls function to get the user-id of the called user
 *  - calls function to check call acceptance (automatic reply)
 *  - calls function to locate the invited user
 *
 *  in:
 *  - inv: pointer to entry in invitation_list
 *  
 ***********************************************************************************/
void process_request (inv_list_t *inv)
{
  static char     func[] = "process_request";
  char            *expanded_name; /* string returned by expand_user_name */
  char            user_login[MAX_USER_NAME_LENGTH]; /* user login */
  char            addr[17];
  char            host[50] = " ";
  char            *p;
  struct passwd   *userEntry,*dummyEntry;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  /***********************************************************************************
   * expand the user name, and check if the name is known 
   ***********************************************************************************/
  if ((expanded_name = expand_user_name (inv)) == NULL)  
    return;  

  /* check whether expanding done by alias or namemapper routine */
  if ((p = strchr (expanded_name, ' ')) != NULL) { 
    /* namemapper routine */
    get_field(p, user_login, sizeof(user_login)-1); 
  } 
  else { 
    strcpy (user_login, expanded_name); 
  } 

  /* check if invited user was expanded to user@host */
  if ((p = strchr (user_login, '@')) != NULL) { 
    *p = '\0';  
    strcpy (host, ++p); 
  }
  
  if ((userEntry = (struct passwd *)malloc(sizeof(struct passwd))) == NULL)  
    return; 
  /*  if (!(userEntry = getpwnam (user_login))) { 
   *  free(userEntry); 
   * 
   * not mt-safe. getpwnam uses a static array...
   * FT123-Team
   */
  if (!(dummyEntry = getpwnam(user_login)))
    {
      free(userEntry);
      reply(inv, NOT_FOUND, "Not Found"); 
      DEBUG (2, syslog (LOG_DEBUG, "%s: NOT_FOUND", func);) 
	return; 
    } 
  
  userEntry->pw_uid=dummyEntry->pw_uid;
  userEntry->pw_gid=dummyEntry->pw_gid;

  if (!(userEntry->pw_name   = strdup(dummyEntry->pw_name)))
    return;
  if (!(userEntry->pw_passwd = strdup(dummyEntry->pw_passwd)))
    return;
  if (!(userEntry->pw_gecos  = strdup(dummyEntry->pw_gecos)))
    return;
  if (!(userEntry->pw_dir    = strdup(dummyEntry->pw_dir)))
    return;
  if (!(userEntry->pw_shell  = strdup(dummyEntry->pw_shell)))
    return;

  inv->userEntry = userEntry;
    
  if (strcmp(host, " ")) {
    /* get IP addr */
    strcpy(addr, inet_ntoa(host2ip(host))); 
    
    /* add addr to location_list */
    if (loc_list_add(inv)) { 
      reply (inv, SERVER_INTERNAL_ERROR, "Server Internal Error"); 
      DEBUG (2, syslog (LOG_DEBUG, "%s: SERVER_INTERNAL_ERROR: malloc",func);) 
      return; 
    } 
    
    /* fill location list entry */
    strcpy(inv->location_list->addr, addr); 
    strcpy(inv->location_list->hostname, host); 
    strcpy(inv->location_list->user, user_login); 
    inv->location_list->inv_parent=inv; 
    inv->locations++;     /* increase counter for locations */
    
    forward_request(inv->location_list); 
    return; 
  } 
  
  /***********************************************************************************
   * check call acceptance 
   ***********************************************************************************/
  if (!automatic_reply(inv)) 
    return;

  /***********************************************************************************
   * locate the user 
   ***********************************************************************************/
  switch (locate_user(inv)) {
    case -1:
      /***********************************************************************************
       * error while locating the user 
       ***********************************************************************************/
      inv->locations = 0;
      reply (inv, SERVER_INTERNAL_ERROR, "Server Internal Error");
      return;
    case 0:
      /***********************************************************************************
       * user was currently not found 
       ***********************************************************************************/
      reply (inv, NOT_CURRENTLY_HERE, "Not Currently Here");
      return;
    case 1:
      /***********************************************************************************
       * exactly one location of the user was found
       * forward request to that location 
       ***********************************************************************************/
      forward_request(inv->location_list);
      return;
    default:
      /***********************************************************************************
       * more than one location of the user was found
       * function proxy will handle the request 
       ***********************************************************************************/
      proxy(inv);
      return;
  }   
} /* process_request */







