/***********************************************************************************
 *
 * functions to 
 * - create replies to the client
 * - handle responses received from the next server
 * - send replies to the client
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "debug.h"
#include "defines.h"
#include "list.h"
#include "misc.h"
#include "notify.h"
#include "multimer.h"
#include "proxy.h"
#include "reply_codes.h"
#include "response.h"
#include "servers.h"
#include "timer.h"
#include "udp.h"



extern char           myAddr[17];
extern inv_list_t     *invitation_list;
extern int            udp_server_socket;



/* get_response_priority *************************************************************
 *
 *  get priority of a response 
 *  a priority depending on the response code is returned to classify the response
 *
 *  in:  
 *  - code: response code
 *
 *  returns:
 *  - priority 
 *
 ***********************************************************************************/
int get_response_priority (int code)
{
  static char     func[] = "get_respnonse_priority";


  DEBUG (4, syslog (LOG_DEBUG, "%s",func);)

  /* final responses, highest priority */
  if ((code >= 200 && code < 500)) 
    return 50;

  switch (code) {
    case SERVICE_UNAVAILABLE:
      return 33;
    case SERVER_INTERNAL_ERROR:
      return 32;
    case NOT_IMPLEMENTED:
      return 31;
    case SEARCH_FAILURE:
      return 23; 
    case NOT_CURRENTLY_HERE:
      return 22;
    case NOT_KNOWN_HERE:
      return 21;
    case RINGING:
      return 12;
    case TRYING:
      return 11;
  }
  
  if ((code >= 500 && code < 600)) 
    return 30;
  if ((code >= 600 && code < 700)) 
    return 20;
  if ((code >= 100 && code < 200)) 
    return 10;

  return 0;
} /* get_response_priority */



/***********************************************************************************
 * 
 *  generates replies which should be send to the client
 *
 *  in:
 *  - inv: pointer to entry in invitation_list
 *  - code: response code
 *  - phrase: reason phrase
 *
 ***********************************************************************************/
void reply (inv_list_t *inv, int code, char *phrase)
{
  static char     func[] = "reply";
  char            new_line[MAX_LINE_LENGTH];
  char            *p;
  loc_list_t      *loc;


  DEBUG (2, syslog (LOG_DEBUG, "%s, code: %d, phrase: %s",func, code, phrase);)

  /***********************************************************************************
   * create response packet 
   ***********************************************************************************/
  sprintf(inv->response, "%s %d %s\r\n", SIP_VERSION, code, phrase);
    
  /* append request header without first line */
  p = strchr(inv->request, '\n');
  strncat(inv->response, ++p, strstr(inv->request, "\r\n\r\n")+4 -p);
 
  /* remove lines which are not needed */
  if ((p = str_search(inv->response, "\nContent-type:")) != NULL)
    delete_SIP_line(++p);
  else if ((p = str_search(inv->response, "\nc:")) != NULL)
    delete_SIP_line(++p);
  if ((p = str_search(inv->response, "\nContent-length:")) != NULL)
    delete_SIP_line(++p);
  if ((p = str_search(inv->response, "\nl:")) != NULL)
    delete_SIP_line(++p);
    
  /* add Contact-host and Content-length line */
  sprintf(new_line, "h:%s", myAddr); 
  p = strstr(inv->response, "\r\n\r\n")+2;
  insert_SIP_line(p, new_line);
  insert_SIP_line(p, "l:0");

  /*********************************************************************************** 
   * append locations if "Alternative Address" or redirect reply
   ***********************************************************************************/
  if ((code == ALTERNATIVE_ADDRESS) || ((code > 300) && (code < 400))) {
    /* append locations */
    for (loc=inv->location_list; loc != NULL; loc=loc->next) {
      sprintf(new_line, "m:sip://%s@%s", loc->user, loc->hostname);
      insert_SIP_line(p,new_line);
    }
  }

  /***********************************************************************************
   * call "check_response_priority" which will send the reply to the client
   ***********************************************************************************/ 
  check_response_priority(inv, inv->response, get_response_priority(code));
} /* reply */
  


/* check_response_priority **********************************************************
 *
 *  checks priority of outgoing and incoming replies and decides whether to send the
 *  response to the client or to wait for other responses
 *
 *  if a parallel search is performed a response will be send to the client if the 
 *  proxy received a response form all locations (responses with higher priority than
 *  informational responses) or if one location has sent a final response (code 2xx 
 *  to 4xx)
 *
 *  the response with highest priority which was received will be stored in 
 *  inv->response
 *
 *  in: 
 *  - inv: pointer to entry in invitation_list
 *  - msg: outgoing or incoming response
 *  - priority: response code priority
 *
 *
 ***********************************************************************************/
void check_response_priority (inv_list_t *inv, char *msg, int priority)
{
  static char     func[] = "check_response_priority";
  char            *p, *s;
  loc_list_t      *loc;
  struct timeval  time;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  /***********************************************************************************
   * check if new response has higher priority than the actual stored one
   ***********************************************************************************/
  if (priority >= inv->response_priority) {
    /* higher priority, store response */
    inv->response_priority = priority;
    if (msg != inv->response) {
      strcpy(inv->response, msg);
      /* delete via header with own address */
      p = str_search(inv->response, "\nv")+1;
      if ((s = strstr(p, myAddr)) != NULL) 
        if (s < strchr(p, '\n')) 
          delete_SIP_line(p);
    }

    /***********************************************************************************
     * send informational responses directly to the client
     ***********************************************************************************/
    if (inv->response_priority < 20) {
      if (inv->transport_to_client == SOCK_STREAM) 
        TCP_send(inv->socket_to_client, inv->response); 
      else 
        UDP_send(inv->addr_of_client, inv->port_to_client, udp_server_socket, 1, inv->response);  
      if (msg != inv->response)
        free(msg);
      return;
    } 
    else {
      /***********************************************************************************
       * when receiving a final response ignore responses from other locations and send 
       * this response to the client
       ***********************************************************************************/
      if (inv->response_priority >= 50) {
        /* final response */
        /* kill all T_retransmit timers */
        for (loc = inv->location_list; loc != NULL; loc = loc->next) 
          timer_set(NULL, (Notify_func)retransmit_timeout, (Notify_client)loc, 1);   
        timer_set(NULL, (Notify_func)request_timeout, (Notify_client)inv, 1);   

        /* remove counter for outstanding responses */
        inv->locations = 0;
      }
    }
  }

  if (msg != inv->response)
    free(msg);
    
  /***********************************************************************************
   * if there are no more outstanding responses, send reply with highest priority
   ***********************************************************************************/
  if (inv->locations <= 0) {
    /* responses from all locations or final response */
    if (inv->response_priority < 20) {
      /* only informational responses are received */
      reply(inv, NOT_CURRENTLY_HERE, "Not Currently Here");
    }
    else {
      if (inv->transport_to_client == SOCK_STREAM) 
        TCP_send(inv->socket_to_client, inv->response);  
      else
        UDP_send(inv->addr_of_client, inv->port_to_client, udp_server_socket, 1, inv->response); 
      inv->state = REPLY_SENT;
       
      /* start T_close_inv timer to remove entry from invitation list */
      time.tv_sec = 60; 
      time.tv_usec = 0;  
      timer_set(&time, (Notify_func)close_event_handler, (Notify_client)inv, 1);      
    }
  } 
} /* check_response_priority */



/* forward_to_alternative_addr  *****************************************************
 *
 *  forward request to alternative address
 *
 *  invoked if an "Alternetive Address" response was received
 *  checks if alternative address was already queried
 *  forwards request to the new location
 *
 *  in:
 *  -  inv: pointer to entry in invitation_list
 *  - l:    pointer to new location
 *  
 *  returns:
 *  - 0: request was forwarded to new location
 *  - 1: request was not forwarded to new location
 * 
 ***********************************************************************************/
int forward_to_alternative_addr (inv_list_t *inv, char *l) 
{
  static char     func[] = "check_location_header";
  char            new_loc[MAX_USER_NAME_LENGTH]; /* location from Alt.Addr. reply */
  char            new_addr[17]; /* addr from Alt. Addr. reply */
  char            *hostname;
  char            *p;
  server_t        *s;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  /***********************************************************************************
   * get new location and check if request was already sent to this address
   ***********************************************************************************/
  if ((p = get_field (strstr(l, "//")+2, new_loc, sizeof (new_loc))) == NULL) 
    return 1;
  
  /* get host part of location */
  p = strchr(new_loc, '@');
  hostname = ++p;
  
  s = servers(hostname);
  strcpy (new_addr, inet_ntoa(s->addr));
  *--p = '\0';
  
  /***********************************************************************************
   * request already forwarded to new_addr  ?
   ***********************************************************************************/
  if (loc_list_search(inv, new_addr) != NULL) 
    return 1;
  else {
    /* append addr to location_list and forward request to addr */
    loc_list_add(inv);
    strcpy(inv->location_list->addr, new_addr);
    strcpy(inv->location_list->hostname, hostname);
    strcpy(inv->location_list->user, new_loc);
    inv->location_list->inv_parent = inv;
    /* increase counter for outstanding responses */
    ++inv->locations;
    forward_request(inv->location_list);
    return 0;
  }
} /* forward_to_alternative_addr */



/* process_response ****************************************************************
 *
 *  handle incoming responses
 *  calls "check_response_priority" which sends responses to the client
 *  
 *  in:  
 *  - msg: incoming response
 *  - addr: address the reply was received from
 *
 ***********************************************************************************/
int process_response (char *msg, char *addr)
{
  static char     func[] = "process_response";
  char            resp_code_str[5];
  int             resp_code;
  int             priority;
  char            call_id[MAX_CALL_ID_LENGTH];
  int             locs; /* counter for alternative locations */
  inv_list_t      *inv;
  loc_list_t      *loc;
  char            *p,*s;


  DEBUG (2, syslog (LOG_DEBUG, "%s, msg: %s \naddr %s",func, msg, addr);)

  /***********************************************************************************
   * get call-id and search appropriate invitation list entry
   ***********************************************************************************/
  if ((p = search_SIP_header(msg, "\nCall-id:","\ni:")) != NULL) {
    if ((p = get_field (p, call_id, sizeof(call_id)-1)) == NULL) {
      free(msg);
      return -1;
    }
  } else {
    free(msg);
    return -1;
  }
  
  /* search for list entry with call-id */
  if ((inv = inv_list_search(call_id)) == NULL) {
    /* there was no request with this call-id */
    free(msg);
    return -1;
  }
  
  if (inv->transport_to_server != SOCK_STREAM) {
    if ((loc = loc_list_search(inv, addr)) == NULL) {
      free(msg);
      return -1;
    }
  
    if (loc->response_received) {
      /* a definite response from this location was already received */
      free(msg);
      return -1;
    }
  }

  
  if (inv->state == REPLY_SENT) {
    /* a response was already sent to the client */
    free(msg);
    return -1;
  }
  
  /***********************************************************************************
   * extract response code
   ***********************************************************************************/
  p = msg + strlen(SIP_VERSION);
  if ((p = get_field (p, resp_code_str, sizeof(resp_code_str) )) == NULL) { 
    free(msg);
    return -1;
  }  
  resp_code = atoi(resp_code_str);
  priority = get_response_priority(resp_code);
  
  /*********************************************************************************** 
   * check the response code
   ***********************************************************************************/
  if (priority >= 20) {
    /***********************************************************************************
     * response was a definite response
     * decrease counter of outstanding responses and kill T_request or T_retransmit timer
     * set response_received flag to avoid decreasing the locations counter if this 
     * response packet was duplicated
     ***********************************************************************************/ 
    --inv->locations;
    if (inv->transport_to_server == SOCK_STREAM)
      timer_set(NULL, (Notify_func)request_timeout, (Notify_client)inv, 1);
    else {
      loc->response_received = TRUE;
      timer_set(NULL, (Notify_func)retransmit_timeout, (Notify_client)loc, 1); 
    }
  }

  if (resp_code == ALTERNATIVE_ADDRESS) {
    /***********************************************************************************
     * an "Alternative Address" response was received
     * get number of alternative locations and expand search to this locations or send a 
     * response to the client
     ***********************************************************************************/
    /* get number of given locations */
    locs = 0;
      
    p=msg;
    /* search for a "Location:" lines */ 
    while ((p = str_search(p, "\nLocation:")) != NULL) { 
      ++locs;
      ++p;
    }
      
    /* search for a "m:" lines */ 
    p=msg;
    while ((p = str_search(p, "\nm:")) != NULL) {
      /* make sure that "m:" line was in the SIP header */
      if  (p < strstr(msg, "\r\n\r\n"))
        break;
      ++locs;
      ++p;
    }

    if ((locs > 1) && (get_via_header_no(inv->request) > 1)) {
      /***********************************************************************************
       * multiple locations are given in the response but this is not the first proxy
       * send the response to the client
       ***********************************************************************************/
      check_response_priority(inv, msg ,priority);
      return;
    }
      
    /***********************************************************************************
     * if only one new location is given in the response query this one,
     * if multiple locations are given in the response, perform a parallel search 
     * because this is the first proxy
     ***********************************************************************************/
    if (locs > 1 )
      inv->state = PARALLEL_SEARCH;
      
    /* expand search to new locations */
    /* search for "Location:" lines */ 
    p=msg;
    while ((p = str_search(p, "\nLocation:")) != NULL) 
      forward_to_alternative_addr(inv, p++);
      
    /* search for "m:" lines */ 
    p=msg;
    while ((p = str_search(p, "\nm:")) != NULL) { 
      /* make sure that "m:" line was in the SIP header */
      if (p > strstr(msg, "\r\n\r\n")) 
        break;
      forward_to_alternative_addr(inv, p++);
    }
      
    /***********************************************************************************
     * call "check_response_priority" to make sure that a reply will be sent to the 
     * client if all given alternative locations were already queried and a response was 
     * received from all locations
     * if inv->locations is zero, the response with the highest priority will be sent to
     * the client
     ***********************************************************************************/
    check_response_priority(inv, msg, 0);
  }
  else {
    /***********************************************************************************
     * the received response was not an  "Alternative Address" response 
     * call "check_response_priority" which will probably send a response to the client
     ***********************************************************************************/
    check_response_priority(inv, msg, priority);
  }
} /* process_response */

