/*
 * Implements the Tcl command "servers" that expands a domain name into
 * a list of suitable servers, in the following order:
 *  (1) SRV records      (negative preferences)
 *  (2) MX  records      (MX preference from 0 to 65535)
 *  (3) A and RR records (given preference 65536)
 *
 * The routine returns a list, with each list element as a list
 *   {host ip-address preference}
 *
 * Can be generalized to any SRV record.
 *
 * Based on DomainSearch by Christian Zahl (1996);
 *   Cleaned up by Henning Schulzrinne, 1997-07-13
 */

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <sys/socket.h>
#include <malloc.h>
#include <netdb.h>
#include <ctype.h>
#include <stdlib.h>   /* qsort() */
#include <string.h>

#include "rr.h"     /* from resparse package */
#include "servers.h"

static server_t *server = 0;
static int server_entries = 0;

/* extern int res_search(char *dname, int class, int type, */
/*   unsigned char *answer, int anslen); */

/*
* Insert host if address if not a duplicate.  Update preference if lower
* one found.
*/
static void server_insert(char *host, struct in_addr addr, int preference)
{
  int i;

  for (i = 0; i < server_entries; i++) {
    if (server[i].addr.s_addr == addr.s_addr) {
      if (preference < server[i].preference)
        server[i].preference = preference; 
      break;
    }
  }
  /* no matching entry found , thus insert */
  if (i == server_entries) {
    server_entries++;   
    server = server ? realloc(server, sizeof(server_t) * server_entries) :
                    malloc(sizeof(server_t));
    server[i].host = host ? strdup(host) : NULL;
    server[i].addr = addr;
    server[i].preference = preference;
  }
} /* server_insert */


/*
 * Sort function: by preference.
 */
static int server_sort(server_t *s1, server_t *s2)
{
  if (s1->preference > s2->preference)
    return 1;
  if (s1->preference < s2->preference)
    return -1;
  return 0;
} /* server_sort */


/*
* Find the address of a host with domain name 'domain' and add it to the
* address list.
*/
static void server_addr(char *domain, int preference)
{
  static int nested = 0;
  struct hostent *h;
  char msg[PACKETSZ];
  s_rr **rr;
  res_response *res;
  int i;

  /* Allow nesting to 5 levels. */
  if (nested++ > 5) return;

  /*
   * Query for any CNAMEs and recurse.
   */
  if ((res_search(domain, C_IN, T_CNAME, msg, sizeof (msg)) != -1)
       && ((res = res_parse((char *)&msg)) != NULL)) {
    res = res_copy(res);
    rr = res->answer;
    for (i = 0; i < ntohs(res->header.ancount); i++) {
      if ((rr[i]->type == T_CNAME)
          && (rr[i]->class == C_IN)
          && (strcasecmp (domain, rr[i]->name) == 0))
        server_addr(rr[i]->rdata.string, preference);
    }
    res_free(res);
  }
  /*
   * Try A RRs next if no CNAME.
   */
  else if ((res_search(domain, C_IN, T_A, msg, sizeof (msg)) != -1)
             && ((res = res_parse((char *)&msg)) != NULL)) {
    rr = res->answer;
    for (i = 0; i < ntohs(res->header.ancount); i++) {
      if ((rr[i]->type == T_A)
          && (rr[i]->class == C_IN)
          && (strcasecmp(domain, rr[i]->name) == 0)) {
        server_insert(rr[i]->name, rr[i]->rdata.address, preference);
      }
    }
  }
  /*
   * If that fails, try gethostbyname (this also uses local resources
   *   such as NIS or /etc/hosts).
   */
  else if (((h = gethostbyname (domain)) != NULL)
            && (h->h_addrtype == AF_INET)) {
    for (i = 0; h->h_addr_list[i] != NULL; i++) {
      server_insert(h->h_name, *(struct in_addr *)(h->h_addr_list[i]),
                   preference);
    }
  }
  nested--;
} /* server_addr */


/*
* Add to address list for 'domain'.  First, try for SRV RRs, then MX,
* the finally A RRs.
*/
void server_hosts(char *domain)
{
  char msg[PACKETSZ];
  char service[1024];
  s_rr **rr;
  res_response *res;
  int i;

  /*
   * Look for SRV RRs of type 'sip' in the domain.
   */
#ifdef T_SRV
  strcpy(service, "sip.tcp.");
  strcat(service, domain);
  if ((res_search(service, C_IN, T_SRV, msg, sizeof (msg)) != -1)
      && ((res = res_parse((char *)&msg)) != NULL)) {
    res = res_copy(res);
    rr = res->answer;
    for (i = 0; i < ntohs(res->header.ancount); i++) {
      if ((rr[i]->type == T_SRV) && (rr[i]->class == C_IN)
          && (strcasecmp(service, rr[i]->rdata.srv.service) == 0))
        server_addr(rr[i]->rdata.srv.target, -rr[i]->rdata.srv.priority);
    }
    res_free(res);
  }
#endif

  /*
   * Look for MX RRs in this domain.
   */
  if ((res_search(domain, C_IN, T_MX, msg, sizeof (msg)) != -1)
       && ((res = res_parse((char *)&msg)) != NULL)) {
    res = res_copy(res);
    rr = res->answer;
    for (i = 0; i < ntohs(res->header.ancount); i++) {
      if ((rr[i]->type == T_MX) && (rr[i]->class == C_IN)
          && (strcasecmp(domain, rr[i]->name) == 0)) {
        server_addr(rr[i]->rdata.mx.exchange, rr[i]->rdata.mx.preference);
      }
    }
    res_free(res);
  } /* if */
  /*
   * Finally, we take any host with that name, but give it the lowest
   * preference.
   */
  server_addr(domain, 65536);
} /* server_hosts */


/*
* External interface; return list of server entries, with last entry
* having a host value of NULL.
*/
server_t *servers(char *domain)
{
  struct in_addr addr;

  server = 0;
  server_entries = 0;
  server_hosts(domain);
  qsort(server, server_entries, sizeof(server_t), server_sort);
  server_insert(NULL, addr, 0);  /* insert terminator */
  return server;
} /* servers */
