/***********************************************************************************
 *
 * functions to perform UDP communication
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus
 *
 ***********************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <syslog.h>
#include <sys/types.h> 
#include <netinet/in.h> 
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/uio.h>
/*#include <sys/varargs.h>*/
#include <stdarg.h>
#include "debug.h"
#include "defines.h"



extern int     isc_socket;
extern int     udp_server_socket;

typedef unsigned char u_int8;



/* UDP_connect ********************************************************************
 *
 *  create a new UDP socket, uni- or multicast
 *
 *  in:
 *  - address: uni- or multicast address of new socket
 *  - port: port of new socket 
 *  - ttl: ttl value of multicast group
 *  
 *  returns:
 *  - new UDP socket
 *  - -1, on failure
 *
 ***********************************************************************************/
int UDP_connect (char *address, int port, u_int8 ttl)
{
  static char           func[] = "UDP_connect";
  unsigned long         addr;
  int                   one=1;
  u_int8                loop=1;
  int                   size = SOCK_BUF_SIZE;
  int                   UDPsocket;
  struct sockaddr_in    sin;
  struct ip_mreq        mreq;

  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)

  if ((UDPsocket = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    return(-1);
  }

  if (setsockopt (UDPsocket, SOL_SOCKET , SO_REUSEADDR, (char *)&one, sizeof (one)) < 0) {
    close (UDPsocket);
    return(-1);
  }

  if (setsockopt (UDPsocket, SOL_SOCKET,SO_SNDBUF,(char *)&size, sizeof(int)) < 0) {
    close (UDPsocket);
    return(-1);
  }

  if (setsockopt (UDPsocket, SOL_SOCKET,SO_RCVBUF,(char *)&size, sizeof(int)) < 0) {
    close (UDPsocket);
    return(-1);
  }

  if ((addr = inet_addr(address)) == -1) {
    close (UDPsocket);
    return(-1);
  }

  if (IN_MULTICAST(htonl(addr)) > 0) {
    mreq.imr_multiaddr.s_addr = addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);

    if (setsockopt(UDPsocket, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char *)&mreq, sizeof(struct ip_mreq)) < 0) {
      close (UDPsocket);
      return(-1);
    }

    if (setsockopt(UDPsocket, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&loop, sizeof(u_int8)) < 0) {
      close (UDPsocket);
      return(-1);
    }

    if (setsockopt(UDPsocket, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(u_int8)) < 0) {
      close (UDPsocket); 
      return(-1);
    }
  }   

  sin.sin_addr.s_addr = addr;
  sin.sin_port = htons(port);

  if (bind(UDPsocket, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    close (UDPsocket);
    return(-1);
  }
  return(UDPsocket);
} /* UDP_connect */ 



/* UDP_send ***********************************************************************
 *
 *  sends UDP packets
 * 
 *  in:
 *  - address: IP address to send to
 *  - port: port to send to
 *  - arg_num: number of strings of which UDP packet will be build
 *  - ... arg_num numbers of strings to send
 *  
 *  returns:
 *  -  bytes sent
 *
 ***********************************************************************************/
int UDP_send (char *address, int port, int socket, int arg_num, ...)
{
  static char            func[] = "UDP_send";
  int                    i;
  va_list                ap;
  unsigned long          addr;
  struct sockaddr_in     sin;
  struct iovec           io[arg_num];
  struct msghdr          msg;


  DEBUG (2, syslog (LOG_DEBUG, "%s", func);)

  addr = inet_addr(address);
  sin.sin_addr.s_addr = addr;
  sin.sin_port = htons(port);
  sin.sin_family=AF_INET;

  /*ap =(char *) &__builtin_va_alist; */
  va_start(ap, arg_num);
  for (i=0; i < arg_num; i++) {
    io[i].iov_base = va_arg(ap, char *);
    io[i].iov_len = strlen(io[i].iov_base);
  }
  /* ((void)0) ;*/
  va_end(ap);
  bzero(&msg,sizeof(msg)); 

  msg.msg_iov = io;
  msg.msg_iovlen = arg_num;
  msg.msg_name = (caddr_t)&sin;
  msg.msg_namelen = sizeof(sin);

  return (sendmsg(socket, &msg, 0));
} /* UDP_send */



/* UDP_read ***********************************************************************
 *
 *  read incoming UDP messages
 *
 *  in:
 *  - fd: socket to read from
 *  - port: pointer to int to store the port from which message was read
 *  - addr: pointer to char to store the address from which message was read
 *  
 *  returns:
 *  - pointer to incoming message
 *  - NULL on failure
 *
 ***********************************************************************************/
char *UDP_read (int fd, int *port, char *addr)
{
  static char          func[] = "UDP_read";
  char                 *msg;
  int                  len;
  int                  l;
  struct sockaddr_in   sin;


  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)

  /* get some memory for holding the message */
  if ((msg = (char *)malloc (MSG_BUF_SIZE)) == NULL) 
    return NULL;
    
  len = sizeof(struct sockaddr);
  if ((l = recvfrom(udp_server_socket, msg, MSG_BUF_SIZE, 0,(struct sockaddr *)&sin, &len)) < 0) {
    free(msg);
    return NULL;
  }
  msg[l] = '\0';
  *port = ntohs(sin.sin_port);  
  strcpy(addr, inet_ntoa((struct in_addr)sin.sin_addr));
  DEBUG (3, syslog (LOG_DEBUG, "%s msg: $s",func, msg);)
  return msg;
} /* UDP_read */




